/* Here are the #define directives that control most of parameters not modifiable in runtime, and the global config variables */
#pragma once
#include <vector>

namespace smile{

typedef std::vector<double> vectord;
typedef std::vector<float> vectorf;
typedef std::pair<double, double> paird;
typedef std::vector< paird > vectorpd;

// actually not a constant but an inline function
inline double sqr(double x) {return (x)*(x);}
// number of significant digits for initial conditions
#define NUM_DIGITS_ROUND 6

#define VERSION "1.1"
/*----------- Parameters for potential and force evaluation -------------*/

// Dimensions of the real world. DO NOT CHANGE (unless you are God)!
#define N_DIM 3

// Black hole smoothing radius (for calculation of potential and forces)
#define BH_SMOOTH 1e-8

// max number of basis-function expansion for Dehnen potential (radial and angular). 
// Note: angular part is expanded in even "l" and "m", so the effective max.harmonic is twice the value below
#define MAX_NCOEFS_RADIAL 20
#define MAX_NCOEFS_ANGULAR 10
// max number of spherical-harmonic expansion terms for scale-free potential (again, only even harmonics are used)
#define MAX_NCOEFS_SF_ANGULAR 10

// whether to use exact or sph-harm.approximated density in Dehnen/Plummer BSE
//#define RHO_DEHNEN_EXACT

/*------------- Orbit integration parameters ----------*/

// relative and absolute accuracy parameters to 8th order Runge-Kutta integrator, default
#define ACCURACY_REL 1e-10
#define ACCURACY_ABS 1e-20

// with black hole, if radius is less than Mbh*{const below}, increase accuracy parameter in integrator (to make timestep smaller)
#define BH_RADIUS_INCREASE_ACCURACY 1e-2

// for Nbody treecode force calculation demand much less accuracy.. default factor in timestep selection
#define TREECODE_TIMESTEP_FACTOR 0.25

// whether to use time-symmetrized leapfrog (Hut, Makino & McMillan 1995) - doubles the number of evaluations but better conserves energy; if turned off, time-symmetrization is still used, but only for r<2*epsilon (otherwise energy conservation degrades severely)
//#define TREECODE_SYMMETRIZE_TIMESTEP

// whether to use quadrupole moments in tree expansion of Nbody potential (no reason to switch it off)
#define TREECODE_QUADRUPOLE

// default value for treecode smooothing length
#define TREECODE_DEFAULT_EPS 0.01

// default value for tree cell opening angle
#define TREECODE_DEFAULT_THETA 0.5

// density estimate in treecode by using distance to k-th neighbour. If using spline, put more (>~20), since this kernel is more concentrated
#define TREECODE_DENSITY_KTH_NEIGHBOUR 32

// method to smooth density: standard SPH spline, or Epanechnikov kernel (both have finite support equal to distance to Kth neighbour)
//#define TREECODE_DENSITY_SPLINE

// if LYAPUNOVVAREQ is defined, orbit integration uses variational equation for deviation vector.
// if not, a nearby trajectory is integrated. Note: variational equation is currently not implemented for scale-free potential
//#define LYAPUNOVVAREQ

// initial value of orbit separation (if variational equation is not used and two nearby trajectories are followed)
#define LYAP_DEV_INIT 1e-10

// if distance between nearby trajectories reaches the following threshold, renormalize it back to LYAP_DEV_INIT
#define LYAP_DEV_MAX 1e-6

// if variational equation is used, perform renormalization of deviation vector if greater than threshold (only for numerical convenience, since the var.equation is anyway linear_
#define LYAP_VAREQ_DEV_MAX 1e5

/*------------- Orbit spectral analysis and classification options ----------*/
// orbit spectrum is calculated by fast fourier transform from GSL.
// to avoid worst-case behavior of FFT, we ensure that its length never contains primes larger that MAX_FFT_PRIME (here equal to 42 minus unity)
#define MAX_FFT_PRIME 41

// Use Hunter's method for precise determination of frequency of a given line if the spectrum neat its maximum is reasonably similar to that of a single line. Otherwise use less precise method without Hanning filtering. Controls threshold for selection.
#define FREQ_PRECISE_THRESHOLD 5.0

// maximal number of spectral lines in each coordinate to search for
#define MAX_SPECTRAL_LINES 15

// do not use lines which amplitude is X times lower than amplitude of the leading line 
#define MIN_SPECTRAL_LINE_AMPLITUDE 1e-2

// method to estimate freq.diffusion rate: 0 - difference in the largest of three frequencies (Valluri&Merritt 98), 
// 1 - average of differences in all three freq, 2 - average over those coords for which |f1-f2|/(f1+f2)<0.5 (more robust to spurious errors in lf determination)
#define FREQ_DIFF_METHOD 2

// accuracy of frequency comparison (to determine resonances), in units of Nyquist frequency
#define FREQ_ACCURACY 0.1

// minimum value of freq.diff threshold separating regular and chaotic orbits (actual threshold depends on integration interval)
#define FREQ_DIFF_REG_MIN 1e-6

// maximum order of m:n frequency resonance (or linear dependence of three frequencies)
#define FREQ_RES_2 20

// maximum order of l*a+m*b+n*c frequency commensurability (thin orbit)
#define FREQ_RES_3 10

// search for leading frequency in X coordinate which is not less than MIN_FREQ_X
#define MIN_FREQ_X 0.98
// search for leading freq. in Y and Z which are not less than MIN_FREQ_YZ * (leading frequency in X and Y correspondingly)
#define MIN_FREQ_YZ 0.99

// threshold value of |<L>| / <|L|> to classify orbit as tube
#define OC_TUBE_AVGL 0.9

// threshold value of |<z>| / rmax to classify orbit as pyramid
#define OC_PYRAMID_AVGZ 0.1

// max number of points in Poincare section
#define MAX_POINCARE_POINTS 10000

// default delta-threshold controlling Pfenniger's iterative method for computing celltimes
#define INTTIME_ADAPTIVE_THRESHOLD 0.05

/*-------------  Schwarzschild modelling options ---------------*/

// if using Schwarzschild stationary start space (3d), put points on a square grid in 3 sectors adjacent to X, Y and Z axes. (same as in Schwarzschild 1993)
// otherwise, put points along lines of constant Z (=constant theta), number of points being proportional to sin(theta). (same as in Terzic 2002)
#define SCHWARZSCHILD_START_SPACE

// if checked, use principal-plane start-space confined to annulus with inner radius of closed loop orbit; otherwise put points in the whole sector
#define ANNULUS_PP_START_SPACE

// calculation of cell mass is done by refining cell into AxAxR smaller cells and computing density in each node. Below specify number of subdivisions
#define NUM_SUBDIV_ANGULAR 5
#define NUM_SUBDIV_RADIAL 10

// max. relative difference in cell mass to consider the cell as feasible: abs(1-Mcell/Mcellrequired)
#define SCHW_MAX_CELLMASS_REL_DIFFERENCE 1e-2

// min. orbit weight to use the orbit after optimization problem solved or iterations finished (divide by number of orbits!)
#define SCHW_MIN_ORBIT_WEIGHT_USED 1e-4

// weight of quadratic terms in objective function
#define SCHW_CONSTRAINT_QUAD_WEIGHT 0.02

// weight of anisotropy constraints in objective function, (weight of orbit-cell-mass constraint is 1.0)
#define SCHW_CONSTRAINT_BETA_WEIGHT 0.01

// min. and max. number of Lucy iterations
#define SCHW_MIN_ITERATIONS 10
#define SCHW_MAX_ITERATIONS 50000
// number of Lucy iterations during which the residual is allowed to grow (if more, then stop and use the best values for weights)
#define SCHW_MAX_GROWING_ITERATIONS 20
// number of iterations to update weights
#define SCHW_ITERATIONS_UPDATE 100

// number of sampling points per orbit (for initializing N-body initial conditions), default (may happen to be greater for some orbits, which then are reintegrated with larger number of sampling points)
#define NUM_SAMPLING_POINTS 1000

// how to perturb initial conditions for orbits in bunch (position only; relative difference)
#define BUNCH_PERTURB_FACTOR 1e-3

// Nbody export: relative importance of nullifying total angular momentum at the expense of pericenter velocity * pericenter offset
#define NBEXPORT_XV_FACTOR 10

/*-------------- Options for GUI and high-level operations ------------*/

// number of points in curve representing equipotential surface (actually *4)
#define NUM_POINTS_PLOT_EQUIPOTENTIAL 100

// number of points in outer bounding curve in Poincare section (actually *4)
#define NUM_POINTS_PLOT_POINCARE 100

// min. weight of orbit to be displayed in GUI
#define SCHW_MIN_ORBIT_WEIGHT_DISPLAYED 1e-4

// width of frequency spectrum displayed (in units of orbital frequency)
#define PLOT_FREQ_MAX_F 5.0

// whether to use 3d plotting library Qwt3D in GUI
#define USEQWT3D

#ifdef USEQWT3D
// whether to perform triangulation and display 'envelope' of an orbit (useless without Qwt3D; requires external program QDelaunay)
#define USE3DFACETS
#endif

// auto-load settings from file
#define DEFAULT_SETTINGS_FILE_NAME "smile.ini"

/*-------------- Global configuration data, stored in ini file, used in many subroutines ----------*/
struct CConfigData // global variables read from ini file
{
    // integration accuracy
    double accuracyRel;              // default: ACCURACY_REL
    double accuracyAbs;              // ACCURACY_ABS
    double accTreecode;              // TREECODE_TIMESTEP_FACTOR
    double nbpotEps;                 // TREECODE_DEFAULT_EPS
    double nbpotTol;                 // TREECODE_DEFAULT_THETA
    // orbit integration data
    double intTimeInPeriods;         // orbit integration time measured in units of long-axis-orbit periods
    double intTimeMaxAdaptive;       // upper limits for adaptive integration time, expressed in energy-dependent orbital time periods
    double adaptiveTimeThreshold;    // if adaptive int.time is used (with Pfenniger's method), controls max difference in cell times for subsequent intervals (default INTTIME_ADAPTIVE_THRESHOLD)
    double intTimeStepsPerPeriod;    // number of timesteps per unit time described above
    bool calcLyapunov;               // whether to calculate max Lyapunov exponent along with trajectory integration
    bool usePS;                      // whether to plot points on Poincare section
    // data for frequency map
    int fm_numOrbitsStationary;      // number of orbits in stationary start-space
    int fm_numOrbitsPrincipalPlane;  // number of orbits in principal-plane start-space
    int fm_numOrbitsYalpha;          // number of orbits in Y-alpha start space
    int fm_numOrbitsRandom;          // number of orbits in random(ergodic) start-space
    // data for Schwarzschild modelling
    int sm_numOrbitsStationary;      // same as above
    int sm_numOrbitsPrincipalPlane;
    int sm_numOrbitsYalpha;
    int sm_numOrbitsRandom;
    int sm_numShells;                // grid size: number of radial shells
    int sm_linesPerSegment;          // grid size: number of lines dividing each of 3 segments of each shell into cells
    double chaoticMinFreqDiff;       // minimal value of frequency diffusion rate to consider orbit as chaotic
    double chaoticMinLambda;         // minimal value of Lyapunov indicator to consider orbit as chaotic (in OR relation with previous criterion)
    bool chaoticSuperOrbit;          // if true, average all chaotic orbits into one super-orbit; if false, treat them individually
    double chaoticWeightFactor;      // weight factor for chaotic (super)orbit in objective function: +1 - disfavor them, -1 - favor, 0 - treat equally as regular
    int maxShellSuperOrbit;          // max number of shell for which all chaotic orbits will be averaged into super-orbit
    double sm_maxWeight;             // upper bound on orbit weight
    int sm_numBunchOrbits;           // number of perturbed orbits in a bunch to average, default 1 (no perturbed orbits)
    int sm_numSamplingPoints;        // number of points sampled from trajectory, default NUM_SAMPLING_POINTS
    bool sm_exportNbodyOrbitsFile;   // whether to export nbody model to a text file in addition to NEMO binary snapshot
    bool sm_constrainBeta;           // whether to constrain anisotropy
    double sm_betaIn, sm_betaOut;    // anisotropy coefficient requested (values for the inner and outer shells, linearly interpolated in between)
    double sm_randSigmaXIn;          // for random IC: value of vel.disp.in the center relative to that of spherical model, in X direction
    double sm_randSigmaXOut;         // same for the outer shell; linearly interpolated in between
    double sm_randSigmaYIn;          // same for other coordinates
    double sm_randSigmaYOut;
    double sm_randSigmaZIn;
    double sm_randSigmaZOut;
    double sm_randRefineFrac;        // for random sampling in energy: increase number of orbits beyond SCHW_RAND_REFINE_FRAC in Lagrangian radius by factor of SCHW_RAND_REFINE_FACT
    double sm_randRefineFact;
    bool sm_useBPMPD;                // whether to use BPMPD solver (if the executable is present), otherwise use GLPK
};

extern CConfigData config;           // global variable holding ALL configuration data (except for potential)
}