#include <QFile>
#include <QTextStream>
#include <QStringList>
#include <QMap>
#include <fstream>
#include <algorithm>
#define _USE_MATH_DEFINES
#include <cmath>
#include <gsl/gsl_fft_real.h>
#include <gsl/gsl_fft_halfcomplex.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_hyperg.h>

#include "orbit.h"

namespace smile{

////-----------------   Below follows COrbit class  -------------////
COrbit::COrbit(const CPotential* _potential, const CSpatialGrid* _grid, double _ip[N_DIM], double _iv[N_DIM], double _timeStep, double _timeUnit, bool _calcLyapunov, bool _usePS, int _PS1, int _PS2):
potential(_potential), grid(_grid), timeStep(_timeStep), timeUnit(_timeUnit), N_dim(_potential->N_dim), calcLyapunov(_calcLyapunov), usePS(_usePS), PS1(_PS1), PS2(_PS2)
{
    ip[2]=fp[2]=iv[2]=fv[2]=0;   // if N_dim<3, Z and Vz will remain zero
    for(int c=0; c<N_dim; c++)
    {
        fp[c]=ip[c]=_ip[c];
        fv[c]=iv[c]=_iv[c];
    }
    intTime=0;
    nActual=nPoints=0;
    hout=0;  // timestep size - autodetect
    Einit = potential->Phi(ip[0], ip[1], ip[2]) + (sqr(iv[0]) + sqr(iv[1]) + sqr(iv[2]))/2;
    Edif=0;
    amz=aamz=amzrms=0;
    amx=aamx=amxrms=0;
    minr=sqrt(sqr(ip[0])+sqr(ip[1])+sqr(ip[2]));
    maxr=minr;
    nperi=0;
    if(calcLyapunov)
    {
        needrenorm=allowrenorm=false;
        // initialize deviation vector
        double w2=0;
        for(int k=0; k<N_dim*2; k++) 
        { 
            devvec[k]=rand()*1.0/RAND_MAX; 
            w2+=sqr(devvec[k]); 
        }
        w2=sqrt(w2);
#ifdef LYAPUNOVVAREQ
        lnwadd=0;
        for(int k=0; k<N_dim*2; k++) devvec[k]/=w2;
#else
        lnwadd = -log(LYAP_DEV_INIT);
        for(int k=0; k<N_dim; k++) 
        {
            devvec[k]=ip[k] + devvec[k]/w2 * LYAP_DEV_INIT;
            devvec[k+N_dim]=iv[k] + devvec[k+N_dim]/w2 * LYAP_DEV_INIT;
        }
#endif
    }
    if(grid)
    {
        celltimes.resize(grid->cellsInModel+1, 0);
        cellVr.resize(grid->cellsInModel+1, 0);
        cellVt.resize(grid->cellsInModel+1, 0);
        cellprev = grid->whichCell(ip[0],ip[1],ip[2]);
    }
}

COrbit::~COrbit(void)
{
}

// simple routine that tests if an integer is prime (in the FFT primes are great evil)
void COrbit::tuneNonPrime(int &n)
{
    int m=n, f=2;
    while(f*f<=m)
        if(m%f==0) m/=f; else f++;
    if(m>=MAX_FFT_PRIME && n>=MAX_FFT_PRIME)
    {
        n--;
        tuneNonPrime(n);
    }
}

// perform FFT of orbit (output amplitudes to 'fa' array) and find 'maxnlines' lines, which will be placed into 'sl' array. 
// If fa=NULL, do not output frequency amplitudes
void COrbit::findLines(vectord::const_iterator traj[N_DIM], int npoints, Spectrum *sl, vectord *fa)
{
    if(npoints==0) return;
    tuneNonPrime(npoints);
    vectord in(npoints);
    gsl_fft_real_wavetable * wt = gsl_fft_real_wavetable_alloc (npoints);
    gsl_fft_real_workspace * ws = gsl_fft_real_workspace_alloc (npoints);
	std::vector<complexd> fc(npoints/2+1);  // temp.array of complex amplitudes

    for(int c=0; c<N_dim; c++)
    {
		vectord::const_iterator ti=traj[c];
        in.assign(traj[c], traj[c]+npoints);
        gsl_fft_real_transform (&(in.front()), 1, npoints, wt, ws);
        fc[0]=complexd(in[0],0)  / (0.5*npoints);
        for(int i = 1; i < npoints/2; i++)
            fc[i]=complexd(in[2*i-1], in[2 * i])/(0.5*npoints);
        if(npoints%2==0)
            fc[npoints/2]=complexd(in[npoints-1], 0)/(0.5*npoints);
        if(fa!=NULL)
        {  // reserve space in vector
            fa[c].resize(npoints/2+1, 0);
            for(int i=0; i<npoints/2+1; i++)
                fa[c][i]=abs(fc[i]);    // store amplitude of spectrum if needed (to display only) - and then forget it; only the true complex spectrum is used hereafter
        }

        // perform frequency analysis
        sl[c].clear();
        bool finish=false;
        double maxamp=0; double ma=0;
        while(!finish)
        {   // find next greatest line
            int m=0; 
            ma=0;
            for(int i=0; i<npoints/2-2; i++)
                if(abs(fc[i])>ma)
                {
                    ma=abs(fc[i]);
                    m=i;
                }
            if(maxamp==0) maxamp=ma;
            if(ma>maxamp*MIN_SPECTRAL_LINE_AMPLITUDE)
            {   // add line
                // peak is somewhere near m.  
                if(m>1)
                {
                    int m1=m;
                    if(abs(fc[m-1])>abs(fc[m+1]))  m1--;
                    //int m2=((abs(fc[m-1])>abs(fc[m+1]))?(m-1):(m+1));
                    //if(m2<m1) { int mm=m2; m2=m1; m1=mm; }   // not necessary. but convenient to have m2=m1+1
                    double s;   // fractional index of peak frequency 
                    complexd Ac;// amplitude
                    complexd f1 = (fc[m1] - (fc[m1-1]+fc[m1+1])/2.0);
                    complexd f2 = (fc[m1+1] - (fc[m1]+fc[m1+2])/2.0);
                    // C=f2/f1; check out that C<>0 and C<>infinity
                    if(abs(f1) < abs(f2)*1e-4)
                    {   
                        m1++;
                        f1=f2;
                        f2 = (fc[m1+1] - (fc[m1]+fc[m1+2])/2.0);
                    }
                    if((abs(f2) < abs(f1)*1e-4) && (m1>0))
                    {   
                        m1--;
                        f2=f1;
                        f1 = (fc[m1] - (fc[m1-1]+fc[m1+1])/2.0);
                    }
                    complexd C = f2 / f1;
                    double Ca=C.real();  // in ideal case, Im C=0
                    double omega=2*M_PI/npoints;
                    bool Precise=true;
                    if((Ca<-1.0/FREQ_PRECISE_THRESHOLD) && (Ca>-FREQ_PRECISE_THRESHOLD)/*abs(C.imag())*0 < abs(Ca)*/)
                    {   // try Hunter's precise method
                        double sgn = (Ca<1)?+1:-1;
                        double t = (Ca*cos(omega)-1 + sgn * sqrt( sqr(Ca*cos(omega)-1) - (1+2*Ca)*sqr(sin(omega))))/sin(omega);
                        s = m1 + npoints/M_PI * atan(t);
                        complexd z = std::polar(1.0, 2*M_PI*(s-m1)/npoints); //complexd((1-sqr(t))/(1+sqr(t)), 2*t/(1+sqr(t)));
                        complexd beta = std::polar(1.0, omega);
                        complexd P = (std::polar(1.0, 2*M_PI*s)-1.0) / (std::polar(1.0, 2*M_PI*(s-m1)/npoints)-1.0) / (1.0*npoints);
                        complexd g1 = P * (1.0 - 0.5*(z-1.0)*(1.0/(z*beta-1.0) + 1.0/(z*std::conj(beta)-1.0)));
                        Ac = f1/g1;
                    }
                    else
                    {   // Follow the method by Carpintero&Aguilar(1998)
                        Precise=false;
                        double alpha = atan(sin(omega/2)/(cos(omega/2)+abs(fc[m1]/fc[m1+1])));
                        s = m1 + npoints*alpha/M_PI;
                        complexd P=(std::polar(1.0, 2*alpha*npoints)-1.0) / (std::polar(1.0, 2*alpha)-1.0) / (1.0*npoints);
                        complexd Q=(std::polar(1.0, -2*M_PI*(s+m1))-1.0) / (std::polar(1.0, -2*M_PI*(s+m1)/npoints)-1.0) / (1.0*npoints);
                        Ac=fc[m1] / P;
                    }
                    sl[c].push_back(SpectralLine(s*timeUnit/(npoints*timeStep)/*intTime*/, abs(Ac), std::arg(Ac), Precise));

                    if(sl[c].size()>=MAX_SPECTRAL_LINES) finish=true;    // not to iterate forever

                    // subtract line from spectrum
                    if(!finish)
                    {
                        if(s-floor(s) < 1e-5)    // frequency was very close to integer value, so subtract only this point
                        {
                            fc[(int)floor(s)]-=Ac;
                        } else
                        if(s-floor(s) > 1-1e-5)
                        {
                            fc[(int)floor(s)+1]-=Ac;
                        }
                        else   // subtract this line with its wings from the whole spectrum
                        {
                            complexd Amul = Ac / (1.0*npoints) * (std::polar(1.0, 2*M_PI*(s))-1.0);
                            complexd Acmul = std::conj(Ac) / (1.0*npoints) * (std::polar(1.0,-2*M_PI*(s))-1.0);
                            int jmin=std::max(0, m-1000);
                            int jmax=std::min(npoints/2+1, m+1000);
                            for(int j=jmin; j<jmax; j++)
                            {
                                complexd linea = Amul / (std::polar(1.0, 2*M_PI*(s-j)/npoints)-1.0) +
                                    Acmul / (std::polar(1.0,-2*M_PI*(s+j)/npoints)-1.0);
                                fc[j] -= linea;
                            }
                        }
                    }
                }
                else
                {   // zero frequency
                    fc[m]=0;
                }
            }
            else finish=true;
        }
        std::sort(sl[c].begin(), sl[c].end());
        if(sl[c].size()==0)
            sl[c].push_back(SpectralLine());  // zero
        int sw=0; SpectralLine ftmp;
        switch(c)   // ensure that the first line in each coordinate has frequency >1 (for X), >X (for Y), and >Y (for Z); tolerate small deviations given in MIN_FREQ_*
        {
        case 0:
            while((sw<sl[0].size()-1) && (sl[0][0].freq<MIN_FREQ_X)) 
            { ftmp=sl[0][sw+1]; sl[0][sw+1]=sl[0][0]; sl[0][0]=ftmp; sw++; }
            break;
        case 1:
            while((sw<sl[1].size()-1) && (sl[1][0].freq<MIN_FREQ_YZ*sl[0][0].freq)) 
            { ftmp=sl[1][sw+1]; sl[1][sw+1]=sl[1][0]; sl[1][0]=ftmp; sw++; }
            break;
        case 2:
            if(N_dim==3) while((sw<sl[2].size()-1) && (sl[2][0].freq<MIN_FREQ_YZ*sl[1][0].freq)) 
            { ftmp=sl[2][sw+1]; sl[2][sw+1]=sl[2][0]; sl[2][0]=ftmp; sw++; }
            break;
        }
    }
    gsl_fft_real_wavetable_free (wt);
    gsl_fft_real_workspace_free (ws);
}

void simplify(int &n1, int &n2)    // remove common factors
{
    int nmin=std::min(n1, n2);
    int f=1;
    for(int k=1; k<nmin/2; k++)
        if((n1%k==0) && (n2%k==0)) f=k;
    n1/=f;
    n2/=f;
}

bool commensurable(double a, double b, double acc, int &na, int &nb)   // check that a:b ~=~ na:nb
{
    if(a*b==0) return(false);
    na=1; nb=0;
    while(na<=FREQ_RES_2)
    {
        int nb1=floor(b/a*na+.5);
        if(nb1 && (fabs(nb1*a-na*b) < acc))
        {
            nb=nb1;
            return true;
        }
        else na++;
    }
    return false;
}

bool lineardepend(double f, double a, double b, double acc, int &ma, int &mb)   // check linear dependency: f ~=~ ma*a + mb*b
{
    ma=mb=0; 
    while(abs(ma)<=FREQ_RES_2)
    {
        mb=floor((f-ma*a)/b+0.5);
        if(fabs(ma*a+mb*b-f) < acc) 
            return true;
        else 
            if(ma>0) ma=-ma; else ma=-ma+1;
    }
    return false;
}

bool lineardepend3(double f, double a, double b, double c, double acc, int &ma, int &mb, int &mc)   // check 3d linear dependency: f ~=~ ma*a + mb*b + mc*c
{
    ma=mb=mc=0; 
    while(abs(ma)<=FREQ_RES_2)
    {
      mb=0;
      while(abs(mb)<=FREQ_RES_2)
      {
        mc=floor((f-ma*a-mb*b)/c+0.5);
        if(fabs(ma*a+mb*b+mc*c-f) < acc) 
            return true;
        else 
            if(mb>0) mb=-mb; else mb=-mb+1;
      }
      if(ma>0) ma=-ma; else ma=-ma+1;
    }
    return false;
}

bool commensurable3(double a, double b, double c, double acc, int &ma, int &mb, int &mc)
{
    for(int na=1; na<=FREQ_RES_3; na++)
        for(int nb=-FREQ_RES_3; nb<=FREQ_RES_3; nb++)
            for(int nc=-FREQ_RES_3; nc<=FREQ_RES_3; nc++)
                if(fabs(a*na+b*nb+c*nc) < acc)
                {
                    ma=na; mb=nb; mc=nc;
                    return true;
                }
    return false;
}

void COrbit::performFrequencyAnalysis(Spectrum *sl, vectord *fa, int start, int count)
{
    if(count<=0) count = nActual - start;   // by default both parameters are zero, so inspect whole orbit
    if(count<=1)  // empty orbit
    {
        Description="?";
        return;
    }
	vectord::const_iterator traj[N_DIM];
    for(int d=0; d<N_dim; d++)
        traj[d] = tp[d].begin()+start;
    findLines(traj, count, sl, fa);

    // calculate avg. angular momentum
    amz=amx=0;
    double avgZ=0, maxz=0;
    for(int i=start; i<start+count; i++)
    {
        double amzc = tp[0][i]*tv[1][i]-tp[1][i]*tv[0][i];
        amz+=amzc;
        aamz+=fabs(amzc);
        amzrms+=sqr(amzc);
        if(N_dim==3)
        {
            double amxc = tp[1][i]*tv[2][i]-tp[2][i]*tv[1][i];
            amx+=amxc;
            aamx+=fabs(amxc);
            amxrms+=sqr(amxc);

            if(potential->Mbh>0)
            {
                avgZ+=tp[2][i];
                if(fabs(tp[2][i])>maxz) maxz=fabs(tp[2][i]);
            }
            else maxz=1;
        }
    }
    avgZ/=count;
    amz/=count;
    aamz/=count;
    amzrms=sqrt(amzrms/count);
    amx/=count;
    aamx/=count;
    amxrms=sqrt(amxrms/count);

    // determine orbit class
    bool tubez=(fabs(amz)>aamz*OC_TUBE_AVGL);  // has definite sense of rotation (more robust indicator than 1:1 correspondence of leading freqencies)
    bool tubex=(N_dim==3) && (fabs(amx)>aamx*OC_TUBE_AVGL);
    double lfX=sl[0][0];
    double lfY=sl[1][0];
    double lfZ=(N_dim==3)?sl[2][0].freq:0;
    double acc = timeUnit/(timeStep*count) * FREQ_ACCURACY;  // accuracy of comparison
    if(acc<FREQ_DIFF_REG_MIN) acc=FREQ_DIFF_REG_MIN;  // don't be too demanding..
    if(N_dim==2)
    {
        // try to find nx:ny correspondence of leading frequencies
        int nx, ny;
        if(commensurable(lfX, lfY, acc, nx, ny))    // boxlet family or tube (in case lfX==lfY)
        {
            double lf = (lfX/nx + lfY/ny)/2;  // mean value of common denominator 
            // find additional (libration) frequency
            double lfa = 0;
            bool regular=true;
            int rank=1;
            int ma, mb;
            for(int c=0; c<N_dim; c++)
            {
                for(int i=1; regular&&(i<sl[c].size()); i++)
                {
                    if(commensurable(sl[c][i], lf, acc, ma, mb)) 
                    {
                        if(((ma==1) || (abs(ma-mb)==1)) && (mb>1))
                            rank=mb;
                    }
                    else
                    {
                        double lfa1=std::min(sl[c][i].freq, fabs(sl[c][0].freq-sl[c][i].freq));
                        if(lfa)
                        {
                            if(!lineardepend(sl[c][i], lf, lfa, acc, ma, mb)) 
                            {
                                regular=false;
                            }
                            else 
                                if(lfa1<lfa) 
                                {
                                    lfa=lfa1;
                                }
                        }
                        else
                        {
                            lfa=lfa1;
                        }
                    }
                }
            }
            Description += " "+QString::number(nx)+":"+QString::number(ny);
            if(tubez)
                Description = "Tube";
            else if((nx==1)&&(ny==2))
                Description = "banana";
            else if((nx==2)&&(ny==3))
                Description = "fish";
            else if((nx==3)&&(ny==4))
                Description = "pretzel";
            else 
                Description = QString::number(nx)+":"+QString::number(ny)+" resonant boxlet";
            Description += " rank "+QString::number(rank);
            if(!regular)
                Description = "chaotic "+Description;
        }
        else 
        {
            // for regular box orbit the remaining lines must be linear-dependent on two leading frequencies
            bool regular=true;
            int ma, mb;
            for(int c=0; c<2; c++)
                for(int i=1; regular&&(i<sl[c].size()); i++)
                    if(!lineardepend(sl[c][i], lfX, lfY, acc, ma, mb)) regular=false;
            if(regular)
                Description = "box";
            else
                Description = "chaotic";
        }
    }
    else    // N_dim==3
    {
        // determine if there are resonances between each pair of coordinates
        bool res[3];
        char resname[4] = "XYZ";
        int nres=0;       // number of resonances
        int na[3], nb[3]; // resonance quotients
        double lf[3];     // leading frequencies (1 for nres=3, 2 for nres=1, 3 for nres=0)
        int nlf;          // number of leading frequencies
        for(int d=0; d<3; d++)
        {
            res[d] = commensurable(sl[d][0], sl[(d+1)%3][0], acc, na[d], nb[d]);
            if(res[d]) nres++;
        }
        
        if(nres==2)  // check for consistency (there cannot be two resonances)
        {   // reconstruct missing resonance number dn
            int dn=!res[0] ? 0 : (!res[1] ? 1 : 2);
            na[dn]=nb[(dn+2)%3]*nb[(dn+1)%3];
            nb[dn]=na[(dn+1)%3]*na[(dn+2)%3];
            simplify(na[dn], nb[dn]);
            nres=3;
        }
        if(nres==0)
        {
            nlf=3;
            lf[0]=sl[0][0];
            lf[1]=sl[1][0];
            lf[2]=sl[2][0];
            int n1, n2, n3;
            if(commensurable3(lf[0],lf[1],lf[2],acc,n1,n2,n3))
            {
                Description = "("+QString::number(n1)+","+QString::number(n2)+","+QString::number(n3)+") thin orbit";
                nlf = 2;
            }
            else
            {
                // either box or pyramid (near black hole)
                if(fabs(avgZ/maxz) > OC_PYRAMID_AVGZ)
                    Description = "pyramid";
                else
                    Description = "box";
            }
        }
        if(nres==1)
        {   
            int de=res[0] ? 0 : (res[1] ? 1 : 2);
            nlf=2;
            lf[0] = (sl[de][0]/na[de] + sl[(de+1)%3][0]/nb[de])/2;
            lf[1] = sl[(de+2)%3][0];
            if(lf[1]==0)
                nlf=1;  // avoid degenerate cases
            switch(de)
            {
            case 0: Description = QString::number(na[0])+":"+QString::number(nb[0])+":*"; break;
            case 1: Description = "*:"+QString::number(na[1])+":"+QString::number(nb[1]); break;
            case 2: Description = QString::number(nb[2])+":*:"+QString::number(na[2]); break;
            }
            Description +=" resonance";
        }
        if(nres==3)
        {
            nlf=1;
            lf[0]=std::min((sl[0][0].freq/na[0] + sl[1][0].freq/nb[0])/2, std::min((sl[1][0].freq/na[1] + sl[2][0].freq/nb[1])/2, (sl[2][0].freq/na[2] + sl[0][0].freq/nb[2])/2));
            int nx=floor(sl[0][0]/lf[0]+0.5); 
            int ny=floor(sl[1][0]/lf[0]+0.5); 
            int nz=floor(sl[2][0]/lf[0]+0.5); 
            Description = QString::number(nx)+":"+QString::number(ny)+":"+QString::number(nz)+" resonance";
        }
        if(tubex) 
        {
            Description = "LAT "+Description;
            // try to determine whether it's inner or outer LAT
            // ..remains to be implemented
        }
        if(tubez) 
        {
            if(fabs(avgZ/maxz) > OC_PYRAMID_AVGZ)
                Description = "SAU "+Description;
            else
                Description = "SAT "+Description;
        }
        // check that every other freq is commensurable with base
        int ma, mb, mc;
        bool regular=true;
        for(int c=0; c<3; c++)
            for(int i=1; regular&&(i<sl[c].size()); i++)
            {
                switch(nlf)
                {
                case 1:
                    if(!commensurable(sl[c][i], lf[0], acc, ma, mb)) 
                    {
                        nlf++;
                        lf[1] = std::min(sl[c][i].freq, fabs(lf[0]-sl[c][i].freq));
                    }
                    break;
                case 2:
                    if(!lineardepend(sl[c][i], lf[0], lf[1], acc, ma, mb))
                    {
                        nlf++;
                        lf[2] = std::min(sl[c][i].freq, std::min(fabs(lf[0]-sl[c][i].freq), fabs(lf[1]-sl[c][i].freq)));
                    }
                    break;
                case 3:
                    if(!lineardepend3(sl[c][i], lf[0], lf[1], lf[2], acc, ma, mb, mc))
                        regular=false;
                }
            }
        if(!regular)
            Description = "chaotic "+Description;
    }
}

bool COrbit::calcLFCCoverInterval(int start, int count, double* sli)
{
    if(start>=nActual || sli==NULL) 
        return false;
    if(count>=nActual-start)
        count=nActual-start;
	vectord::const_iterator tpi[N_DIM];
    for(int c=0; c<N_dim; c++)
        tpi[c] = tp[c].begin() + start;
    Spectrum lfi[N_DIM];
    findLines(tpi, count, lfi, NULL);
    for(int c=0; c<N_dim; c++)
        sli[c]=(lfi[c].size()>0)? lfi[c][0].freq : -1;
    return true;
}

// calculate rate of diffusion of leading frequencies in cartesian coordinates (LFCCs), which are related to the degree of chaos in the orbit
// determines the LFCCs for first and second halves of orbit, computes difference, returns delta f/f averaged over all coordinates
double COrbit::calcLFCCdiffusion(int start, int count)
{
    if(count<=0) count=nActual-start;  // whole orbit by default
    int count2=count/2;
    double lf1[N_DIM], lf2[N_DIM];  // leading frequencies of 1st and 2nd half for each coordinate
    calcLFCCoverInterval(start, count2, lf1);
    calcLFCCoverInterval(start+count2, count2, lf2);
    double FreqDiff = 0;
#if FREQ_DIFF_METHOD == 0
    for(int c=N_dim-1; c>=0 && FreqDiff==0; c++)
      if(lf1[c] + lf2[c] > 0 )
        FreqDiff = fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]);
#else
    int nf=0;
    for(int c=0; c<N_dim; c++)
        if(lf1[c] + lf2[c] > 0 )
#if FREQ_DIFF_METHOD == 2
          if(fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]) < 1)
#endif
        {
            FreqDiff += fabs(lf1[c] - lf2[c]) * 2 / (lf1[c] + lf2[c]);
            nf++;
        }
    if(nf>0) FreqDiff /= nf;
#endif
    double maxFreqDiff = sqr(timeUnit/intTime);  // rough estimate of regular/chaotic divide
    if(maxFreqDiff<FREQ_DIFF_REG_MIN) maxFreqDiff=FREQ_DIFF_REG_MIN;
    // here one may also add the estimate based on energy conservation error (deltaE) and associated minimum frequency error d(Torbx^-1)/dE * (deltaE)

    if(FreqDiff<1*maxFreqDiff)
        DescriptionLFCCdiff = "regular";
    else 
    if(FreqDiff<2*maxFreqDiff)
        DescriptionLFCCdiff = "probably regular";
    else 
    if(FreqDiff<4*maxFreqDiff)
        DescriptionLFCCdiff = "probably chaotic";
    else 
        DescriptionLFCCdiff = "chaotic";
    return FreqDiff;
}

bool COrbit::calcInertiaOverInterval(int start, int count, double* inert)
{
    if(start>=nActual || inert==NULL) 
        return false;
    if(count>=nActual-start)
        count=nActual-start;
    for(int c=0; c<N_dim; c++)
        inert[c] = 0;
    for(int i=0; i<count; i++)
    {
        for(int c=0; c<N_dim; c++)
            inert[c] += sqr(tp[c][i+start]);
    }
    for(int c=0; c<N_dim; c++)
        inert[c] = sqrt(inert[c]/count);
    return true;
}

// calculates maximal Lyapunov characteristic exponent (if the deviation equation was integrated along with orbit)
double COrbit::calcLambda(void)
{
    if(!calcLyapunov)
        return 0;
    double sumavg=0;     // to calculate average value of ln(|w|/t) over the linear-growth interval
    int cntavg=0;
    double sumlambda=0;  // to calculate average value of ln(|w|)/t over the exponential growth interval
    int cntlambda=0;
    int timeblock=timeUnit/timeStep*2;    // averaging interval (roughly orbital period * 2)
    std::vector<double> interval(timeblock);
    bool chaotic=false;
    for(int t=timeblock; t<nActual; t+=timeblock)
    {
        int cnt = std::min(timeblock, nActual-t);
        if(!chaotic)
        {
            // compute median value over 'timeblock' points  to avoid outliers
            interval.resize(cnt);
            for(int i=0; i<cnt; i++)
                interval[i] = (lnw[i+t]-log((i+t)*timeStep));
            std::sort(interval.begin(), interval.end());
            double mid = interval[cnt/2];
            sumavg+=mid;
            cntavg++;
            if((cntavg>4) && (mid > sumavg/cntavg + 2))
                chaotic=true;     // from here on, assume that we got into t->infinity regime and start collecting data for estimating lambda
        }
        if(chaotic)
        {
            for(int i=0; i<cnt; i++)
            {
                sumlambda += lnw[i+t]/(i+t);
                cntlambda++;
            }
        }
    }
    if(chaotic)
        return (sumlambda / cntlambda * timeUnit / timeStep);
    else
        return 0;
}

double finddrdt(double t, void* params)
{
    COrbit* o=((COrbit*)params);
    int N_dim=o->N_dim;
    return o->contd8(0,t)*o->contd8(N_dim,t) + o->contd8(1,t)*o->contd8(N_dim+1,t) + (N_dim==3?o->contd8(2,t)*o->contd8(5,t):0);
}

double findPS(double t, void* params)
{
    return ((COrbit*)params)->contd8(((COrbit*)params)->PS2,t);
}

// performs regular-timestep output (store trajectory, populate Poincare section, etc). Called from dop853 integration routine
void COrbit::solout(long nr, double told, double t, double* y, unsigned n)
{
    int i1=told/timeStep; 
    int i2=t/timeStep;

    double r2=sqr(y[0]) + sqr(y[1]) + (N_dim>2 ? sqr(y[2]) : 0);
    double sqrtr2=sqrt(r2);
    if(r2<sqr(minr)) minr=sqrtr2;
    if(r2>sqr(maxr)) maxr=sqrtr2;

    // watch for pericenter passage
    int prevsign=drdtsign;
    drdtsign= y[0]*y[N_dim]+y[1]*y[N_dim+1]+(N_dim==3?y[2]*y[5]:0) >0 ? 1 : -1;
    if(prevsign>0 && drdtsign<0)  // apocenter
    {
        if(needrenorm) allowrenorm=true;
    }
    // check energy conservation
    if(i2>i1 && i2%10==0)  // do not check too frequently
    {
        double Z=(N_dim==3)?y[2]:0;
        double vZ=(N_dim==3)?y[N_dim+2]:0;
        double Ecur = potential->Phi(y[0], y[1], Z) + (sqr(y[N_dim]) + sqr(y[N_dim+1]) + sqr(vZ))/2;
        double Edifcur = fabs(Einit - Ecur);
        if(Edifcur>Edif)
            Edif=Edifcur;
    }
    
    for (int i=i1; i<=i2; i++)
    {
        double tt=timeStep*i;
        if ((tt>=told) && (tt<t) && (nActual<nPoints))
        {
            for(int k=0; k<N_dim; k++)
            {
                fp[k]=contd8(k,tt);
                tp[k].push_back(fp[k]);
                fv[k]=contd8(k+N_dim,tt);
                tv[k].push_back(fv[k]);
            }

            if(calcLyapunov)
            {
                double wx =contd8(N_dim*2,tt);
                double wy =contd8(N_dim*2+1,tt);
                double wz =(N_dim==3)?contd8(N_dim*2+2,tt):0;
                double wvx=contd8(N_dim*3,tt);
                double wvy=contd8(N_dim*3+1,tt);
                double wvz=(N_dim==3)?contd8(N_dim*3+2,tt):0;
#ifdef LYAPUNOVVAREQ
                lnw.push_back(log(sqr(wx) + sqr(wy) + sqr(wz) + sqr(wvx) + sqr(wvy) + sqr(wvz))/2 + lnwadd);
#else
                lnw.push_back(log(sqr(fp[0]-wx) + sqr(fp[1]-wy) + sqr(fp[2]-wz) + sqr(fv[0]-wvx) + sqr(fv[1]-wvy) + sqr(fv[2]-wvz))/2 + lnwadd);
#endif
            }
            nActual++;
        }
    }
    // Poincare section
    double yprev=(told>0)?contd8(PS2, told):y[PS2];  // previous y-value
    if(usePS && (ps.size()<MAX_POINCARE_POINTS) && (yprev<=0) && (y[PS2]>=0))  // put a point
    {
        if(y[PS2]>yprev)
        {
            // interpolate coordinates at Poincare section crossing
            gsl_function F;
            F.function=&findPS;
            F.params=this;
            gsl_root_fsolver *s = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
            gsl_root_fsolver_set (s, &F, told, t);
            int status=0, iter=0;
            double tcross, tlow, tupp;
            do{
                iter++;
                status = gsl_root_fsolver_iterate (s);
                tcross = gsl_root_fsolver_root (s);
                tlow = gsl_root_fsolver_x_lower (s);
                tupp = gsl_root_fsolver_x_upper (s);
                status = gsl_root_test_interval (tlow, tupp, 0, 1e-8*(t-told));
            }
            while (status == GSL_CONTINUE && iter < 100);
            gsl_root_fsolver_free (s);
            ps.push_back(paird(contd8(PS1,tcross), contd8(PS1+N_dim, tcross)));
        }
        else
        {
            ps.push_back(paird(y[PS1], y[PS1+N_dim]));
        }
    }
    if(grid)
    {
        double Lx=(N_dim==3)? y[1]*y[5]-y[2]*y[4] : 0;
        double Ly=(N_dim==3)? y[2]*y[3]-y[0]*y[5] : 0;
        double Lz=(N_dim==3)? y[0]*y[4]-y[1]*y[3] : y[0]*y[3]-y[1]*y[2];
        double L=sqrt(sqr(Lx)+sqr(Ly)+sqr(Lz));
        double vr2= sqr(y[0]*y[3]+y[1]*y[4]+y[2]*y[5])/r2;
        double vt2= sqr(L)/r2;
        if(r2==0) { vr2=sqr(y[0])+sqr(y[1])+sqr(y[2]); vt2=0; }
        int cell=grid->whichCell(fp[0], fp[1], fp[2]);
        if(cell==cellprev)
        {
            celltimes[cell] += (t-told);
            cellVr[cell] += (t-told)*vr2;
            cellVt[cell] += (t-told)*vt2;
            if(latestcelltimes.size()>0)
                latestcelltimes[cell] += (t-told);
        }
        else  // calculate fractional time spent in cell
            calcCellTime(told, t, cellprev, cell, vr2, vt2);
        cellprev=cell;
    }
}

void COrbit::calcCellTime(double tlow, double tupp, int clow, int cupp, double vr2, double vt2)
{
    double tl=tlow, tu=tupp;
    int niter=0;
    double tmid;
    do{
        tmid=(tu+tl)/2;
        int cmid=grid->whichCell(contd8(0,tmid), contd8(1,tmid), contd8(2,tmid));
        if(cmid==clow)
            tl=tmid;
        else if(cmid==cupp)
            tu=tmid;
        else
        {   // neither of the two cells - go deeper into recursion
            calcCellTime(tlow, tmid, clow, cmid, vr2, vt2);
            calcCellTime(tmid, tupp, cmid, cupp, vr2, vt2);
            return;
        }
        niter++;
    }
    while((tu-tl>1.0e-6*timeStep) && (niter<20));
    celltimes[clow] += tmid-tlow;
    celltimes[cupp] += tupp-tmid;
    cellVr[clow] += (tmid-tlow)*vr2;
    cellVr[cupp] += (tupp-tmid)*vr2;
    cellVt[clow] += (tmid-tlow)*vt2;
    cellVt[cupp] += (tupp-tmid)*vt2;
    if(latestcelltimes.size()>0)
    {
        latestcelltimes[clow] += tmid-tlow;
        latestcelltimes[cupp] += tupp-tmid;
    }
}

double COrbit::diffCellTimes()
{
    if(celltimes.size()<=0) return -1;
    if(latestcelltimes.size()<=0) return 0;
    double delta=0;
    for(int c=0; c<grid->cellsInModel; c++)
    {
        delta += fabs(celltimes[c]/intTime - latestcelltimes[c]/latestIntTime);
    }
    return delta;
}

void COrbit::integrateAdaptiveTime(double minTime, double maxTime, double adaptiveTimeThreshold)
{
    if(grid==NULL)    // used only in the context of creating Schwarzschild model orbit library
    {
        integrateToTime(minTime);
        return;
    }
    if(maxTime<minTime) maxTime=minTime;
    double interval=minTime/2;
    integrateToTime(interval);
    double delta=1;
    int numint=1;
    bool goon=true;
    while(goon)
    {
        numint++;
        interval = std::min((minTime/2)*pow(1.1, numint-2), maxTime-intTime);
        integrateToTime(intTime+interval);
        delta=diffCellTimes();
        goon = intTime < maxTime-1e-5 && delta*3.0/(2+numint)>adaptiveTimeThreshold;
    }
}

// initialize variables, then integrate using dop853 routine
void COrbit::integrateToTime(double time)
{
    int i,j,len,err;
    double t,vars[N_DIM*4];
    double tend;           /* final t-value */
    double rtoler=config.accuracyRel; /* relative error tolerance */
    double atoler=config.accuracyAbs; /* absolute error tolerance */
    if(potential->PotentialType()==CPotential::PT_NB) 
    {
        leapfrog=true;     /* treecode has discontinuous potential, so low-order leapfrog integrator performs better */
        if(N_dim!=N_DIM || calcLyapunov) return;  // allow run only in 3d without Lyapunov exp
    }
    else leapfrog=false;
    int itoler=0;          /* switch for rtoler and atoler */
    int iout=2;            /* switch for calling solout */
    double uround=0.0;     /* rounding unit */
    double safe=0.0;       /* safety factor */
    double fac1=0.0;       /* parameters for step size selection */
    double fac2=0.0;
    double beta=0.0;       /* for stabilized step size control */
    double hmax=0.0;       /* maximal step size */
    double h=hout;         /* initial step size */
    long nmax=1e8;         /* maximal number of allowed steps */
    int meth=0;            /* switch for the choice of the coefficients */
    long nstiff=0;         /* test for stiffness */
    unsigned neq;          /* number of equations */
    unsigned nrdens;       /* number of components for which dense output is required */
    unsigned licont=0;     /* declared length of icont */
    unsigned* icont=NULL;  /* indexes for which dense output is required  NULL=all */
    //FILE* fileout=stderr;    /* messages stream */

    t = intTime;  // current value 
    tend = time;  // value at the end of integration
    nPoints=(tend==0?0:ceil(tend/timeStep-1e-8));
    neq=nrdens = N_dim*2*(calcLyapunov?2:1);
    for(i=0; i<N_dim; i++)
    {
        vars[i]      = fp[i];  // current final values
        vars[i+N_dim]= fv[i];
        tp[i].reserve(nPoints);
        tv[i].reserve(nPoints);
    }
    drdtsign= vars[0]*vars[N_dim]+vars[1]*vars[N_dim+1]+(N_dim==3?vars[2]*vars[5]:0) >0 ? 1 : -1;

    if(calcLyapunov)
    {
        double w2=0;
        for(int k=0; k<2*N_dim; k++)
        {
            vars[N_dim*2+k]=devvec[k];  // stored value
#ifdef LYAPUNOVVAREQ
            w2+=sqr(devvec[k]);
#else
            w2+=sqr(devvec[k]-vars[k]);
#endif
        }
        lnw.push_back(log(w2)/2 + lnwadd);
    }
    if(grid && t>0)
    {
        latestcelltimes.clear();
        latestcelltimes.resize(grid->cellsInModel+1, 0);
    }

    dop853(neq, /*fcn,*/t, vars, tend, &rtoler, &atoler, itoler, /*solout,*/iout,/*fileout,*/
        uround, safe, fac1,fac2, beta, hmax,h , nmax, meth, nstiff, nrdens, icont, licont);
    latestIntTime = xout-t;  // duration of integration interval
    intTime = xout;  // latest time value reached in the integrator
    for(i=0; i<N_dim; i++)
    {
        fp[i]=vars[i];  // store values
        fv[i]=vars[i+N_dim];
    }
}

void COrbit::finishIntegration()
{
    if(grid)
    {
        for(int c=0; c<=grid->cellsInModel; c++)
        {
            if(celltimes[c]>0)
            {
                cellVr[c] /= celltimes[c];
                cellVt[c] /= celltimes[c];
            }
            celltimes[c] /= intTime;    // normalize to unit time
        }
        if(celltimes[grid->cellsInModel]>0 && celltimes[grid->cellsInModel]<1)
        {   // normalize to unity for cells inside model (excluding the last cell)
            double mult = 1/(1-celltimes[grid->cellsInModel]);
            for(int c=0; c<=grid->cellsInModel; c++)
                celltimes[c] *= mult;
        }
    }
}

inline void COrbit::fcn(unsigned n, double t, double *v, double *f)
{
    potential->DiffEq(n, t, v, f);
}

// leap-frog integrator for frozen-N treecode potential
int COrbit::intlf(unsigned int n, double x, double *y, double xend)
{
    double f[4*N_DIM+2];
    double hprev=0, h=hprev;
    fcn(n, x, y, f);
    naccpt=nrejct=nstep=0; nfcn=1;
    while(x<xend)
    {
        for(int d=0; d<N_DIM; d++)
        {
            yprev[d] = y[d];
            yprev[d+N_DIM] = y[d+N_DIM];
            aprev[d] = f[d+N_DIM];
        }
        xold=x;
        // first estimate timestep based on two criteria: minimum distance to a particle and maximum acceleration; these were computed during tree=walk
        double vel = sqrt(sqr(y[N_DIM])+sqr(y[N_DIM+1])+sqr(y[N_DIM+2]));
        double acc = sqrt(sqr(f[N_DIM])+sqr(f[N_DIM+1])+sqr(f[N_DIM+2]));
        double ts1 = vel>0 ? f[N_DIM*2]/vel : xend;
        double ts2 = acc>0 ? sqrt(f[N_DIM*2]/acc) : xend;
        h = std::min<double> (ts1, ts2) * config.accTreecode;
#ifndef TREECODE_SYMMETRIZE_TIMESTEP
        if((sqr(y[0])+sqr(y[1])+sqr(y[2])) < sqr(3*config.nbpotEps)) 
#endif
        {
            // iteratively estimate timestep according to (Hut, Makino & McMillan 1995)
            for(int d=0; d<N_DIM; d++)  // first step for position
                y[d] = yprev[d] + yprev[d+N_DIM]*h + aprev[d]*h*h/2;
            fcn(n, x+h, y, f);
            nfcn++;
            for(int d=0; d<N_DIM; d++)
                y[d+N_DIM] = yprev[d+N_DIM] + (aprev[d]+f[d+N_DIM])*h/2;
            vel = sqrt(sqr(y[N_DIM])+sqr(y[N_DIM+1])+sqr(y[N_DIM+2]));
            acc = sqrt(sqr(f[N_DIM])+sqr(f[N_DIM+1])+sqr(f[N_DIM+2]));
            ts1 = vel>0 ? f[N_DIM*2]/vel : xend;
            ts2 = acc>0 ? sqrt(f[N_DIM*2]/acc) : xend;
            double h1 = std::min<double> (ts1, ts2) * config.accTreecode;
            h = (h+h1)/2;
        }
        if(x+h>xend) h=xend-x;
        x+=h;
        for(int d=0; d<N_DIM; d++)  // first step for position
            y[d] = yprev[d] + yprev[d+N_DIM]*h + aprev[d]*h*h/2;
        fcn(n, x, y, f);
        nfcn++;
        for(int d=0; d<N_DIM; d++)
        {
            anew[d] = f[d+N_DIM];
            y[d+N_DIM] = yprev[d+N_DIM] + (aprev[d]+anew[d])*h/2;
        }
        hprev=h;
        hout=h;
        solout (naccpt+1, xold, x, y, n);
        naccpt++; nstep++;
    }
    xout=x;
    return 1;
}
double COrbit::contlf(unsigned int d, double x)
{
    double dh=x-xold;
    if(d<N_DIM)  // interpolate position
        return yprev[d] + yprev[d+N_DIM]*dh + aprev[d]*dh*dh/2/* + (anew[d]-aprev[d])/hout*dh*dh*dh/6*/;
    else   // interpolate velocity
        return yprev[d] + aprev[d-N_DIM]*dh + (anew[d-N_DIM]-aprev[d-N_DIM])/hout*dh*dh/2;
}
    
///////////////////////////////////////////////////////
////  below follow dop853.c encapsulated routines  ////
static double sign (double a, double b)
{
  return (b < 0.0)? -fabs(a) : fabs(a);

} /* sign */


static double min_d (double a, double b)
{
  return (a < b)?a:b;

} /* min_d */


static double max_d (double a, double b)
{
  return (a > b)?a:b;

} /* max_d */


double COrbit::hinit (unsigned n, /*FcnEqDiff fcn,*/ double x, double* y,
              double posneg, double* f0, double* f1, double* yy1, int iord,
              double hmax, double* atoler, double* rtoler, int itoler)
{
  double   dnf, dny, atoli, rtoli, sk, h, h1, der2, der12, sqre;
  unsigned i;

  dnf = 0.0;
  dny = 0.0;
  atoli = atoler[0];
  rtoli = rtoler[0];

  if (!itoler)
    for (i = 0; i < n; i++)
    {
      sk = atoli + rtoli * fabs(y[i]);
      sqre = f0[i] / sk;
      dnf += sqre*sqre;
      sqre = y[i] / sk;
      dny += sqre*sqre;
    }
  else
    for (i = 0; i < n; i++)
    {
      sk = atoler[i] + rtoler[i] * fabs(y[i]);
      sqre = f0[i] / sk;
      dnf += sqre*sqre;
      sqre = y[i] / sk;
      dny += sqre*sqre;
    }

  if ((dnf <= 1.0E-10) || (dny <= 1.0E-10))
    h = 1.0E-6;
  else
    h = sqrt (dny/dnf) * 0.01;

  h = min_d (h, hmax);
  h = sign (h, posneg);

  /* perform an explicit Euler step */
  for (i = 0; i < n; i++)
    yy1[i] = y[i] + h * f0[i];
  fcn (n, x+h, yy1, f1);

  /* estimate the second derivative of the solution */
  der2 = 0.0;
  if (!itoler)
    for (i = 0; i < n; i++)
    {
      sk = atoli + rtoli * fabs(y[i]);
      sqre = (f1[i] - f0[i]) / sk;
      der2 += sqre*sqre;
    }
  else
    for (i = 0; i < n; i++)
    {
      sk = atoler[i] + rtoler[i] * fabs(y[i]);
      sqre = (f1[i] - f0[i]) / sk;
      der2 += sqre*sqre;
    }
  der2 = sqrt (der2) / h;

  /* step size is computed such that h**iord * max_d(norm(f0),norm(der2)) = 0.01 */
  der12 = max_d (fabs(der2), sqrt(dnf));
  if (der12 <= 1.0E-15)
    h1 = max_d (1.0E-6, fabs(h)*1.0E-3);
  else
    h1 = pow (0.01/der12, 1.0/(double)iord);
  h = min_d (100.0 * h, min_d (h1, hmax));

  return sign (h, posneg);

} /* hinit */


/* core integrator */
int COrbit::dopcor (unsigned n, /*FcnEqDiff fcn,*/ double x, double* y, double xend,
                   double hmax, double h, double* rtoler, double* atoler,
                   int itoler, /*FILE* fileout, SolTrait solout,*/ int iout,
                   long nmax, double uround, int meth, long nstiff, double safe,
                   double beta, double fac1, double fac2, unsigned* icont)
{
  double   facold, expo1, fac, facc1, facc2, fac11, posneg, xph;
  double   atoli, rtoli, hlamb, err, sk, hnew, yd0, ydiff, bspl;
  double   stnum, stden, sqre, err2, erri, deno;
  int      iasti, iord, irtrn, reject, last, nonsti;
  unsigned i, j;
  double   c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c14, c15, c16;
  double   b1, b6, b7, b8, b9, b10, b11, b12, bhh1, bhh2, bhh3;
  double   er1, er6, er7, er8, er9, er10, er11, er12;
  double   a21, a31, a32, a41, a43, a51, a53, a54, a61, a64, a65, a71, a74, a75, a76;
  double   a81, a84, a85, a86, a87, a91, a94, a95, a96, a97, a98;
  double   a101, a104, a105, a106, a107, a108, a109;
  double   a111, a114, a115, a116, a117, a118, a119, a1110;
  double   a121, a124, a125, a126, a127, a128, a129, a1210, a1211;
  double   a141, a147, a148, a149, a1410, a1411, a1412, a1413;
  double   a151, a156, a157, a158, a1511, a1512, a1513, a1514;
  double   a161, a166, a167, a168, a169, a1613, a1614, a1615;
  double   d41, d46, d47, d48, d49, d410, d411, d412, d413, d414, d415, d416;
  double   d51, d56, d57, d58, d59, d510, d511, d512, d513, d514, d515, d516;
  double   d61, d66, d67, d68, d69, d610, d611, d612, d613, d614, d615, d616;
  double   d71, d76, d77, d78, d79, d710, d711, d712, d713, d714, d715, d716;

  /* initialisations */
  switch (meth)
  {
    case 1:

      c2  = 0.526001519587677318785587544488E-01;
      c3  = 0.789002279381515978178381316732E-01;
      c4  = 0.118350341907227396726757197510E+00;
      c5  = 0.281649658092772603273242802490E+00;
      c6  = 0.333333333333333333333333333333E+00;
      c7  = 0.25E+00;
      c8  = 0.307692307692307692307692307692E+00;
      c9  = 0.651282051282051282051282051282E+00;
      c10 = 0.6E+00;
      c11 = 0.857142857142857142857142857142E+00;
      c14 = 0.1E+00;
      c15 = 0.2E+00;
      c16 = 0.777777777777777777777777777778E+00;

      b1 =   5.42937341165687622380535766363E-2;
      b6 =   4.45031289275240888144113950566E0;
      b7 =   1.89151789931450038304281599044E0;
      b8 =  -5.8012039600105847814672114227E0;
      b9 =   3.1116436695781989440891606237E-1;
      b10 = -1.52160949662516078556178806805E-1;
      b11 =  2.01365400804030348374776537501E-1;
      b12 =  4.47106157277725905176885569043E-2;

      bhh1 = 0.244094488188976377952755905512E+00;
      bhh2 = 0.733846688281611857341361741547E+00;
      bhh3 = 0.220588235294117647058823529412E-01;

      er1  =  0.1312004499419488073250102996E-01;
      er6  = -0.1225156446376204440720569753E+01;
      er7  = -0.4957589496572501915214079952E+00;
      er8  =  0.1664377182454986536961530415E+01;
      er9  = -0.3503288487499736816886487290E+00;
      er10 =  0.3341791187130174790297318841E+00;
      er11 =  0.8192320648511571246570742613E-01;
      er12 = -0.2235530786388629525884427845E-01;

      a21 =    5.26001519587677318785587544488E-2;
      a31 =    1.97250569845378994544595329183E-2;
      a32 =    5.91751709536136983633785987549E-2;
      a41 =    2.95875854768068491816892993775E-2;
      a43 =    8.87627564304205475450678981324E-2;
      a51 =    2.41365134159266685502369798665E-1;
      a53 =   -8.84549479328286085344864962717E-1;
      a54 =    9.24834003261792003115737966543E-1;
      a61 =    3.7037037037037037037037037037E-2;
      a64 =    1.70828608729473871279604482173E-1;
      a65 =    1.25467687566822425016691814123E-1;
      a71 =    3.7109375E-2;
      a74 =    1.70252211019544039314978060272E-1;
      a75 =    6.02165389804559606850219397283E-2;
      a76 =   -1.7578125E-2;

      a81 =    3.70920001185047927108779319836E-2;
      a84 =    1.70383925712239993810214054705E-1;
      a85 =    1.07262030446373284651809199168E-1;
      a86 =   -1.53194377486244017527936158236E-2;
      a87 =    8.27378916381402288758473766002E-3;
      a91 =    6.24110958716075717114429577812E-1;
      a94 =   -3.36089262944694129406857109825E0;
      a95 =   -8.68219346841726006818189891453E-1;
      a96 =    2.75920996994467083049415600797E1;
      a97 =    2.01540675504778934086186788979E1;
      a98 =   -4.34898841810699588477366255144E1;
      a101 =   4.77662536438264365890433908527E-1;
      a104 =  -2.48811461997166764192642586468E0;
      a105 =  -5.90290826836842996371446475743E-1;
      a106 =   2.12300514481811942347288949897E1;
      a107 =   1.52792336328824235832596922938E1;
      a108 =  -3.32882109689848629194453265587E1;
      a109 =  -2.03312017085086261358222928593E-2;

      a111 =  -9.3714243008598732571704021658E-1;
      a114 =   5.18637242884406370830023853209E0;
      a115 =   1.09143734899672957818500254654E0;
      a116 =  -8.14978701074692612513997267357E0;
      a117 =  -1.85200656599969598641566180701E1;
      a118 =   2.27394870993505042818970056734E1;
      a119 =   2.49360555267965238987089396762E0;
      a1110 = -3.0467644718982195003823669022E0;
      a121 =   2.27331014751653820792359768449E0;
      a124 =  -1.05344954667372501984066689879E1;
      a125 =  -2.00087205822486249909675718444E0;
      a126 =  -1.79589318631187989172765950534E1;
      a127 =   2.79488845294199600508499808837E1;
      a128 =  -2.85899827713502369474065508674E0;
      a129 =  -8.87285693353062954433549289258E0;
      a1210 =  1.23605671757943030647266201528E1;
      a1211 =  6.43392746015763530355970484046E-1;

      a141 =  5.61675022830479523392909219681E-2;
      a147 =  2.53500210216624811088794765333E-1;
      a148 = -2.46239037470802489917441475441E-1;
      a149 = -1.24191423263816360469010140626E-1;
      a1410 =  1.5329179827876569731206322685E-1;
      a1411 =  8.20105229563468988491666602057E-3;
      a1412 =  7.56789766054569976138603589584E-3;
      a1413 = -8.298E-3;

      a151 =  3.18346481635021405060768473261E-2;
      a156 =  2.83009096723667755288322961402E-2;
      a157 =  5.35419883074385676223797384372E-2;
      a158 = -5.49237485713909884646569340306E-2;
      a1511 = -1.08347328697249322858509316994E-4;
      a1512 =  3.82571090835658412954920192323E-4;
      a1513 = -3.40465008687404560802977114492E-4;
      a1514 =  1.41312443674632500278074618366E-1;
      a161 = -4.28896301583791923408573538692E-1;
      a166 = -4.69762141536116384314449447206E0;
      a167 =  7.68342119606259904184240953878E0;
      a168 =  4.06898981839711007970213554331E0;
      a169 =  3.56727187455281109270669543021E-1;
      a1613 = -1.39902416515901462129418009734E-3;
      a1614 =  2.9475147891527723389556272149E0;
      a1615 = -9.15095847217987001081870187138E0;

      d41  = -0.84289382761090128651353491142E+01;
      d46  =  0.56671495351937776962531783590E+00;
      d47  = -0.30689499459498916912797304727E+01;
      d48  =  0.23846676565120698287728149680E+01;
      d49  =  0.21170345824450282767155149946E+01;
      d410 = -0.87139158377797299206789907490E+00;
      d411 =  0.22404374302607882758541771650E+01;
      d412 =  0.63157877876946881815570249290E+00;
      d413 = -0.88990336451333310820698117400E-01;
      d414 =  0.18148505520854727256656404962E+02;
      d415 = -0.91946323924783554000451984436E+01;
      d416 = -0.44360363875948939664310572000E+01;

      d51  =  0.10427508642579134603413151009E+02;
      d56  =  0.24228349177525818288430175319E+03;
      d57  =  0.16520045171727028198505394887E+03;
      d58  = -0.37454675472269020279518312152E+03;
      d59  = -0.22113666853125306036270938578E+02;
      d510 =  0.77334326684722638389603898808E+01;
      d511 = -0.30674084731089398182061213626E+02;
      d512 = -0.93321305264302278729567221706E+01;
      d513 =  0.15697238121770843886131091075E+02;
      d514 = -0.31139403219565177677282850411E+02;
      d515 = -0.93529243588444783865713862664E+01;
      d516 =  0.35816841486394083752465898540E+02;

      d61 =  0.19985053242002433820987653617E+02;
      d66 = -0.38703730874935176555105901742E+03;
      d67 = -0.18917813819516756882830838328E+03;
      d68 =  0.52780815920542364900561016686E+03;
      d69 = -0.11573902539959630126141871134E+02;
      d610 =  0.68812326946963000169666922661E+01;
      d611 = -0.10006050966910838403183860980E+01;
      d612 =  0.77771377980534432092869265740E+00;
      d613 = -0.27782057523535084065932004339E+01;
      d614 = -0.60196695231264120758267380846E+02;
      d615 =  0.84320405506677161018159903784E+02;
      d616 =  0.11992291136182789328035130030E+02;

      d71  = -0.25693933462703749003312586129E+02;
      d76  = -0.15418974869023643374053993627E+03;
      d77  = -0.23152937917604549567536039109E+03;
      d78  =  0.35763911791061412378285349910E+03;
      d79  =  0.93405324183624310003907691704E+02;
      d710 = -0.37458323136451633156875139351E+02;
      d711 =  0.10409964950896230045147246184E+03;
      d712 =  0.29840293426660503123344363579E+02;
      d713 = -0.43533456590011143754432175058E+02;
      d714 =  0.96324553959188282948394950600E+02;
      d715 = -0.39177261675615439165231486172E+02;
      d716 = -0.14972683625798562581422125276E+03;

      break;
  }

  facold = 1.0E-4;
  expo1 = 1.0/8.0 - beta * 0.2;
  facc1 = 1.0 / fac1;
  facc2 = 1.0 / fac2;
  posneg = sign (1.0, xend-x);

  /* initial preparations */
  atoli = atoler[0];
  rtoli = rtoler[0];
  last  = 0;
  hlamb = 0.0;
  iasti = 0;
  fcn (n, x, y, k1);
  hmax = fabs (hmax);
  iord = 8;
  if (h == 0.0)
    h = hinit (n, /*fcn,*/ x, y, posneg, k1, k2, k3, iord, hmax, atoler, rtoler, itoler);
  nfcn += 2;
  reject = 0;
  xold = x;
  
  if (iout)
  {
    irtrn = 1;
    hout = 1.0;
    xout = x;
    if(x>xold) solout (naccpt+1, xold, x, y, n); 
  }

  /* basic integration step */
  while (1)
  {
    if (nstep > nmax)
    {
      /*if (fileout)
        fprintf (fileout, "Exit of dop853 at x = %.16e, more than nmax = %li are needed\r\n", x, nmax);*/
      xout = x;
      hout = h;
      return -2;
    }

    if (!gsl_finite(h) || 0.1 * fabs(h) <= fabs(x) * uround)
    {
      /*if (fileout)
        fprintf (fileout, "Exit of dop853 at x = %.16e, step size too small h = %.16e\r\n", x, h);*/
      xout = x;
      hout = h;
      return -3;
    }

    if ((x + 1.01*h - xend) * posneg > 0.0)
    {
      h = xend - x;
      last = 1;
    }

    nstep++;

    /* the twelve stages */
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * a21 * k1[i];
    fcn (n, x+c2*h, yy1, k2);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a31*k1[i] + a32*k2[i]);
    fcn (n, x+c3*h, yy1, k3);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a41*k1[i] + a43*k3[i]);
    fcn (n, x+c4*h, yy1, k4);
    for (i = 0; i <n; i++)
      yy1[i] = y[i] + h * (a51*k1[i] + a53*k3[i] + a54*k4[i]);
    fcn (n, x+c5*h, yy1, k5);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a61*k1[i] + a64*k4[i] + a65*k5[i]);
    fcn (n, x+c6*h, yy1, k6);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a71*k1[i] + a74*k4[i] + a75*k5[i] + a76*k6[i]);
    fcn (n, x+c7*h, yy1, k7);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a81*k1[i] + a84*k4[i] + a85*k5[i] + a86*k6[i] +
                          a87*k7[i]);
    fcn (n, x+c8*h, yy1, k8);
    for (i = 0; i <n; i++)
      yy1[i] = y[i] + h * (a91*k1[i] + a94*k4[i] + a95*k5[i] + a96*k6[i] +
                          a97*k7[i] + a98*k8[i]);
    fcn (n, x+c9*h, yy1, k9);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a101*k1[i] + a104*k4[i] + a105*k5[i] + a106*k6[i] +
                          a107*k7[i] + a108*k8[i] + a109*k9[i]);
    fcn (n, x+c10*h, yy1, k10);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a111*k1[i] + a114*k4[i] + a115*k5[i] + a116*k6[i] +
                          a117*k7[i] + a118*k8[i] + a119*k9[i] + a1110*k10[i]);
    fcn (n, x+c11*h, yy1, k2);
    xph = x + h;
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a121*k1[i] + a124*k4[i] + a125*k5[i] + a126*k6[i] +
                          a127*k7[i] + a128*k8[i] + a129*k9[i] +
                          a1210*k10[i] + a1211*k2[i]);
    fcn (n, xph, yy1, k3);
    nfcn += 11;
    for (i = 0; i < n; i++)
    {
      k4[i] = b1*k1[i] + b6*k6[i] + b7*k7[i] + b8*k8[i] + b9*k9[i] +
              b10*k10[i] + b11*k2[i] + b12*k3[i];
      k5[i] = y[i] + h * k4[i];
    }
     
    /* error estimation */
    err = 0.0;
    err2 = 0.0;
    if (!itoler)
    {
      /*   !!!! alteration of original dop853 code !!!!   */
      rtoli=rtoler[0]; atoli=atoler[0];
      if(potential->Mbh!=0)
      {
          double r=sqrt(sqr(y[0])+sqr(y[1])+(N_dim==3?sqr(y[2]):0.0));
          if(r < potential->Mbh*BH_RADIUS_INCREASE_ACCURACY)   // take measures to increase accuracy
          {
              rtoli=rtoler[0]*(r/(potential->Mbh*BH_RADIUS_INCREASE_ACCURACY));
              atoli=atoler[0]*(r/(potential->Mbh*BH_RADIUS_INCREASE_ACCURACY));
          }
      }
      for (i = 0; i < n; i++)
      {
        sk = atoli + rtoli * max_d (fabs(y[i]), fabs(k5[i]));
        erri = k4[i] - bhh1*k1[i] - bhh2*k9[i] - bhh3*k3[i];
        sqre = erri / sk;
        err2 += sqre*sqre;
        erri = er1*k1[i] + er6*k6[i] + er7*k7[i] + er8*k8[i] + er9*k9[i] +
               er10 * k10[i] + er11*k2[i] + er12*k3[i];
        sqre = erri / sk;
        err += sqre*sqre;
      }
    }
    else
      for (i = 0; i < n; i++)
      {
        sk = atoler[i] + rtoler[i] * max_d (fabs(y[i]), fabs(k5[i]));
        erri = k4[i] - bhh1*k1[i] - bhh2*k9[i] - bhh3*k3[i];
        sqre = erri / sk;
        err2 += sqre*sqre;
        erri = er1*k1[i] + er6*k6[i] + er7*k7[i] + er8*k8[i] + er9*k9[i] +
               er10 * k10[i] + er11*k2[i] + er12*k3[i];
        sqre = erri / sk;
        err += sqre*sqre;
      }
    deno = err + 0.01 * err2;
    if (deno <= 0.0)
      deno = 1.0;
    err = fabs(h) * err * sqrt (1.0 / (deno*(double)n));

    /* computation of hnew */
    fac11 = pow (err, expo1);
    /* Lund-stabilization */
    fac = fac11 / pow(facold,beta);
    /* we require fac1 <= hnew/h <= fac2 */
    fac = max_d (facc2, min_d (facc1, fac/safe));
    hnew = h / fac;

    if (err <= 1.0)
    {
      /* step accepted */

      facold = max_d (err, 1.0E-4);
      naccpt++;
      fcn (n, xph, k5, k4);
      nfcn++;
      
      /* stiffness detection */
      if (!(naccpt % nstiff) || (iasti > 0))
      {
        stnum = 0.0;
        stden = 0.0;
        for (i = 0; i < n; i++)
        {
          sqre = k4[i] - k3[i];
          stnum += sqre*sqre;
          sqre = k5[i] - yy1[i];
          stden += sqre*sqre;
        }
        if (stden > 0.0)
          hlamb = h * sqrt (stnum / stden);
        if (hlamb > 6.1)
        {
          nonsti = 0;
          iasti++;
          if (iasti == 15)
            /*if (fileout)
              fprintf (fileout, "The problem seems to become stiff at x = %.16e\r\n", x);
            else*/
            {
              xout = x;
              hout = h;
              return -4;
            }
        }
        else
        {
          nonsti++;
          if (nonsti == 6)
            iasti = 0;
        }
      }
       
      /* final preparation for dense output */
      if (iout == 2)
      {
        /* save the first function evaluations */
        if (nrds == n)
          for (i = 0; i < n; i++)
          {
            rcont1[i] = y[i];
            ydiff = k5[i] - y[i];
            rcont2[i] = ydiff;
            bspl = h * k1[i] - ydiff;
            rcont3[i] = bspl;
            rcont4[i] = ydiff - h*k4[i] - bspl;
            rcont5[i] = d41*k1[i] + d46*k6[i] + d47*k7[i] + d48*k8[i] +
                        d49*k9[i] + d410*k10[i] + d411*k2[i] + d412*k3[i];
            rcont6[i] = d51*k1[i] + d56*k6[i] + d57*k7[i] + d58*k8[i] +
                        d59*k9[i] + d510*k10[i] + d511*k2[i] + d512*k3[i];
            rcont7[i] = d61*k1[i] + d66*k6[i] + d67*k7[i] + d68*k8[i] +
                        d69*k9[i] + d610*k10[i] + d611*k2[i] + d612*k3[i];
            rcont8[i] = d71*k1[i] + d76*k6[i] + d77*k7[i] + d78*k8[i] +
                        d79*k9[i] + d710*k10[i] + d711*k2[i] + d712*k3[i];
          }
        else
          for (j = 0; j < nrds; j++)
          {
            i = icont[j];
            rcont1[j] = y[i];
            ydiff = k5[i] - y[i];
            rcont2[j] = ydiff;
            bspl = h * k1[i] - ydiff;
            rcont3[j] = bspl;
            rcont4[j] = ydiff - h*k4[i] - bspl;
            rcont5[j] = d41*k1[i] + d46*k6[i] + d47*k7[i] + d48*k8[i] +
                        d49*k9[i] + d410*k10[i] + d411*k2[i] + d412*k3[i];
            rcont6[j] = d51*k1[i] + d56*k6[i] + d57*k7[i] + d58*k8[i] +
                        d59*k9[i] + d510*k10[i] + d511*k2[i] + d512*k3[i];
            rcont7[j] = d61*k1[i] + d66*k6[i] + d67*k7[i] + d68*k8[i] +
                        d69*k9[i] + d610*k10[i] + d611*k2[i] + d612*k3[i];
            rcont8[j] = d71*k1[i] + d76*k6[i] + d77*k7[i] + d78*k8[i] +
                        d79*k9[i] + d710*k10[i] + d711*k2[i] + d712*k3[i];
          }

        /* the next three function evaluations */
        for (i = 0; i < n; i++)
          yy1[i] = y[i] + h * (a141*k1[i] + a147*k7[i] + a148*k8[i] +
                              a149*k9[i] + a1410*k10[i] + a1411*k2[i] +
                              a1412*k3[i] + a1413*k4[i]);
        fcn (n, x+c14*h, yy1, k10);
        for (i = 0; i < n; i++)
          yy1[i] = y[i] + h * (a151*k1[i] + a156*k6[i] + a157*k7[i] + a158*k8[i] +
                              a1511*k2[i] + a1512*k3[i] + a1513*k4[i] +
                              a1514*k10[i]);
        fcn (n, x+c15*h, yy1, k2);
        for (i = 0; i < n; i++)
          yy1[i] = y[i] + h * (a161*k1[i] + a166*k6[i] + a167*k7[i] + a168*k8[i] +
                              a169*k9[i] + a1613*k4[i] + a1614*k10[i] +
                              a1615*k2[i]);
        fcn (n, x+c16*h, yy1, k3);
        nfcn += 3;

        /* final preparation */
        if (nrds == n)
          for (i = 0; i < n; i++)
          {
            rcont5[i] = h * (rcont5[i] + d413*k4[i] + d414*k10[i] +
                             d415*k2[i] + d416*k3[i]);
            rcont6[i] = h * (rcont6[i] + d513*k4[i] + d514*k10[i] +
                             d515*k2[i] + d516*k3[i]);
            rcont7[i] = h * (rcont7[i] + d613*k4[i] + d614*k10[i] +
                             d615*k2[i] + d616*k3[i]);
            rcont8[i] = h * (rcont8[i] + d713*k4[i] + d714*k10[i] +
                             d715*k2[i] + d716*k3[i]);
          }
        else
          for (j = 0; j < nrds; j++)
          {
            i = icont[j];
            rcont5[j] = h * (rcont5[j] + d413*k4[i] + d414*k10[i] +
                             d415*k2[i] + d416*k3[i]);
            rcont6[j] = h * (rcont6[j] + d513*k4[i] + d514*k10[i] +
                             d515*k2[i] + d516*k3[i]);
            rcont7[j] = h * (rcont7[j] + d613*k4[i] + d614*k10[i] +
                             d615*k2[i] + d616*k3[i]);
            rcont8[j] = h * (rcont8[j] + d713*k4[i] + d714*k10[i] +
                             d715*k2[i] + d716*k3[i]);
          }
      }

      memcpy (k1, k4, n * sizeof(double)); 
      memcpy (y, k5, n * sizeof(double));
      xold = x;
      x = xph;

      if (iout)
      {
        hout = h;
        xout = x;
        solout (naccpt+1, xold, x, y, n);
      }

      /* normal exit */
      if (last)
      {
        hout=hnew;
        xout = x;
        return 1;
      }

      if (fabs(hnew) > hmax)
        hnew = posneg * hmax;
      if (reject)
        hnew = posneg * min_d (fabs(hnew), fabs(h));

      reject = 0;

      /*   !!!! alteration of original dop853 code !!!!   */
      if(calcLyapunov)
      {
          // check if deviation vector requires renormalization
          double wnorm=0;
#ifdef LYAPUNOVVAREQ
          for(int indx=N_dim*2; indx<N_dim*4; indx++) wnorm+=sqr(y[indx]);
          wnorm=sqrt(wnorm);
          if(wnorm>LYAP_VAREQ_DEV_MAX)
          {
            lnwadd+=log(wnorm);
            for(int indx=N_dim*2; indx<N_dim*4; indx++)
            {
                y[indx]/=wnorm;
                k1[indx]/=wnorm;
                k2[indx]/=wnorm;
                k3[indx]/=wnorm;
                k4[indx]/=wnorm;
                k5[indx]/=wnorm;
                k6[indx]/=wnorm;
                k7[indx]/=wnorm;
                k8[indx]/=wnorm;
                k9[indx]/=wnorm;
                k10[indx]/=wnorm;
            }
          }
#else
          for(int indx=N_dim*2; indx<N_dim*4; indx++)
              wnorm+=gsl_pow_2(y[indx-N_dim*2]-y[indx]); 
          wnorm=sqrt(wnorm)/LYAP_DEV_INIT;
          if(wnorm>LYAP_DEV_MAX/LYAP_DEV_INIT)
          {
            needrenorm=true;   // it will be performed later, when allowed
          }
          if(needrenorm && allowrenorm)
          {
            // ensure that we do not decrease normalization
            if(wnorm<2) wnorm=2;
            lnwadd+=log(wnorm);
            for(int indx=N_dim*2; indx<N_dim*4; indx++)
            {
                y[indx] = y[indx-N_dim*2] + (y[indx]-y[indx-N_dim*2])/wnorm;
                k1[indx] = k1[indx-N_dim*2] + (k1[indx]-k1[indx-N_dim*2])/wnorm;
                k2[indx] = k2[indx-N_dim*2] + (k2[indx]-k2[indx-N_dim*2])/wnorm;
                k3[indx] = k3[indx-N_dim*2] + (k3[indx]-k3[indx-N_dim*2])/wnorm;
                k4[indx] = k4[indx-N_dim*2] + (k4[indx]-k4[indx-N_dim*2])/wnorm;
                k5[indx] = k5[indx-N_dim*2] + (k5[indx]-k5[indx-N_dim*2])/wnorm;
                k6[indx] = k6[indx-N_dim*2] + (k6[indx]-k6[indx-N_dim*2])/wnorm;
                k7[indx] = k7[indx-N_dim*2] + (k7[indx]-k7[indx-N_dim*2])/wnorm;
                k8[indx] = k8[indx-N_dim*2] + (k8[indx]-k8[indx-N_dim*2])/wnorm;
                k9[indx] = k9[indx-N_dim*2] + (k9[indx]-k9[indx-N_dim*2])/wnorm;
                k10[indx] = k10[indx-N_dim*2] + (k10[indx]-k10[indx-N_dim*2])/wnorm;
            }
            needrenorm = allowrenorm = false;
          }
#endif
      }
    }
    else
    {
      /* step rejected */
      hnew = h / min_d (facc1, fac11/safe);
      reject = 1;
      if (naccpt >= 1)
        nrejct=nrejct + 1;
      last = 0;
    }

    h = hnew;
  }

} /* dopcor */


/* front-end */
int COrbit::dop853
 (unsigned n, /*FcnEqDiff fcn,*/ double x, double* y, double xend, double* rtoler,
  double* atoler, int itoler, /*SolTrait solout,*/ int iout, /*FILE* fileout,*/ double uround,
  double safe, double fac1, double fac2, double beta, double hmax, double h,
  long nmax, int meth, long nstiff, unsigned nrdens, unsigned* icont, unsigned licont)
{
  if(leapfrog)
    return intlf(n, x, y, xend);
  int       arret, idid;
  unsigned  i;

  /* initialisations */
  nfcn = nstep = naccpt = nrejct = arret = 0;
  rcont1 = rcont2 = rcont3 = rcont4 = rcont5 = rcont6 = rcont7 = rcont8 = NULL;
  indir = NULL;

  /* n, the dimension of the system */
  if (n == UINT_MAX)
  {
    /*if (fileout)
      fprintf (fileout, "System too big, max. n = %u\r\n", UINT_MAX-1);*/
    arret = 1;
  }

  /* nmax, the maximal number of steps */
  if (!nmax)
    nmax = 100000;
  else if (nmax <= 0)
  {
    /*if (fileout)
      fprintf (fileout, "Wrong input, nmax = %li\r\n", nmax);*/
    arret = 1;
  }

  /* meth, coefficients of the method */
  if (!meth)
    meth = 1;
  else if ((meth <= 0) || (meth >= 2))
  {
    /*if (fileout)
      fprintf (fileout, "Curious input, meth = %i\r\n", meth);*/
    arret = 1;
  }

  /* nstiff, parameter for stiffness detection */
  if (!nstiff)
    nstiff = 1000;
  else if (nstiff < 0)
    nstiff = nmax + 10;

  /* iout, switch for calling solout */
  if ((iout < 0) || (iout > 2))
  {
    /*if (fileout)
      fprintf (fileout, "Wrong input, iout = %i\r\n", iout);*/
    arret = 1;
  }

  /* nrdens, number of dense output components */
  if (nrdens > n)
  {
    /*if (fileout)
      fprintf (fileout, "Curious input, nrdens = %u\r\n", nrdens);*/
    arret = 1;
  }
  else if (nrdens)
  {
    /* is there enough memory to allocate rcont12345678&indir ? */
    rcont1 = (double*) malloc (nrdens*sizeof(double));
    rcont2 = (double*) malloc (nrdens*sizeof(double));
    rcont3 = (double*) malloc (nrdens*sizeof(double));
    rcont4 = (double*) malloc (nrdens*sizeof(double));
    rcont5 = (double*) malloc (nrdens*sizeof(double));
    rcont6 = (double*) malloc (nrdens*sizeof(double));
    rcont7 = (double*) malloc (nrdens*sizeof(double));
    rcont8 = (double*) malloc (nrdens*sizeof(double));
    if (nrdens < n)
      indir = (unsigned*) malloc (n*sizeof(unsigned));

    if (!rcont1 || !rcont2 || !rcont3 || !rcont4 || !rcont5 ||
        !rcont6 || !rcont7 || !rcont8 || (!indir && (nrdens < n)))
    {
      /*if (fileout)
        fprintf (fileout, "Not enough free memory for rcont12345678&indir\r\n");*/
      arret = 1;
    }

    /* control of length of icont */
    if (nrdens == n)
    {
      /*if (icont && fileout)
        fprintf (fileout, "Warning : when nrdens = n there is no need allocating memory for icont\r\n");*/
      nrds = n;
    }
    else if (licont < nrdens)
    {
      /*if (fileout)
        fprintf (fileout, "Insufficient storage for icont, min. licont = %u\r\n", nrdens);*/
      arret = 1;
    }
    else
    {
      /*if ((iout < 2) && fileout)
        fprintf (fileout, "Warning : put iout = 2 for dense output\r\n");*/
      nrds = nrdens;
      for (i = 0; i < n; i++)
        indir[i] = UINT_MAX;
      for (i = 0; i < nrdens; i++)
        indir[icont[i]] = i;
    }
  }

  /* uround, smallest number satisfying 1.0+uround > 1.0 */
  if (uround == 0.0)
    uround = 2.3E-16;
  else if ((uround <= 1.0E-35) || (uround >= 1.0))
  {
    /*if (fileout)
      fprintf (fileout, "Which machine do you have ? Your uround was : %.16e\r\n", uround);*/
    arret = 1;
  }

  /* safety factor */
  if (safe == 0.0)
    safe = 0.9;
  else if ((safe >= 1.0) || (safe <= 1.0E-4))
  {
    /*if (fileout)
      fprintf (fileout, "Curious input for safety factor, safe = %.16e\r\n", safe);*/
    arret = 1;
  }

  /* fac1, fac2, parameters for step size selection */
  if (fac1 == 0.0)
    fac1 = 0.333;
  if (fac2 == 0.0)
    fac2 = 6.0;

  /* beta for step control stabilization */
  if (beta == 0.0)
    beta = 0.0;
  else if (beta < 0.0)
    beta = 0.0;
  else if (beta > 0.2)
  {
    /*if (fileout)
      fprintf (fileout, "Curious input for beta : beta = %.16e\r\n", beta);*/
    arret = 1;
  }

  /* maximal step size */
  if (hmax == 0.0)
    hmax = xend - x;

  /* is there enough free memory for the method ? */
  yy1 = (double*) malloc (n*sizeof(double));
  k1 = (double*) malloc (n*sizeof(double));
  k2 = (double*) malloc (n*sizeof(double));
  k3 = (double*) malloc (n*sizeof(double));
  k4 = (double*) malloc (n*sizeof(double));
  k5 = (double*) malloc (n*sizeof(double));
  k6 = (double*) malloc (n*sizeof(double));
  k7 = (double*) malloc (n*sizeof(double));
  k8 = (double*) malloc (n*sizeof(double));
  k9 = (double*) malloc (n*sizeof(double));
  k10 = (double*) malloc (n*sizeof(double));

  if (!yy1 || !k1 || !k2 || !k3 || !k4 || !k5 || !k6 || !k7 || !k8 || !k9 || !k10)
  {
    /*if (fileout)
      fprintf (fileout, "Not enough free memory for the method\r\n");*/
    arret = 1;
  }

  /* when a failure has occured, we return -1 */
  if (arret)
  {
    if (k10)
      free (k10);
    if (k9)
      free (k9);
    if (k8)
      free (k8);
    if (k7)
      free (k7);
    if (k6)
      free (k6);
    if (k5)
      free (k5);
    if (k4)
      free (k4);
    if (k3)
      free (k3);
    if (k2)
      free (k2);
    if (k1)
      free (k1);
    if (yy1)
      free (yy1);
    if (indir)
      free (indir);
    if (rcont8)
      free (rcont8);
    if (rcont7)
      free (rcont7);
    if (rcont6)
      free (rcont6);
    if (rcont5)
      free (rcont5);
    if (rcont4)
      free (rcont4);
    if (rcont3)
      free (rcont3);
    if (rcont2)
      free (rcont2);
    if (rcont1)
      free (rcont1);

    return -1;
  }
  else
  {
    idid = dopcor (n, /*fcn,*/ x, y, xend, hmax, h, rtoler, atoler, itoler, /*fileout,
                   solout,*/ iout, nmax, uround, meth, nstiff, safe, beta, fac1, fac2, icont);
    free (k10);
    free (k9);
    free (k8);
    free (k7);
    free (k6);
    free (k5);    /* reverse order freeing too increase chances */
    free (k4);    /* of efficient dynamic memory managing       */
    free (k3);
    free (k2);
    free (k1);
    free (yy1);
    if (indir)
      free (indir);
    if (rcont8)
    {
      free (rcont8);
      free (rcont7);
      free (rcont6);
      free (rcont5);
      free (rcont4);
      free (rcont3);
      free (rcont2);
      free (rcont1);
    }

    return idid;
  }

} /* dop853 */


/* dense output function */
double COrbit::contd8 (unsigned ii, double x)
{
  if(leapfrog)
      return contlf(ii, x);
  unsigned i, j;
  double   s, s1;

  i = UINT_MAX;

  if (!indir)
    i = ii;
  else
    i = indir[ii];

  if (i == UINT_MAX)
  {
    printf ("No dense output available for %uth component", ii);
    return 0.0;
  }

  s = (x - xold) / hout;
  s1 = 1.0 - s;

  return rcont1[i]+s*(rcont2[i]+s1*(rcont3[i]+s*(rcont4[i]+s1*(rcont5[i]+
         s*(rcont6[i]+s1*(rcont7[i]+s*rcont8[i]))))));
} /* contd8 */

////--------------  COrbitDesc class - shortened orbit data -------------////
COrbitDesc::~COrbitDesc()
{
}

COrbitDesc::COrbitDesc():
potential(NULL), grid(NULL), timeStep(0), timeUnit(0), calcLyapunov(false)
{
    for(int i=0; i<N_DIM; i++)
        ip[i]=iv[i]=lf[i]=inertia[i]=0;
    time=maxtime=0;
    state=0;  
    Description="?";
    lfccdiff=lambda=0;
    orbitWeight=0;
    shellIndex=-1;
    orbitIndex=0;
    isChaotic=false;
    numSamplingPoints=config.sm_numSamplingPoints;  // default
    Energy=0; Ediff=0;
}

COrbitDesc::COrbitDesc(const COrbitDesc &src):
potential(src.potential), grid(src.grid), timeStep(src.timeStep), timeUnit(src.timeUnit), calcLyapunov(src.calcLyapunov)
{
    for(int i=0; i<N_DIM; i++)
    {
        ip[i]=src.ip[i];
        iv[i]=src.iv[i];
        lf[i]=src.lf[i];
        inertia[i]=src.inertia[i];
    }
    time=src.time;
    maxtime=src.maxtime;
    state=src.state;
    Description=src.Description;
    lfccdiff=src.lfccdiff;
    lambda=src.lambda;
    orbitWeight=src.orbitWeight;
    shellIndex=src.shellIndex;
    orbitIndex=src.orbitIndex;
    isChaotic=src.isChaotic;
	celltimes=src.celltimes;
    cellVr=src.cellVr;
    cellVt=src.cellVt;
    for(int d=0; d<N_DIM*2; d++) trajsample[d]=src.trajsample[d];
    numSamplingPoints=src.numSamplingPoints;
    Energy=src.Energy;
    Ediff=src.Ediff;
}
COrbitDesc& COrbitDesc::operator=(const COrbitDesc &src)
{
    if(&src!=this)
    {
        potential=src.potential;
        grid=src.grid;
        timeStep=src.timeStep;
        timeUnit=src.timeUnit;
        calcLyapunov=src.calcLyapunov;
        for(int i=0; i<N_DIM; i++)
        {
            ip[i]=src.ip[i];
            iv[i]=src.iv[i];
            lf[i]=src.lf[i];
            inertia[i]=src.inertia[i];
        }
        time=src.time;
        maxtime=src.maxtime;
        state=src.state;
        Description=src.Description;
        lfccdiff=src.lfccdiff;
        lambda=src.lambda;
        orbitWeight=src.orbitWeight;
        shellIndex=src.shellIndex;
        orbitIndex=src.orbitIndex;
        isChaotic=src.isChaotic;
        celltimes=src.celltimes;
        cellVr=src.cellVr;
        cellVt=src.cellVt;
        for(int d=0; d<N_DIM*2; d++) trajsample[d]=src.trajsample[d];
        numSamplingPoints=src.numSamplingPoints;
        Energy=src.Energy;
        Ediff=src.Ediff;
    }
    return *this;
}

//function that rounds double to NUM_DIGITS_ROUND significant digits
double xround(double x)
{   // sorry, this is the only compliant way to store the same value in text file...
    if(x==0) return 0;
    return QString::number(x,'e', NUM_DIGITS_ROUND).toDouble();
}

COrbitDesc::COrbitDesc(const CPotential* _potential, const CSpatialGrid* _grid, double _ip[N_DIM], double _iv[N_DIM], double _time, double _timeStep, double _timeUnit, bool _calcLyapunov, double _maxtime):
    potential(_potential), grid(_grid), 
    time(xround(_time)), timeStep(xround(_timeStep)), timeUnit(xround(_timeUnit)), calcLyapunov(_calcLyapunov)
{
    for(int i=0; i<N_DIM; i++)
    {
        ip[i]=xround(_ip[i]);
        iv[i]=xround(_iv[i]);
        lf[i]=0;
        inertia[i]=0;
    }
    maxtime=(_maxtime>time)?_maxtime:time;
    state=0;
    Description="?";
    lfccdiff=0;
    lambda=0;
    orbitWeight=0;
    shellIndex=-1;
    orbitIndex=0;
    isChaotic=false;
    numSamplingPoints=config.sm_numSamplingPoints;
    Energy = potential->Phi(ip[0], ip[1], ip[2]) + (sqr(iv[0])+sqr(iv[1])+sqr(iv[2]))/2;
    Ediff=0;
}

void COrbitDesc::run()
{
    if(timeStep==0)
        return;
    state=1;
    try{
    COrbit* orbit;
    if(grid!=NULL)
    {
        celltimes.assign(grid->cellsInModel+1, 0);
        cellVr.assign(grid->cellsInModel+1, 0);
        cellVt.assign(grid->cellsInModel+1, 0);
        for(int d=0; d<2*N_DIM; d++) 
        {
            trajsample[d].clear();
            trajsample[d].reserve(numSamplingPoints);
        }
    }
    int numOrbits=(grid==NULL? 1 : std::max<int>(config.sm_numBunchOrbits, 1));
    double initip[N_DIM], initiv[N_DIM];
    for(int numCurrentOrbit=0; numCurrentOrbit<numOrbits; numCurrentOrbit++)
    {
        if(numCurrentOrbit>0)
        {   // perturb IC
            const int perturbArraySize=8;
            double perturbation[perturbArraySize][3] = { {1,1,1}, {1,-1,-1}, {-1,-1,1}, {-1,1,-1}, {1,-1,1}, {-1,1,1}, {1,1,-1}, {-1,-1,-1} };
            int perturbIndex=(numCurrentOrbit-1)%perturbArraySize;
            double perturbFactor = BUNCH_PERTURB_FACTOR / (1 + (numCurrentOrbit-1)/perturbArraySize);  // 1 for first 8 orbits, then 1/2, then 1/3 etc
            for(int d=0; d<N_DIM; d++)
                initip[d]=ip[d]*(1 + perturbation[perturbIndex][d]*perturbFactor);
        }
        else
        for(int d=0; d<N_DIM; d++)
        {
            initip[d]=ip[d];
            initiv[d]=iv[d];
        }
        orbit = new COrbit(potential, grid, initip, initiv, timeStep, timeUnit, calcLyapunov);
        if(grid!=NULL && maxtime>time)
            orbit->integrateAdaptiveTime(time, maxtime, config.adaptiveTimeThreshold);
        else
            orbit->integrateToTime(time);
        orbit->finishIntegration();
        if(numCurrentOrbit==0 && orbit->nActual>0)
        {   // if there are several orbits in the bunch, store only the main one
            Spectrum sl[N_DIM];
            double inert[N_DIM];
            orbit->performFrequencyAnalysis(sl);
            orbit->calcInertiaOverInterval(0, orbit->nActual, inert);
            for(int i=0; i<potential->N_dim; i++)
            {
                lf[i] = sl[i][0];
                inertia[i] = inert[i];
            }
            if(potential->N_dim==2) lf[2]=0;
            lfccdiff = orbit->calcLFCCdiffusion();
            lambda = orbit->calcLambda();
            Description = orbit->Description;
            Ediff=orbit->Edif;
            Energy=orbit->Einit;
            time=orbit->intTime;  // may be longer than desired
        }
        if(grid!=NULL)
        {   
            for(int c=0; c<orbit->celltimes.size(); c++)
            {
                celltimes[c] += orbit->celltimes[c];     // convert from double to float
                cellVr[c] += orbit->cellVr[c];
                cellVt[c] += orbit->cellVt[c];
            }
            if(numSamplingPoints>0)
            {
                // sample trajectory only if run within context of Schwarzschild modelling
                int numCurrentSamplingPoints = (numSamplingPoints-1)/numOrbits+1;
                if(numCurrentSamplingPoints>orbit->nActual-1) numCurrentSamplingPoints=orbit->nActual-1;
                std::vector<int> samplingpoints;
                samplingpoints.reserve(numCurrentSamplingPoints);
                for(int i=0; i<numCurrentSamplingPoints; i++)
                {
                    int np;
                    bool exists;
                    do{
                        exists=false;
                        np=floor(rand()*1.0/RAND_MAX*(orbit->nActual-1));
                        for(int j=0; j<samplingpoints.size(); j++) 
                            if(samplingpoints[j]==np) exists=true;
                    }while(exists && (orbit->nActual-1>samplingpoints.size()));
                    samplingpoints.push_back(np);   // try to avoid duplication
                    for(int d=0; d<N_DIM; d++)
                    {
                        trajsample[d].push_back(orbit->tp[d][np]);
                        trajsample[d+N_DIM].push_back(orbit->tv[d][np]);
                    }
                }
            }
        }
        delete(orbit); 
    }
    if(grid!=NULL)
    {
        for(int c=0; c<grid->cellsInModel+1; c++)
        {
            celltimes[c] /= numOrbits;
            cellVr[c] /= numOrbits;
            cellVt[c] /= numOrbits;
        }
        numSamplingPoints=trajsample[0].size();
    }
    }
    catch(const std::bad_alloc&) {
        Description="N/A";
        // do nothing..
    }
    state=2;
}

////--------------  COrbitLibrary class - array of COrbitDesc -------------////

COrbitLibrary::COrbitLibrary(const CPotential* _potential, const CSpatialGrid* _grid): 
potential(_potential?_potential->clone():NULL), grid(_grid)
{
    nComplete=numChaotic=0;
}
COrbitLibrary::~COrbitLibrary() { delete potential; }

bool COrbitLibrary::findPeriodicOrbit(double E, int coord1, int coord2, double& cross1)
{
    if(coord1>coord2) { int tmp=coord1; coord1=coord2; coord1=tmp; }
    double Xlo=0, Xhi=potential->longaxisradius(E);
    double tol=Xhi*1e-8;
    double Torb=potential->longaxisperiod(E, Xhi);
    double IP[N_DIM], IV[N_DIM];
    IP[0]=IP[1]=IP[2]=0; IV[0]=IV[1]=IV[2]=0;
    cross1=0;
    int niter=0;
    do{
        double v2;
        do{
            IP[coord1]=(Xlo+Xhi)/2; 
            v2=2*(E-potential->Phi(IP[0],IP[1],IP[2]));
            if(v2<0) Xhi=IP[coord1];
            else IV[coord2]=sqrt(v2);
        } 
        while(v2<0);
        COrbit orb(potential, NULL, IP, IV, Torb/10, Torb, false, true, coord1, coord2);
        orb.integrateToTime(Torb*2);   // should perform more than one revolution
        if(orb.ps.size()<2)
            return false;
        if(orb.ps[1].first > IP[coord1])
        {   Xlo=IP[coord1]; }
        else
        {   Xhi=IP[coord1]; }
        niter++;
    }
    while(niter<25 && Xhi-Xlo>tol && Xhi>tol);
    cross1=(Xhi>tol) ? IP[coord1] : 0;
    return (Xhi>tol);
}

void COrbitLibrary::prepareInitialConditions(int NStationary, int NPrincipalPlane, int NYalpha, int NRandom, double E)
{
    double Xmax = potential->longaxisradius(E);
    if(Xmax<0) return;  // error...
    double Torbx= xround(potential->longaxisperiod(E, Xmax));
    double intTime = xround(config.intTimeInPeriods * Torbx);
    double maxTime = xround(config.intTimeMaxAdaptive * Torbx);
    double intTimeStep = xround(Torbx / config.intTimeStepsPerPeriod);
    // Prepare initial conditions
    int nfms=NStationary;            // number of stationary start-space elements
    int nfmp=NPrincipalPlane/potential->N_dim;  // number of principal-plane start-space elements
    double ip[N_DIM], iv[N_DIM];
    ip[2]=iv[2]=0;
    if(potential->N_dim==3)
    {
        // stationary
#ifdef SCHWARZSCHILD_START_SPACE
        int ngrd=sqrt(nfms/3.0);
#else
        int ngrd=sqrt(1.0*nfms*M_PI/2);
#endif
        for(int ith=0; ith<ngrd; ith++)
        {
#ifdef SCHWARZSCHILD_START_SPACE
            double theta = tan((ith+0.5)/ngrd * M_PI/4);
            for(int iph=0; iph<ngrd*3; iph++)
            {
                double phi = tan((iph/3+0.5)/ngrd * M_PI/4);
                double X, Y, Z;
                switch(iph%3)
                {
                case 0:
                    X=Xmax;
                    Y=Xmax*theta;
                    Z=Xmax*phi;
                    break;
                case 1:
                    Y=Xmax;
                    Z=Xmax*theta;
                    X=Xmax*phi;
                    break;
                case 2:
                    Z=Xmax;
                    X=Xmax*theta;
                    Y=Xmax*phi;
                    break;
                }
#else
            double theta = (1 - (ith+0.5)/ngrd) * M_PI/2;
            int nph = ngrd * sin(theta) + 1 ;
            for(int iph=0; iph<nph; iph++)
            {
                double phi = (iph+0.5)/nph * M_PI/2;
                double X, Y, Z;

                X=Xmax*sin(theta)*cos(phi);
                Y=Xmax*sin(theta)*sin(phi)/* *q*/;
                Z=Xmax*cos(theta)/* *p*/;
#endif
                double r=potential->findintersection(E, X, Y, Z);
                ip[0]=X*r; ip[1]=Y*r; ip[2]=Z*r;
                iv[0]=iv[1]=iv[2]=0;
                OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
                OrbitList.back().orbitIndex = OrbitList.size()-1;
            }
        }
        // principal-plane
        if(NPrincipalPlane>0)
        {
#ifdef ANNULUS_PP_START_SPACE
            double rmin=Xmax, rminc;
            findPeriodicOrbit(E, 0, 1, rminc);
            if(rminc<rmin) rmin=rminc;
            if(rmin>0) findPeriodicOrbit(E, 1, 2, rminc);
            if(rminc<rmin) rmin=rminc;
            if(rmin>0) findPeriodicOrbit(E, 0, 2, rminc);
            if(rminc<rmin) rmin=rminc;
            rminc=rmin/Xmax;
#else
            double rminc=0;
#endif
            // use elliptic annulus with inner radius rmin and outer xmax
            ngrd = sqrt(4*nfmp/M_PI/(1+rminc))+0.5;
            for(int ir=0; ir<ngrd; ir++)
            {
                double r = (ir+0.5)/ngrd * (1-rminc) + rminc;
                int nph = r * ngrd * M_PI/2 + 0.22;
                for(int iph=0; iph<nph; iph++)
                {
                    double phi = (iph+0.5)/nph * M_PI/2;
                    double rmax=potential->findintersection(E, Xmax*cos(phi), Xmax*sin(phi), 0);
                    ip[0]=r*rmax*Xmax*cos(phi); ip[1]=r*rmax*Xmax*sin(phi); ip[2]=0;
                    iv[0]=iv[1]=0; iv[2]=sqrt(2*(E-potential->Phi(ip[0], ip[1], ip[2])));
                    if(iv[2]>0){
                        OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
                        OrbitList.back().orbitIndex = OrbitList.size()-1;
                    }
                    rmax=potential->findintersection(E, 0, Xmax*cos(phi), Xmax*sin(phi));
                    ip[0]=0; ip[1]=r*rmax*Xmax*cos(phi); ip[2]=r*rmax*Xmax*sin(phi);
                    iv[1]=iv[2]=0; iv[0]=sqrt(2*(E-potential->Phi(ip[0], ip[1], ip[2])));
                    if(iv[0]>0){
                        OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
                        OrbitList.back().orbitIndex = OrbitList.size()-1;
                    }            
                    rmax=potential->findintersection(E, Xmax*sin(phi), 0, Xmax*cos(phi));
                    ip[0]=r*rmax*Xmax*sin(phi); ip[1]=0; ip[2]=r*rmax*Xmax*cos(phi);
                    iv[0]=iv[2]=0; iv[1]=sqrt(2*(E-potential->Phi(ip[0], ip[1], ip[2])));
                    if(iv[1]>0){
                        OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
                        OrbitList.back().orbitIndex = OrbitList.size()-1;
                    }
                }
            }
        }
    }
    else   // N_dim==2
    {
        for(int iph=0; iph<nfms; iph++)
        {
            double phi = (iph+0.5)/nfms * M_PI/2;
            double X=Xmax*cos(phi);
            double Y=Xmax*sin(phi) * potential->q;
            double r=potential->findintersection(E, X, Y, 0);
            ip[0]=X*r; ip[1]=Y*r; ip[2]=0;
            iv[0]=iv[1]=iv[2]=0;
            OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
            OrbitList.back().orbitIndex = OrbitList.size()-1;
        }

        for(int ir=0; ir<nfmp; ir++)
        {
            double r = (ir+0.5)/nfmp;
            ip[0]=r*Xmax; ip[1]=0; ip[2]=0;
            iv[0]=iv[2]=0; iv[1]=sqrt(2*(E-potential->Phi(ip[0], ip[1], ip[2])));
            if(iv[1]>0){
                OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
                OrbitList.back().orbitIndex = OrbitList.size()-1;
            }
            ip[0]=0; ip[1]=r*Xmax*potential->q; ip[2]=0;
            iv[1]=iv[2]=0; iv[0]=sqrt(2*(E-potential->Phi(ip[0], ip[1], ip[2])));
            if(iv[0]>0){
                OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
                OrbitList.back().orbitIndex = OrbitList.size()-1;
            }
        }

    }

    if(NYalpha>0)
    {
        // Y-alpha start space 
        int ny=sqrt(NYalpha*1.0), na=NYalpha/ny;
        double maxx=potential->longaxisradius(E);
        double maxy=0;
        for(int niter=0; niter<50 && maxx-maxy>maxx*1e-8; niter++)
        {
            double Ey=potential->Phi(0,(maxy+maxx)/2,0);
            if(Ey>E) maxx=(maxx+maxy)/2; else maxy=(maxx+maxy)/2;
        }   // thus we have found approximately the intermediate axis radius at given energy;
        for(int iy=0; iy<ny; iy++)
        {
            double Y = (iy+0.5)/ny * maxy;
            double V=E-potential->Phi(0,Y,0);
            if(V<0) continue;   // should not happen though..
            V=sqrt(2*V);
            ip[0]=0; ip[1]=Y; ip[2]=0; iv[1]=0;
            for(int ia=0; ia<na; ia++)
            {
                double alpha = (ia+0.5)/na * M_PI;
                iv[0]=V*cos(alpha); iv[2]=V*sin(alpha);
                OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
                OrbitList.back().orbitIndex = OrbitList.size()-1;
            }
        }
    }

    // random initial conditions (if desired)
    if(NRandom<=0) return;

    const int npsampling = std::max(NRandom*10, 1000);
    std::vector<double> Xs(npsampling), Ys(npsampling), Zs(npsampling), Ws(npsampling);
    double X, Y, Z, Et, totalWeight=0, Lmax2=0, rcirc;
    for(int s=0; s<npsampling; s++)
    {   // sample accessible configuration space with appropriate weight
        double v2;
        do{
            X=rand()*Xmax/RAND_MAX;
            Y=rand()*Xmax/RAND_MAX;
            Z=rand()*Xmax/RAND_MAX;
            v2=2*(E-potential->Phi(X,Y,Z));
        } while(v2<0);
        double W=sqrt(v2);
        Xs[s]=X; Ys[s]=Y; Zs[s]=Z; Ws[s]=W; totalWeight+=W;
      
    }
    // now select points from sampled
    for(int np=0; np<NRandom; np++)
    {
        double indw=rand() * totalWeight / RAND_MAX;
        int i=0;
        while(i<npsampling-1 && indw>=Ws[i]) 
        {
            indw-=Ws[i];
            i++;
        }
        ip[0]=Xs[i]; ip[1]=Ys[i]; ip[2]=Zs[i]; 
        double V=Ws[i], L2;
        double costh=(potential->N_dim==3)? (rand()*2.0/RAND_MAX-1) : 0;
        double sinth=sqrt(1-sqr(costh));
        double phi=rand()*2*M_PI/RAND_MAX;
        iv[0]=V*sinth*cos(phi); iv[1]=V*sinth*sin(phi); iv[2]=V*costh;
        OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
        OrbitList.back().orbitIndex = OrbitList.size()-1;
        //OrbitList.back().orbitWeight=L2;
        // remove this point from sample list, to avoid duplication
        totalWeight-=Ws[i];
        Ws[i]=0;  // it won't occur again
    }
}

// prepare IC for the whole Schw.grid
void COrbitLibrary::prepareInitialConditions(int NStationary, int NPrincipalPlane, int NYalpha, int NRandom, CSchwarzschildModel* model)
{
    if(grid==NULL) return;
    OrbitList.clear();
    if(NStationary==0 && NPrincipalPlane==0 && NYalpha==0 && NRandom>0 && (potential->PotentialType()==CPotential::PT_DEHNEN || potential->PotentialType()==CPotential::PT_DEHNENBSE || potential->PotentialType()==CPotential::PT_PLUMMERBSE))
    {
        double Gamma=grid->potentialGamma();
        int numOrbits=NRandom*grid->numShells;
        gsl_rng* r=gsl_rng_alloc(gsl_rng_default);
        double abc=pow(potential->p*potential->q, 1.0/3);
        for(int o=0; o<numOrbits; o++)
        {
            double f=grid->totalMass;  // Lagrangian radius of current particle
            if(config.sm_randRefineFact!=1)   // do we need to put more orbits in the outer parts?
            {
                if((o+0.5)/numOrbits < config.sm_randRefineFrac/(config.sm_randRefineFrac + (1-config.sm_randRefineFrac)*config.sm_randRefineFact))
                    f*=(o+0.5)/numOrbits * (config.sm_randRefineFrac + (1-config.sm_randRefineFrac)*config.sm_randRefineFact);
                else
                    f*=(config.sm_randRefineFrac*(1-1/config.sm_randRefineFact) + (o+0.5)/numOrbits * (1-config.sm_randRefineFrac*(1-1/config.sm_randRefineFact)));
            }
            else
                f*=(o+0.5)/numOrbits;
            if(f>=1) f=1-1.0/grid->numShells;  // fool protection..
            double R;
            if(Gamma==-1) R = 1/sqrt(pow(f,-2.0/3)-1);   // plummer
            else {
                double b=pow(f, 1/(3-Gamma));
                R = b/(1-b);
            }
            double ip[N_DIM], iv[N_DIM], E, Phi;
            int niter=0;
            do{
                double costh=rand()*2.0/RAND_MAX-1;
                double sinth=sqrt(1-sqr(costh));
                double phi=rand()*2*M_PI/RAND_MAX;
                ip[0]=R*sinth*cos(phi); 
                ip[1]=R*sinth*sin(phi) * potential->q;
                ip[2]=R*costh * potential->p;
                Phi = potential->Phi(ip[0], ip[1], ip[2]);
            } while(Phi>model->shellEnergy[grid->numShells] && niter++<1000);  // fool protection..
            // determine velocity dispersion from spherical equivalent of Dehnen model. Rescale R to scale radius=1.
            R = R/abc;
            double K;
            if(Gamma==-1)
            {
                K =(-Phi)/6;  // for Plummer model
            }
            else
            if(Gamma==0)
            {
                K = (1+6*R)/(30+60*R+30*R*R);
            }
            else
            if(Gamma==1)
            {
                double p=log((R+1)/R);
                K = R/(1+R) * (p*pow(R,4) + (4*p-1)*pow(R,3) + (6*p-3.5)*R*R + (4*p-13.0/3)*R + p - 25.0/12);
            }
            else
            if(Gamma==2)
            {
                double p=log((R+1)/R);
                K = 6*p*pow(R, 4) + (12*p-6)*pow(R,3) + (6*p-9)*R*R -2*R + 0.5;
            }
            else
            if(Gamma==0.5)
            {
                K = sqrt(R)/5/pow(1+R, 1.5);
            }
            else
            if(Gamma==1.5)
            {
                double p=log((R+1)/R);
                K = sqrt(R/(1+R)) * (-4*p*pow(R,4) + (4-12*p)*pow(R,3) + (10-12*p)*R*R + (22.0/3-4*p)*R + 1);
            }
            else 
            {
                K = 0.2 * pow(R, Gamma-5) * pow(R+1, 4-Gamma) * gsl_sf_hyperg_2F1(5, 7-2*Gamma, 6, -1/R);
            }
            double sigma = sqrt(K/abc);  // sigma holds the 1d velocity dispersion
            double sigmax=sigma * (config.sm_randSigmaXIn + (config.sm_randSigmaXOut-config.sm_randSigmaXIn)*f);
            double sigmay=sigma * (config.sm_randSigmaYIn + (config.sm_randSigmaYOut-config.sm_randSigmaYIn)*f);
            double sigmaz=sigma * (config.sm_randSigmaZIn + (config.sm_randSigmaZOut-config.sm_randSigmaZIn)*f);
            do{
                iv[0]=gsl_ran_gaussian(r, sigmax);
                iv[1]=gsl_ran_gaussian(r, sigmay);
                iv[2]=gsl_ran_gaussian(r, sigmaz);
                E = Phi + (iv[0]*iv[0]+iv[1]*iv[1]+iv[2]*iv[2])/2;
            }
            while(E>0 && E>model->shellEnergy[grid->numShells]);  // do not produce runaway particles

            double Torbx= xround(potential->longaxisperiod(E));
            double intTime = xround(config.intTimeInPeriods * Torbx);
            double maxTime = xround(config.intTimeMaxAdaptive * Torbx);
            double intTimeStep = xround(Torbx / config.intTimeStepsPerPeriod);
            OrbitList.push_back(COrbitDesc(potential, grid, ip, iv, intTime, intTimeStep, Torbx, config.calcLyapunov, maxTime));
            // determine shell index
            int sh=0;
            while((sh < grid->numShells) && (model->shellEnergy[sh+1] < E)) sh++;
            OrbitList.back().shellIndex=sh;
            OrbitList.back().Energy=E;
        }
        // sort orbits by energy
        std::sort(OrbitList.begin(), OrbitList.end());
        for(int o=0; o<OrbitList.size(); o++) 
        {
            OrbitList[o].orbitIndex = o;
        }
    }
    else   // ordinary initial conditions at Nshell energy levels
    {
        int oc=0;
        for(int e=0; e<grid->numShells; e++)
        {
            double EnergyLevel = model->shellEnergy[e+1];
            prepareInitialConditions(NStationary, NPrincipalPlane, NYalpha, NRandom, EnergyLevel);
            for(int oi=oc; oi<OrbitList.size(); oi++)
                OrbitList[oi].shellIndex=e;
            oc=OrbitList.size();
        }
    }
}

void COrbitLibrary::modifyInitialConditions(CPotential* _potential, CSpatialGrid* _grid)
{
    if(OrbitList.size()==0) return;
    if(!_potential) return;
    if(potential) delete potential;
    potential = _potential->clone();
    grid = _grid;
    for(std::vector<COrbitDesc>::iterator o=OrbitList.begin(); o!=OrbitList.end(); o++)
    {
        o->state=0;
        o->potential=potential;
        o->grid=_grid;
        if(config.intTimeInPeriods>0)    // otherwise leave the parameters unchanged (as read from text file)
        {
            o->time=xround(config.intTimeInPeriods*o->timeUnit);
            o->timeStep=xround(o->timeUnit/config.intTimeStepsPerPeriod);
            o->maxtime=xround(config.intTimeMaxAdaptive*o->timeUnit);
        }
        o->calcLyapunov=config.calcLyapunov;
    }
}

void COrbitLibrary::removeUnused()
{
    if(OrbitList.size()==0) return;
    for(std::vector<COrbitDesc>::iterator o=OrbitList.end()-1; o!=OrbitList.begin(); )
        if((o->state)!=2)
            OrbitList.erase(o--);
        else
            --o;
}

int COrbitLibrary::markChaoticOrbits()
{
    numChaotic=0;
    for(int o=0; o<OrbitList.size(); o++)
    {
        bool ch=(OrbitList[o].lfccdiff > config.chaoticMinFreqDiff) || (OrbitList[o].lambda > config.chaoticMinLambda);
        OrbitList[o].isChaotic=ch;
        if(ch) numChaotic++;
    }
    return numChaotic;
}

bool COrbitLibrary::exportOrbits(QString fileName, bool includeCellTimes)
{
    QFile file(fileName);
    if (!file.open(QFile::WriteOnly | QFile::Text)) 
        return false;
    QTextStream out(&file);
    out << "x\ty\tz\tvx\tvy\tvz\tweight\tindex\tshell\ttime\ttimestep\ttimeunit\tlfx\tlfy\tlfz\tlfdiff\tlambda\tdescription\tchaotic\ttmax\tEtot\tEdiff\trminC\trminA\trminDev\tinertx\tinerty\tinertz";
    out << endl;

    std::ofstream strmocw, strmsam;
    int numsam=0;
    if(includeCellTimes)
    {
        QString fileNameOCW=fileName+".ocw";
        strmocw.open(fileNameOCW.toAscii().data(), std::ios::out | std::ios::binary);
        if(!strmocw) return false;
        // check if need to store sampling points
        for(std::vector<COrbitDesc>::const_iterator orb=OrbitList.begin(); orb!=OrbitList.end(); orb++)
            numsam+=orb->trajsample[0].size();
        if(numsam>0)
        {
            QString fileNameSAM=fileName+".sam";
            strmsam.open(fileNameSAM.toAscii().data(), std::ios::out | std::ios::binary);
        }
    }

    for(std::vector<COrbitDesc>::const_iterator orb=OrbitList.begin(); orb!=OrbitList.end(); orb++)
    {
        QString Description = orb->Description;
        Description.replace(' ','_');
        out << QString::number(orb->ip[0], 'f', 10) << "\t" << QString::number(orb->ip[1], 'f', 10) << "\t" << QString::number(orb->ip[2], 'f', 10) << "\t" <<
               QString::number(orb->iv[0], 'f', 10) << "\t" << QString::number(orb->iv[1], 'f', 10) << "\t" << QString::number(orb->iv[2], 'f', 10) << "\t" <<
               orb->orbitWeight << "\t" << (orb->orbitIndex+1) << "\t" << (orb->shellIndex+1) << "\t" << 
               orb->time  << "\t" << orb->timeStep << "\t" << orb->timeUnit << "\t" <<
               orb->lf[0] << "\t" << orb->lf[1] << "\t" << orb->lf[2] << "\t" <<
               orb->lfccdiff << "\t" << (orb->calcLyapunov ? orb->lambda : -1.0) << "\t" <<
               Description << "\t" << (orb->isChaotic?"true":"false") << "\t" <<
               orb->maxtime << "\t" << 
               (potential->Phi(orb->ip[0], orb->ip[1], orb->ip[2]) + (sqr(orb->iv[0]) + sqr(orb->iv[1]) + sqr(orb->iv[2]))/2) << "\t" <<
               orb->Ediff << "\t" <<
               orb->inertia[0] << "\t" << orb->inertia[1] << "\t" << orb->inertia[2] << "\t" <<
               endl;
        if(includeCellTimes && orb->celltimes.size()!=0)
        {
            strmocw.write((char*)&orb->celltimes.front(), orb->celltimes.size()*sizeof(float));
            strmocw.write((char*)&orb->cellVr.front(), orb->cellVr.size()*sizeof(float));
            strmocw.write((char*)&orb->cellVt.front(), orb->cellVt.size()*sizeof(float));
            int NSP=orb->trajsample[0].size();
            strmsam.write((char*)&NSP, sizeof(int));
            if(NSP>0)
                for(int d=0; d<2*N_DIM; d++)
                    strmsam.write((char*)&orb->trajsample[d].front(), NSP*sizeof(float));
        }
    }
    file.close();
    if(includeCellTimes)
    {
        strmocw.close();
        if(numsam)
            strmsam.close();
    }
    return true;
}

bool COrbitLibrary::importOrbits(QString fileName, bool includeCellTimes)
{
    try{
    QFile file(fileName);
    if (!file.open(QFile::ReadOnly | QFile::Text)) 
        return false;
    QTextStream strm(&file);
    QString ln0 = strm.readLine();
    QStringList fields0=ln0.split(QRegExp("\\s+"));
    if(fields0.size()<6)
        return false;
    OrbitList.clear();
    numChaotic=0;

    std::ifstream strmocw, strmsam;
    if(includeCellTimes)
    {
        QString fileNameOCW=fileName+".ocw";
        strmocw.open(fileNameOCW.toAscii().data(), std::ios::in | std::ios::binary);
        if(!strmocw) return false;
        // reserve memory for vector OrbitList, based on the number of orbits estimated from filesize of .ocw
        strmocw.seekg(0,std::ios_base::end);
        OrbitList.reserve( strmocw.tellg() / (3*(grid->cellsInModel+1)*sizeof(float)) );
        strmocw.seekg(0,std::ios_base::beg);
        QString fileNameSAM=fileName+".sam";
        strmsam.open(fileNameSAM.toAscii().data(), std::ios::in | std::ios::binary);
    }

    while(!strm.atEnd())
    {
        QString ln = strm.readLine();
        QStringList fields=ln.split(QRegExp("\\s+"));
        int s=fields.size();
        if(s>=6)
        {
            double IP[N_DIM], IV[N_DIM];
            float LF[N_DIM];
            for(int i=0; i<N_DIM; i++)
            {
                IP[i]=fields[i].toDouble();
                IV[i]=fields[i+3].toDouble();
                if(s>=14)  LF[i]=fields[i+12].toFloat();
            }
            double orbitWeight = (s>6)? fields[6].toDouble() : 0;
            int orbitIndex     = (s>7)? fields[7].toInt()-1  : OrbitList.size();
            int shellIndex     = (s>8)? fields[8].toInt()-1  : -1;
            double intTime     = (s>9)? fields[9].toDouble() : 0;
            double intTimeStep = (s>10)?fields[10].toDouble(): 0;
            double intTimeUnit = (s>11)?fields[11].toDouble(): 0;
            if(intTimeUnit==0)  // calculate it
            {
                double E = potential->Phi(IP[0], IP[1], IP[2]) + (IV[0]*IV[0] + IV[1]*IV[1] + IV[2]*IV[2])/2;
                intTimeUnit = potential->longaxisperiod(E);
                if(intTimeUnit<0) intTimeUnit=0; 
            }
            double lambda      = (s>16)?fields[16].toDouble(): 0;
            bool calcLyapunov=false;
            if(lambda<0) lambda=0; else calcLyapunov=true;
            COrbitDesc od=COrbitDesc(potential, grid, IP, IV, intTime, intTimeStep, intTimeUnit, calcLyapunov);
            
            for(int c=0; c<N_DIM; c++) od.lf[c]=LF[c];
            od.lfccdiff    = (s>15)?fields[15].toDouble(): 0;
            od.lambda      = lambda;
            od.Description = (s>17)?fields[17].replace('_',' '):"?";
            od.state = (od.Description=="" || od.Description=="?")?0:2;
            od.isChaotic   = (s>18) && (fields[18]=="true" || fields[18]=="TRUE" || fields[18]=="1");
            if(od.isChaotic) numChaotic++;
            od.maxtime = (s>19)? fields[19].toDouble() : intTime;
            od.Energy = (s>20)? fields[20].toDouble(): 0;
            od.Ediff = (s>21)? fields[21].toDouble(): 0;
            od.shellIndex= shellIndex;
            od.orbitIndex= orbitIndex;
            od.orbitWeight=orbitWeight;
            od.inertia[0] =(s>24)? fields[22].toDouble(): 0;
            od.inertia[1] =(s>24)? fields[23].toDouble(): 0;
            od.inertia[2] =(s>24)? fields[24].toDouble(): 0;
            if(includeCellTimes) 
            {
                od.celltimes.resize(grid->cellsInModel+1);
                strmocw.read((char*)&od.celltimes.front(), (grid->cellsInModel+1)*sizeof(float));
                od.cellVr.resize(grid->cellsInModel+1);
                strmocw.read((char*)&od.cellVr.front(), (grid->cellsInModel+1)*sizeof(float));
                od.cellVt.resize(grid->cellsInModel+1);
                strmocw.read((char*)&od.cellVt.front(), (grid->cellsInModel+1)*sizeof(float));
                if(!strmocw) return false;
                if(strmsam)
                {
                    int NSP = NUM_SAMPLING_POINTS;
                    strmsam.read((char*)&NSP, sizeof(int));
                    if(NSP>0)
                      for(int d=0; d<N_DIM*2; d++)
                      {
                        od.trajsample[d].resize(NSP);
                        strmsam.read((char*)&od.trajsample[d].front(), NSP*sizeof(float));
                        if(!strmsam)
                            od.trajsample[d].clear();
                        else
                            od.numSamplingPoints=NSP;
                      }
                }
            }
            OrbitList.push_back(od);
        }
    }

    file.close();
    nComplete=OrbitList.size();
    return (nComplete>0);
    }
    catch(const std::bad_alloc&) {
        return false;
    }
}

QString COrbitLibrary::getOrbitPopulation(int numshellOrbit, int numcellMin, int numcellMax)
{
    if(OrbitList.size()==0) return "Error";
    QMap<QString, double> families, familiesr, familiesc;
    double totalweight=0, weightreg=0, weightch=0;
    for(int o=0; o<OrbitList.size(); o++)
    {
        double w=0;
        if(numcellMin>=0 && OrbitList[o].celltimes.size()>0)
        {   // look for orbit population in given cell(s)
            for(int c=numcellMin; c<=numcellMax; c++)
                w += OrbitList[o].orbitWeight * OrbitList[o].celltimes[c];
        }
        else   // look for orbit population in energy shell(s)
            if(numshellOrbit==0 || OrbitList[o].shellIndex==numshellOrbit-1) w=1.0+rand()*0.00001/RAND_MAX;   // add random to avoid duplicates
        if(w==0) continue;
        QString family=OrbitList[o].Description;
        family.replace("chaotic ", "", Qt::CaseInsensitive);
        if(family.contains("SAT")&&family.contains("resonance")) family="SAT";  // merge all subgroups like 'SAT 6:6:7' into one
        if(family.contains("LAT")&&family.contains("resonance")) family="LAT";
        families[family] += w;
        if(OrbitList[o].isChaotic)
        {
            familiesc[family] += w;
            weightch+=w;
        }
        else
        {
            familiesr[family] += w;
            weightreg+=w;
        }
        totalweight+=w;
    }
    QList<double> weights = families.values();
    qSort(weights.begin(), weights.end(), qGreater<double>());
    int nf=0; 
    QList<double>::ConstIterator iter=weights.constBegin();
    QString info = QString::number(weightreg/totalweight*100, 'f', 2)+"% regular\t";
    while(nf<8 && iter!=weights.constEnd())
    {
        QString f=families.key(*iter);
        info +=  f + ": " + QString::number((familiesr.value(f)+familiesc.value(f))/totalweight*100, 'f', 2) + "% (" + QString::number(familiesc.value(f)/totalweight*100, 'f', 2) + "%)\t";
        nf++;
        ++iter;
    }
    return info;
}

}