#pragma once
#include <QString>
#include <complex>
#include <vector>
#include "potential.h"
#include "schwarzschild.h"

namespace smile{

struct SpectralLine
{
    double freq;  // frequency
    double ampl;  // amplitude
    double phase; // phase (complex amplitude = ampl * exp(I*phase)
    bool precise; // whether it was obtained using more precise Hunter's method
    SpectralLine(double Freq=0, double Ampl=0, double Phase=0, bool Precise=true)
    {   freq=Freq; ampl=Ampl; phase=Phase; precise=Precise; }
    inline bool operator<(SpectralLine S) const
    {   return (ampl>S.ampl); }    // sort in reverse order of amplitude
    inline operator double() const { return freq; }
};
typedef std::vector<SpectralLine> Spectrum;
typedef std::complex<double> complexd;

double xround(double x);  // rounds float number to NUM_DIGITS_ROUND significant digits. Used to initialize ICs

// declare CSpatialGrid and CSchwarzschildModel classes defined in schwarzschild.h
class CSpatialGrid;
class CSchwarzschildModel;

class COrbit
{
public:
    COrbit(const CPotential* _potential, const CSpatialGrid* _grid, double _ip[N_DIM], double _iv[N_DIM], double _timeStep, double _timeUnit, bool _calcLyapunov, bool _usePS=false, int _PS1=0, int _PS2=1);
    ~COrbit(void);

    const int N_dim;                 // copied from potential->N_dim, for convenience
    double ip[N_DIM], iv[N_DIM];     // initial conditions (position and velocity)
    double fp[N_DIM], fv[N_DIM];     // final coordinates
    double intTime;                  // integration time 
    const double timeStep;           // constant timestep of output trajectory
    const double timeUnit;           // period of x-axis orbit with given energy (unit of time used in dimensionless frequencies)
    const bool calcLyapunov;         // whether to calculate Lyapunov exponent
    const bool usePS;                // whether to use Poincare section
    const int PS1, PS2;              // index of coordinates to use in Poincare section: (0, 1) gives  x, x' plane for motion in x,y plane
    int nPoints, nActual;            // number of points in output trajectory
    vectord tp[N_DIM], tv[N_DIM];    // calculated trajectory
    vectord celltimes;               // time spent in each cell of Schwarzschild's model
    vectord cellVr, cellVt;          // sum of squared velocity * time spent in cell; for radial and tangential velocity;
    double amz, aamz, amzrms;        // mean values of z-axis angular momentum, its absolute magnitude and its root-mean-square
    double amx, aamx, amxrms;        // mean values of x-axis angular momentum, its absolute magnitude and its root-mean-square
    double minr, maxr;               // minimal distance from origin
    vectord lnw;                     // ln(|w|) - norm of deviation vector, lambda - finite-time Lyapunov exponent
    double lnwadd;                   // addition term - required for renormalization so that |w| remains not too large
    QString Description;             // description of orbit type
    QString DescriptionLFCCdiff;     // description based on rate of LFCC diffusion
    vectorpd ps;                     // Poincare section  (x - xdot plane)
    double Einit, Edif;              // energy conservation check

    void integrateToTime(double time);   // start or continue integration up to given time
    void finishIntegration();        // perform normalization of some statistical quantities such as Lz or celltimes
    void integrateAdaptiveTime(double minTime, double maxTime, double ctdelta);  // perform integration for a variable time interval, using Pfenniger's method for checking convergence of orbit cell occupation
    double calcLambda();             // estimate Lyapunov exponent
	void findLines(vectord::const_iterator traj[N_DIM], int npoints, Spectrum *sl, vectord *fa=NULL);  // perform Fourier transform of trajectory, find spectral lines and store in sl[N_dim]; if fa!=NULL, store amplitude of spectrum in fa[N_dim]
    void performFrequencyAnalysis(Spectrum *sl, vectord *fa=NULL, int start=0, int count=0);           // extract frequencies and perform orbit classification; by default on whole orbit, but may use part of orbit as well. Output spectral lines in sl, if necessary - output whole spectrum in fa 
    bool calcLFCCoverInterval(int start, int count, double* sli);   // calculate LFCCs over a part of orbit: output sli=double[N_DIM] - array of LFCCs
    double calcLFCCdiffusion(int start=0, int count=0);             // evaluate rate of diffusion of fundamental frequencies over given interval (whole orbit by default)
    bool calcInertiaOverInterval(int start, int count, double* sli);   // calculate LFCCs over a part of orbit: output sli=double[N_DIM] - array of LFCCs

    const CPotential* potential;     // handler to potential used in calculations
    long      nfcn, nstep, naccpt, nrejct;   // statistics of force evaluations, timesteps, etc.
private:
    const CSpatialGrid* grid;        // handler to grid used in Schwarzschild modelling (if needed, otherwise 0)
    int cellprev;                    // cell index at previous timestep (for recording time-spent-in-cell)
    int drdtsign;                    // direction of motion at prev timestep
    int nperi;                       // number of pericenter passages
    bool needrenorm, allowrenorm;    // flag set when renormalization of deviation vector required (usually close to pericenter), and when it is allowed to do (in apocenter)
    double devvec[N_DIM*2];          // holds value of deviation vector between integration periods
    vectord latestcelltimes;         // used in Pfenniger method: holds celltimes on the latest interval of integration
    double latestIntTime;            // duration of latest integration interval
    void calcCellTime(double tlow, double tupp, int clow, int cupp, double vr2, double vt2);  // record time-in-cell with fraction-of-period accuracy
    double diffCellTimes();          // calculate difference in celltimes between latest interval and whole integration time
    static void tuneNonPrime(int &n);// ensure that n does not have large prime divisors, if needed, decrease it

    // dop853.c encapsulated variables and functions - 7/8 order Runge-Kutta integrator
    double    hout, xold, xout;
    unsigned  nrds, *indir;
    double    *yy1, *k1, *k2, *k3, *k4, *k5, *k6, *k7, *k8, *k9, *k10;
    double    *rcont1, *rcont2, *rcont3, *rcont4;
    double    *rcont5, *rcont6, *rcont7, *rcont8;

    void fcn(unsigned n, double x, double *y, double *f);   // rhs of the differential equation
    void solout(long nr, double xold, double x, double* y, unsigned n);  // performs action at each timestep

    int dop853
     (unsigned n,      /* dimension of the system <= UINT_MAX-1*/
      //  FcnEqDiff fcn,   /* function computing the value of f(x,y) */
      double x,        /* initial x-value */
      double* y,       /* initial values for y */
      double xend,     /* final x-value (xend-x may be positive or negative) */
      double* rtoler,  /* relative error tolerance */
      double* atoler,  /* absolute error tolerance */
      int itoler,      /* switch for rtoler and atoler */
      //  SolTrait solout, /* function providing the numerical solution during integration */
      int iout,        /* switch for calling solout */
      //  FILE* fileout,   /* messages stream */
      double uround,   /* rounding unit */
      double safe,     /* safety factor */
      double fac1,     /* parameters for step size selection */
      double fac2,
      double beta,     /* for stabilized step size control */
      double hmax,     /* maximal step size */
      double h,        /* initial step size */
      long nmax,       /* maximal number of allowed steps */
      int meth,        /* switch for the choice of the coefficients */
      long nstiff,     /* test for stiffness */
      unsigned nrdens, /* number of components for which dense outpout is required */
      unsigned* icont, /* indexes of components for which dense output is required, >= nrdens */
      unsigned licont  /* declared length of icon */
     );
public:    
    double contd8
     (unsigned ii,     /* index of desired component */
      double x         /* approximation at x */
    );
private:
    // internal integration initialization routine
    double hinit (unsigned n, /*FcnEqDiff fcn,*/ double x, double* y,
	  double posneg, double* f0, double* f1, double* yy1, int iord,
	  double hmax, double* atoler, double* rtoler, int itoler);
    
    // internal integration routine
    int dopcor (unsigned n, /*FcnEqDiff fcn,*/ double x, double* y, double xend,
	  double hmax, double h, double* rtoler, double* atoler,
	  int itoler, /*FILE* fileout, SolTrait solout,*/ int iout,
	  long nmax, double uround, int meth, long nstiff, double safe,
	  double beta, double fac1, double fac2, unsigned* icont);

    // low-order leapfrog method provided for treecode, which has discontinuous potential/forces and is poorly handled by RK integrator
    bool leapfrog;
    double yprev[4*N_DIM], aprev[2*N_DIM], anew[N_DIM];
    int intlf(unsigned n, double x, double* y, double xend);
    double contlf
     (unsigned ii,     /* index of desired component */
      double x         /* approximation at x */
    );

};

class COrbitDesc                  // shortened orbit data; used as an item in OrbitLibrary
{
public:
    const CPotential* potential;  // potential used for orbit integration
    const CSpatialGrid* grid;     // if used for Schwarzschild modelling, this holds a reference to spatial grid to compute cell occupation numbers
    double timeStep, timeUnit;    // parameters for orbit integration
    bool calcLyapunov;            // whether to calculate Lyapunov exponent
    double ip[N_DIM], iv[N_DIM];  // initial conditions
    double time, maxtime;         // integration time
    int state;                    // 0 - not started, 1 - computing, 2 - finished
    float Energy;                 // (initial) total energy
    // results:
    float Ediff;                  // energy conservation error
    float lf[N_DIM];              // leading frequencies
    float lfccdiff;               // rate of LFCC diffusion
    float lambda;                 // Lyapunov exponent
    float inertia[N_DIM];         // inertia tensor components (only diagonal in Cartesian coordinates)
    QString Description;          // text description
    vectorf celltimes;            // time fraction spent in each cell (for Schwarzschild modelling)
    vectorf cellVr, cellVt;       // velocity dispersion in each cell
    vectorf trajsample[N_DIM*2];  // sample points on trajectory (for initializing N-body initial conditions)
    double orbitWeight;           // weight in Schwarzschild model
    int shellIndex;               // index of energy shell to which this orbit belongs
    int orbitIndex;               // index of parent orbit in the case of N-body exported model, or simply the number of orbit
    bool isChaotic;               // whether orbit is considered chaotic in Schwarzschild modelling
    int numSamplingPoints;        // number of stored points sampled from trajectory (used to generate Nbody initial conditions)
    COrbitDesc();
    COrbitDesc(const COrbitDesc &src);
    COrbitDesc& operator=(const COrbitDesc &src);
    COrbitDesc(const CPotential* _potential, const CSpatialGrid* _grid, double _ip[N_DIM], double _iv[N_DIM], 
        double _time, double _timeStep, double _timeUnit, bool _calcLyapunov, double _maxtime=0);      // initialize new orbit
    ~COrbitDesc();
    void run();                   // create internal COrbit instance and run orbit integration and analysis
    inline bool operator<(COrbitDesc D) const
    {   return (Energy<D.Energy); }    // sort by total energy (used in creating random initial conditions)
};

class COrbitLibrary
{
public:
    std::vector<COrbitDesc> OrbitList; // array of shortened orbit info
    volatile int nComplete;            // number of orbits already processed
    int numChaotic;                    // number of chaotic orbits
    CPotential* potential;             // potential for all orbits
    const CSpatialGrid* grid;          // if doing integration for Schwarzschild model, this holds the reference to spatial grid to compute cell occupation numbers

    COrbitLibrary(const CPotential* _potential, const CSpatialGrid* _grid);
    ~COrbitLibrary();
    bool findPeriodicOrbit(double E, int coord1, int coord2, double& cross1);    // find 1:1 periodic orbit at given energy in coord1-coord2 plane, return starting position in 1st coordinate
    void prepareInitialConditions(int NStationary, int NPrincipalPlane, int NYalpha, int NRandom, double E);   // create initial conditions for given energy
    void prepareInitialConditions(int NStationary, int NPrincipalPlane, int NYalpha, int NRandom, CSchwarzschildModel* model);  // create initial conditions for the whole Schwarzschild model
    void modifyInitialConditions(CPotential* _potential, CSpatialGrid* _grid=NULL);    // change potential and grid pointers and integration time/timestep for each orbit
    void removeUnused();               // if integration of the whole library was terminated prematurely, delete unfinished orbits
    int markChaoticOrbits();           // mark chaotic orbits based on criteria in config (min.freq.diff. or Lyapunov exp.)
    bool exportOrbits(QString fileName, bool includeCellTimes);    // export orbits to text file; if includeCellTimes=true, export also binary data (cell file and trajectory sample file)
    bool importOrbits(QString fileName, bool includeCellTimes);    // import orbits from text file; if includeCellTimes=true, import also binary data
    QString getOrbitPopulation(int numshellOrbit, int numcellMin=-1, int numcellMax=-1);   // return a string containing orbit population (5 most abundant orbit families)
};

}