#define _USE_MATH_DEFINES
#include <math.h>
#include <algorithm>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_legendre.h>
#include <gsl/gsl_sf_gegenbauer.h>
#include <gsl/gsl_sf_gamma.h>
#include "potential.h"

#ifdef WIN32
double _HUGE=1e303;   // !!! hack: MSVC2005 linker cannot find that symbol elsewhere for some strange reason
#endif

namespace smile{

// base class
CPotential::CPotential(): N_dim(3), q(1), p(1), Mbh(0) {}
CPotential::CPotential(int _N_dim, double _q, double _p, double _Mbh): N_dim(_N_dim), q(_q), p(_p), Mbh(_Mbh) {}

CPotential::~CPotential() {}

struct CPotentialParam1{
    const CPotential* P;
    double E;
};
struct CPotentialParam4{
    const CPotential* P;
    double E, X, Y, Z;
};

// general functions for any potential
double twoV2(double x, void* params)
{
    double E=((CPotentialParam1*)params)->E;
    return E-((CPotentialParam1*)params)->P->Phi(x,0,0);
}
double Torb(double x, void* params)
{
    if(x==0) return 0;
    double E=((CPotentialParam1*)params)->E;
    double V2=2*(E-((CPotentialParam1*)params)->P->Phi(x,0,0));
    if(V2<=0) return 0; else return 1/sqrt(V2);
}
double k_intersect(double k, void* params)
{
    double E= ((CPotentialParam4*)params)->E;
    double X= ((CPotentialParam4*)params)->X;
    double Y= ((CPotentialParam4*)params)->Y;
    double Z= ((CPotentialParam4*)params)->Z;
    return E - ((CPotentialParam4*)params)->P->Phi(X*k,Y*k,Z*k);
}
double minusPhi(double x, void* params)
{
    if(((CPotentialParam4*)params)->E)
        return -((CPotentialParam1*)params)->P->Phi(0,x,0);
    else
        return -((CPotentialParam1*)params)->P->Phi(x,0,0);
}

double CPotential::longaxisradius(double E) const
{
    // find maximal x-axis elongation for given energy
    double Xlow=1, Xupp=1, Xmax;
    while ((Phi(Xlow,0,0)>E) && (Xlow>1e-6)) Xlow/=2;
    if(Xlow<1e-6) 
    {
        return -1;  // signal of error (E lower than minimum of the potential)
    }
    else
    {
        while ((Phi(Xupp,0,0)<E) && (Xupp<1e6)) Xupp*=2;
        if(Xupp>1e6) 
        {
            return -1;  // signal of error (e.g. E>0 and potential is everywhere negative)
        }
    }
    gsl_function F;
    CPotentialParam1 PP;
    PP.E=E;
    PP.P=this;
    F.function=&twoV2;
    F.params=&PP;
    gsl_root_fsolver *s = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
    gsl_root_fsolver_set (s, &F, Xlow, Xupp);
    int status=0, iter=0;
    do{
        iter++;
        status = gsl_root_fsolver_iterate (s);
        Xmax = gsl_root_fsolver_root (s);
        Xlow = gsl_root_fsolver_x_lower (s);
        Xupp = gsl_root_fsolver_x_upper (s);
        status = gsl_root_test_interval (Xlow, Xupp, 0, 1e-9);
    }
    while (status == GSL_CONTINUE && iter < 100);
    gsl_root_fsolver_free (s);
    return Xmax;    
}

double CPotential::longaxisperiod(double E, double Xmax) const
{
    // calculate T_orb (for x-axis orbit)
    if(Xmax<0)
        Xmax=longaxisradius(E);
    if(Xmax<0) return -1;  // error
    double result, error;
    gsl_function F;
    CPotentialParam1 PP;
    PP.E=E;
    PP.P=this;
    F.function=&Torb;
    F.params=&PP;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);
    gsl_integration_qags (&F, 0, Xmax, 0, 1e-7, 1000, w, &result, &error); 
    gsl_integration_workspace_free (w);
    /*size_t neval;
    gsl_integration_qng (&F, 0, Xmax, 0, 1e-7, &result, &error, &neval); */
    return 4*result;
}

double CPotential::findintersection(double E, double X, double Y, double Z) const
{
    double klow=1, kupp=1, k;
    while ((Phi(klow*X,klow*Y,klow*Z)>E) && (klow>1e-6)) klow/=2;
    if(klow<1e-6) 
        return -1;  // signal of error (E lower than minimum of the potential)
    while ((Phi(kupp*X,kupp*Y,kupp*Z)<E) && (kupp<1e4)) kupp*=2;
    if(kupp>1e4) 
        return -1;  // signal of error (e.g. E>0 and potential is everywhere negative)
    gsl_function F;
    CPotentialParam4 PP;
    PP.E=E;
    PP.P=this;
    PP.X=X;
    PP.Y=Y;
    PP.Z=Z;
    F.function=&k_intersect;
    F.params=&PP;
    gsl_root_fsolver *s = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);
    gsl_root_fsolver_set (s, &F, klow, kupp);
    int status=0, iter=0;
    do{
        iter++;
        status = gsl_root_fsolver_iterate (s);
        k = gsl_root_fsolver_root (s);
        klow = gsl_root_fsolver_x_lower (s);
        kupp = gsl_root_fsolver_x_upper (s);
        status = gsl_root_test_interval (klow, kupp, 0, 1e-9);
    }
    while (status == GSL_CONTINUE && iter < 100);
    gsl_root_fsolver_free (s);
    return k;
}

// Logarithmic potential
CPotentialLog::CPotentialLog(int _N_dim, double _q, double _p, double _Mbh, double _Rc): CPotential(_N_dim, _q, _p, _Mbh), Rc(_Rc) {}

CPotential* CPotentialLog::clone() const { return new CPotentialLog(*this); }
CPotential::POTENTIALTYPE CPotentialLog::PotentialType() const { return PT_LOG; }
char* CPotentialLog::myName() { return "Logarithmic"; }
char* CPotentialLog::PotentialName() const { return myName(); }

double CPotentialLog::Rho(double X, double Y, double Z, double t) const
{
    double m2=sqr(Rc) + sqr(X) + sqr(Y)/sqr(q) + sqr(Z)/sqr(p);
    return (0.5*(1+1/sqr(q)+1/sqr(p))/m2 - (sqr(X) + sqr(Y/sqr(q)) + sqr(Z/sqr(p)))/sqr(m2))/M_PI;
}

double CPotentialLog::Phi(double X, double Y, double Z, double t) const
{
    return log(sqr(Rc) + sqr(X) + sqr(Y)/sqr(q) + sqr(Z)/sqr(p)) - ((Mbh>0)?(Mbh/sqrt(sqrt(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z))):0);
}

// r.h.s. of the differential equation: for given time 't' and variables 'v', return their difference 'f'
void CPotentialLog::DiffEq(unsigned n, double t, double *v, double *f) const
{
    bool calcLyapunov = n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double m2=(sqr(X) + sqr(Y)/sqr(q) + sqr(Z)/sqr(p) + sqr(Rc));
    double r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
    f[N_dim]   = -2*X/m2    - ((Mbh>0)?(Mbh*X/r3):0);
    f[N_dim+1] = -2*Y/m2/sqr(q) - ((Mbh>0)?(Mbh*Y/r3):0);
    if(N_dim==3)
        f[N_dim+2] = -2*Z/m2/sqr(p) - ((Mbh>0)?(Mbh*Z/r3):0);
    if(calcLyapunov)
    {
#ifdef LYAPUNOVVAREQ
        // variational equation
        double r5=pow(sqr(X) + sqr(Y) + sqr(Z), 2.5);
        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        double d2Vdxdy = -X*Y * (4/sqr(q*m2) + ((Mbh>0)?(3*Mbh/r5):0));
        double d2Vdxdz = (N_dim==3)?( -X*Z * (4/sqr(p*m2)   + ((Mbh>0)?(3*Mbh/r5):0)) ):0;
        double d2Vdydz = (N_dim==3)?( -Y*Z * (4/sqr(q*p*m2) + ((Mbh>0)?(3*Mbh/r5):0)) ):0;
        f[N_dim*3]   = -wx * (2/m2   -4*sqr(X/m2   ) - 3*Mbh*sqr(X)/r5 + Mbh/r3) - wy*d2Vdxdy - wz*d2Vdxdz;
        f[N_dim*3+1] = -wy * (2/m2/sqr(q)-4*sqr(Y/m2/sqr(q)) - 3*Mbh*sqr(Y)/r5 + Mbh/r3) - wx*d2Vdxdy - wz*d2Vdydz;
        if(N_dim==3)
            f[N_dim*3+2] = -wz * (2/m2/sqr(p)-4*sqr(Z/m2/sqr(p)) - 3*Mbh*sqr(Z)/r5 + Mbh/r3) - wx*d2Vdxdz - wy*d2Vdydz;
#else
        // integrate nearby orbit
        X=v[N_dim*2];
        Y=v[N_dim*2+1];
        Z=(N_dim==3)?v[N_dim*2+2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2] = v[N_dim*3+2];
        m2=(sqr(X) + sqr(Y)/sqr(q) + sqr(Z)/sqr(p) + sqr(Rc));
        r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
        f[N_dim*3]   = -2*X/m2    - ((Mbh>0)?(Mbh*X/r3):0);
        f[N_dim*3+1] = -2*Y/m2/sqr(q) - ((Mbh>0)?(Mbh*Y/r3):0);
        if(N_dim==3)
            f[N_dim*3+2] = -2*Z/m2/sqr(p) - ((Mbh>0)?(Mbh*Z/r3):0);
#endif
    }
}

// Dehnen potential
CPotentialDehnen::CPotentialDehnen(int _N_dim, double _q, double _p, double _Mbh, double _Gamma): 
    CPotential(_N_dim, _q, _p, _Mbh), 
    Gamma(std::max<double>(0, std::min<double>(_Gamma, 2.0))) {}

CPotential* CPotentialDehnen::clone() const { return new CPotentialDehnen(*this); }
CPotential::POTENTIALTYPE CPotentialDehnen::PotentialType() const { return PT_DEHNEN; }
char* CPotentialDehnen::myName() { return "Dehnen"; }
char* CPotentialDehnen::PotentialName() const { return myName(); }

double CPotentialDehnen::Rho(double X, double Y, double Z, double t) const
{
    double m = sqrt(sqr(X) + sqr(Y/q) + sqr(Z/p));
    return (3-Gamma)/(4*M_PI*p*q) * pow(m, -Gamma) * pow(1+m, Gamma-4);
}

struct CPotentialParamD{
    const CPotentialDehnen* P;
    double X, Y, Z;
};
struct CPotentialParamD2{
    const CPotentialDehnen* P;
    double X, Y, Z;
    double wx, wy, wz;
};

double intPhiD(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    const CPotentialDehnen* P=((CPotentialParamD*)params)->P;
    double m = s * sqrt( sqr(X) + sqr(Y)/(1-(1-sqr(P->q))*sqr(s)) + sqr(Z)/(1-(1-sqr(P->p))*sqr(s)));
    double numerator = (P->Gamma==2)? (log((1+m)*s/m) - 1/(1+m) - log(s)) :
        (1 - (3-P->Gamma)*pow(m/(m+1), 2-P->Gamma) + (2-P->Gamma)*pow(m/(m+1), 3-P->Gamma))/(2-P->Gamma);
    return numerator / sqrt( (1-(1-sqr(P->q))*sqr(s)) * (1-(1-sqr(P->p))*sqr(s)));
}

double CPotentialDehnen::Phi(double X, double Y, double Z, double t) const
{
    double result, error;
    CPotentialParamD PP;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.P=this;
    gsl_function F;
    F.function=&intPhiD;
    F.params=&PP;
    size_t neval;
    gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
    return -result - ((Mbh>0)?(Mbh/sqrt(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z))):0);
}

double intForceDx(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    const CPotentialDehnen* P=((CPotentialParamD*)params)->P;
    double m = s * sqrt( sqr(X) + sqr(Y)/(1+(sqr(P->q)-1)*sqr(s)) + sqr(Z)/(1+(sqr(P->p)-1)*sqr(s)));
    return sqr(s) * pow(m, -P->Gamma) * pow(1+m, P->Gamma-4) /
        sqrt( (1+(sqr(P->q)-1)*sqr(s)) * (1+(sqr(P->p)-1)*sqr(s)));
}
double intForceDy(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    const CPotentialDehnen* P=((CPotentialParamD*)params)->P;
    double m = s * sqrt( sqr(X)/(sqr(P->q)+(1-sqr(P->q))*sqr(s)) + sqr(Y)/sqr(P->q) + sqr(Z)/(sqr(P->q)+(sqr(P->p)-sqr(P->q))*sqr(s)));
    return sqr(s) * pow(m, -P->Gamma) * pow(1+m, P->Gamma-4) /
        sqrt( (sqr(P->q)+(sqr(P->p)-sqr(P->q))*sqr(s)) * (sqr(P->q)+(1-sqr(P->q))*sqr(s)));
}
double intForceDz(double s, void* params)
{
    double X= ((CPotentialParamD*)params)->X;
    double Y= ((CPotentialParamD*)params)->Y;
    double Z= ((CPotentialParamD*)params)->Z;
    const CPotentialDehnen* P=((CPotentialParamD*)params)->P;
    double m = s * sqrt( sqr(X)/(sqr(P->p)+(1-sqr(P->p))*sqr(s)) + sqr(Y)/(sqr(P->p)+(sqr(P->q)-sqr(P->p))*sqr(s)) + sqr(Z)/sqr(P->p));
    return sqr(s) * pow(m, -P->Gamma) * pow(1+m, P->Gamma-4) /
        sqrt( (sqr(P->p)+(1-sqr(P->p))*sqr(s)) * (sqr(P->p)+(sqr(P->q)-sqr(P->p))*sqr(s)));
}

double intForceD2x(double s, void* params)
{
    double X= ((CPotentialParamD2*)params)->X;
    double Y= ((CPotentialParamD2*)params)->Y;
    double Z= ((CPotentialParamD2*)params)->Z;
    double wx=((CPotentialParamD2*)params)->wx;
    double wy=((CPotentialParamD2*)params)->wy;
    double wz=((CPotentialParamD2*)params)->wz;
    const CPotentialDehnen* P=((CPotentialParamD*)params)->P;
    double m = s * sqrt( sqr(X) + sqr(Y)/(1+(sqr(P->q)-1)*sqr(s)) + sqr(Z)/(1+(sqr(P->p)-1)*sqr(s)));
    return sqr(sqr(s)) * pow(m, -P->Gamma-2) * pow(1+m, P->Gamma-5) * (P->Gamma+4*m) * 
      ( X * wx / sqrt( (1+(sqr(P->q)-1)*sqr(s)) * (1+(sqr(P->p)-1)*sqr(s))) +
        Y * wy * pow( 1 + (sqr(P->q)-1)*sqr(s), -1.5) / sqrt(1 + (sqr(P->p)-1)*sqr(s)) +
        Z * wz * pow( 1 + (sqr(P->p)-1)*sqr(s), -1.5) / sqrt(1 + (sqr(P->q)-1)*sqr(s)) );
}
double intForceD2y(double s, void* params)
{
    double X= ((CPotentialParamD2*)params)->X;
    double Y= ((CPotentialParamD2*)params)->Y;
    double Z= ((CPotentialParamD2*)params)->Z;
    double wx=((CPotentialParamD2*)params)->wx;
    double wy=((CPotentialParamD2*)params)->wy;
    double wz=((CPotentialParamD2*)params)->wz;
    const CPotentialDehnen* P=((CPotentialParamD*)params)->P;
    double m = s * sqrt( sqr(X)/(sqr(P->q)+(1-sqr(P->q))*sqr(s)) + sqr(Y)/sqr(P->q) + sqr(Z)/(sqr(P->q)+(sqr(P->p)-sqr(P->q))*sqr(s)));
    return sqr(sqr(s)) * pow(m, -P->Gamma-2) * pow(1+m, P->Gamma-5) * (P->Gamma+4*m) * 
      ( Y * wy / sqr(P->q) / sqrt( (sqr(P->q)+(sqr(P->p)-sqr(P->q))*sqr(s)) * (sqr(P->q)+(1-sqr(P->q))*sqr(s))) +
        Z * wz * pow( sqr(P->q) + (sqr(P->p)-sqr(P->q))*sqr(s), -1.5) / sqrt(sqr(P->q) + (1-sqr(P->q))*sqr(s)) +
        X * wx * pow( sqr(P->q) + (1-sqr(P->q))*sqr(s), -1.5) / sqrt(sqr(P->q) + (sqr(P->p)-sqr(P->q))*sqr(s)) );
}
double intForceD2z(double s, void* params)
{
    double X= ((CPotentialParamD2*)params)->X;
    double Y= ((CPotentialParamD2*)params)->Y;
    double Z= ((CPotentialParamD2*)params)->Z;
    double wx=((CPotentialParamD2*)params)->wx;
    double wy=((CPotentialParamD2*)params)->wy;
    double wz=((CPotentialParamD2*)params)->wz;
    const CPotentialDehnen* P=((CPotentialParamD*)params)->P;
    double m = s * sqrt( sqr(X)/(sqr(P->p)+(1-sqr(P->p))*sqr(s)) + sqr(Y)/(sqr(P->p)+(sqr(P->q)-sqr(P->p))*sqr(s)) + sqr(Z)/sqr(P->p));
    return sqr(sqr(s)) * pow(m, -P->Gamma-2) * pow(1+m, P->Gamma-5) * (P->Gamma+4*m) * 
      ( Z * wz / sqr(P->p) / sqrt( (sqr(P->p)+(1-sqr(P->p))*sqr(s)) * (sqr(P->p)+(sqr(P->q)-sqr(P->p))*sqr(s))) +
        X * wx * pow( sqr(P->p) + (1-sqr(P->p))*sqr(s), -1.5) / sqrt(sqr(P->p) + (sqr(P->q)-sqr(P->p))*sqr(s)) +
        Y * wy * pow( sqr(P->p) + (sqr(P->q)-sqr(P->p))*sqr(s), -1.5) / sqrt(sqr(P->p) + (1-sqr(P->p))*sqr(s)) );
}

void CPotentialDehnen::DiffEq(unsigned n, double t, double *v, double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
    // force calculation requires integration
    double resultx, resulty, resultz, error; 
    size_t neval;
    CPotentialParamD PP;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.P=this;
    gsl_function F;
    F.params=&PP;
    F.function=&intForceDx;
    gsl_integration_qng (&F, 0, 1, 0, 1e-7, &resultx, &error, &neval); 
    f[N_dim]   = -(3-Gamma)*X * resultx   - ((Mbh>0)?(Mbh*X/r3):0);
    F.function=&intForceDy;
    gsl_integration_qng (&F, 0, 1, 0, 1e-7, &resulty, &error, &neval); 
    f[N_dim+1] = -(3-Gamma)*Y/q * resulty - ((Mbh>0)?(Mbh*Y/r3):0);
    if(N_dim==3)
    {
        F.function=&intForceDz;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &resultz, &error, &neval); 
        f[N_dim+2] = -(3-Gamma)*Z/p * resultz - ((Mbh>0)?(Mbh*Z/r3):0);
    }
    if(calcLyapunov)
    {
#ifdef LYAPUNOVVAREQ
        // variational equation
        double r5=pow(sqr(X) + sqr(Y) + sqr(Z), 2.5);
        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        CPotentialParamD2 P2;
        P2.X=X; P2.Y=Y; P2.Z=Z;
        P2.wx=wx; P2.wy=wy; P2.wz=wz;
        P2.P=this;
        double result;
        F.params=&P2;
        F.function=&intForceD2x;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
        f[N_dim*3]   = -(3-Gamma) * (wx*resultx - X*result) + ((Mbh>0)?Mbh*( (3*sqr(X)/r5-1/r3)*wx + 3*X*Y/r5*wy + 3*X*Z/r5*wz ):0);
        F.function=&intForceD2y;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
        f[N_dim*3+1] = -(3-Gamma)/q * (wy*resulty - Y*result) + ((Mbh>0)?Mbh*( (3*sqr(Y)/r5-1/r3)*wy + 3*Y*X/r5*wx + 3*Y*Z/r5*wz ):0);
        if(N_dim==3)
        {
            F.function=&intForceD2z;
            gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
            f[N_dim*3+2] = -(3-Gamma)/p * (wz*resultz - Z*result) + ((Mbh>0)?Mbh*( (3*sqr(Z)/r5-1/r3)*wz + 3*Z*X/r5*wx + 3*Z*Y/r5*wy ):0);
        }
        /*double d2Vdxdy = -X*Y * (4/sqr(q*m2) + ((Mbh>0)?(3*Mbh/r5):0));
        double d2Vdxdz = (N_dim==3)?( -X*Z * (4/sqr(p*m2)   + ((Mbh>0)?(3*Mbh/r5):0)) ):0;
        double d2Vdydz = (N_dim==3)?( -Y*Z * (4/sqr(q*p*m2) + ((Mbh>0)?(3*Mbh/r5):0)) ):0;
        f[N_dim*3]   = -wx * (2/m2   -4*sqr(X/m2   ) - 3*Mbh*sqr(X)/r5 + Mbh/r3) - wy*d2Vdxdy - wz*d2Vdxdz;
        f[N_dim*3+1] = -wy * (2/m2/sqr(q)-4*sqr(Y/m2/sqr(q)) - 3*Mbh*sqr(Y)/r5 + Mbh/r3) - wx*d2Vdxdy - wz*d2Vdydz;
        if(N_dim==3)
            f[N_dim*3+2] = -wz * (2/m2/sqr(p)-4*sqr(Z/m2/sqr(p)) - 3*Mbh*sqr(Z)/r5 + Mbh/r3) - wx*d2Vdxdz - wy*d2Vdydz;*/
#else
        // integrate nearby orbit
        X=v[N_dim*2];
        Y=v[N_dim*2+1];
        Z=(N_dim==3)?v[N_dim*2+2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2] = v[N_dim*3+2];
        r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
        // force calculation requires integration
        PP.X=X; PP.Y=Y; PP.Z=Z;
        F.function=&intForceDx;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &resultx, &error, &neval); 
        f[N_dim*3]   = -(3-Gamma)*X * resultx    - ((Mbh>0)?(Mbh*X/r3):0);
        F.function=&intForceDy;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &resulty, &error, &neval); 
        f[N_dim*3+1] = -(3-Gamma)*Y/q * resulty    - ((Mbh>0)?(Mbh*Y/r3):0);
        if(N_dim==3)
        {
            F.function=&intForceDz;
            gsl_integration_qng (&F, 0, 1, 0, 1e-7, &resultz, &error, &neval); 
            f[N_dim*3+2] = -(3-Gamma)*Z/p * resultz    - ((Mbh>0)?(Mbh*Z/r3):0);
        }
#endif
    }
}

// Basis-set expansion for arbitrary potential (using Hernquist&Ostriker 1992 basis set)

CPotentialBSE::CPotentialBSE(int _N_dim, double _q, double _p, double _Mbh, double _Gamma, double _Alpha, int _Ncoefs_radial, int _Ncoefs_angular, 
    int npoints, double X[], double Y[], double Z[], double mass[]):
    CPotential(_N_dim, _q, _p, _Mbh), Gamma(_Gamma), Alpha(_Alpha),
	Ncoefs_radial(std::min(MAX_NCOEFS_RADIAL, _Ncoefs_radial)), 
	Ncoefs_angular(std::min(MAX_NCOEFS_ANGULAR, _Ncoefs_angular))
{
    for(int n=0; n<Ncoefs_radial; n++)
    {
        for(int l=0; l<Ncoefs_angular; l++)
            for(int m=0; m<Ncoefs_angular; m++)
                SHcoefs[n][l][m]=0;
    }
    prepareCoefs(npoints, X, Y, Z, mass);
}

CPotentialBSE::CPotentialBSE(int _N_dim, double _q, double _p, double _Mbh, double _Gamma, double _Alpha, int _Ncoefs_radial, int _Ncoefs_angular, SHcoefsArray _SHcoefs):
    CPotential(_N_dim, _q, _p, _Mbh), Gamma(_Gamma), Alpha(_Alpha), 
	Ncoefs_radial(std::min(MAX_NCOEFS_RADIAL, _Ncoefs_radial)), 
	Ncoefs_angular(std::min(MAX_NCOEFS_ANGULAR, _Ncoefs_angular))
{
    for(int n=0; n<Ncoefs_radial; n++)
    {
        for(int l=0; l<Ncoefs_angular; l++)
            for(int m=0; m<=l; m++)
                SHcoefs[n][l][m] = _SHcoefs[n][l][m];
    }
}

CPotential* CPotentialBSE::clone() const { return new CPotentialBSE(*this); }
CPotential::POTENTIALTYPE CPotentialBSE::PotentialType() const { return PT_BSE; }
char* CPotentialBSE::myName() { return "BSE"; }
char* CPotentialBSE::PotentialName() const { return myName(); }

void CPotentialBSE::prepareCoefs(int npoints, double X[], double Y[], double Z[], double mass[])
{
    if(npoints<=0)
    {
        SHcoefs[0][0][0] = 1;
        return;
    }
    double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR*2-1];
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    for(int i=0; i<npoints; i++)
    {
        double massi=(mass==NULL)?1.0/npoints:mass[i];
        double costheta=(X[i]==0&&Y[i]==0&&Z[i]==0)?0:(Z[i]/sqrt(sqr(X[i])+sqr(Y[i])+sqr(Z[i])));
        double phi=atan2(Y[i], X[i]);
        double r=sqrt(X[i]*X[i]+Y[i]*Y[i]+Z[i]*Z[i]);
        double ralpha=pow(r, 1/Alpha);
        double xi=(ralpha-1)/(ralpha+1);
        for(int m2=0; m2<Ncoefs_angular; m2++)
            gsl_sf_legendre_sphPlm_array(Ncoefs_angular*2-2, m2*2, costheta, legendre_array[m2]);

        for(int l2=0; l2<Ncoefs_angular; l2++)
        {
            int l=2*l2;
            double w=(2*l+1)*Alpha+0.5;
            double gammaw2=gsl_pow_2(gsl_sf_gamma(w));
            double phil=pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha);
            gsl_sf_gegenpoly_array(Ncoefs_radial-1, w, xi, gegenpoly_array);
            for(int n=0; n<Ncoefs_radial; n++)
                {
                    double Inl = 2*M_SQRTPI * Alpha * pow(2.0, -4*w) * gsl_sf_gamma(2*w+n) / gsl_sf_fact(n) / (n+w) / gammaw2 * (4*(n+w)*(n+w)-1)/(8*Alpha*Alpha);
                    double mult= massi * gegenpoly_array[n] * phil / Inl;
                    for(int m2=0; m2<=l2; m2++)
                        SHcoefs[n][l2][m2] += mult * legendre_array[m2][(l2-m2)*2] * cos(m2*2*phi) /* (1+(m2>0))*/;
                }
        }
    }
}

double CPotentialBSE::Phi(double X, double Y, double Z, double t) const
{
    double costheta=(X==0&&Y==0&&Z==0)?0:(Z/sqrt(sqr(X)+sqr(Y)+sqr(Z)));
    double phi=atan2(Y, X);
    double r=sqrt(X*X+Y*Y+Z*Z);
    double ralpha=pow(r, 1/Alpha);
    double xi=(ralpha-1)/(ralpha+1);
    double res=0;
    double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR*2-1];
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    for(int m2=0; m2<Ncoefs_angular; m2++)
        gsl_sf_legendre_sphPlm_array(Ncoefs_angular*2-2, m2*2, costheta, legendre_array[m2]);
    for(int l2=0; l2<Ncoefs_angular; l2++)
    {
        int l=2*l2;
        double w=(2*l+1)*Alpha+0.5;
        double resn=0;
        gsl_sf_gegenpoly_array(Ncoefs_radial-1, w, xi, gegenpoly_array);
        for(int n=0; n<Ncoefs_radial; n++)
            {
                double resm=0;
                for(int m2=0; m2<=l2; m2++)
                    resm += SHcoefs[n][l2][m2] * legendre_array[m2][(l2-m2)*2] * cos(m2*2*phi) * (1+(m2>0));
                resn += resm * gegenpoly_array[n];
            }
        res += resn * pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha) * 2*M_SQRTPI;
    }
    return -res - ((Mbh>0)?(Mbh/sqrt(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z))):0);
}

double CPotentialBSE::Rho(double X, double Y, double Z, double t) const
{   // density coming from self-consistent SHE
    double costheta=(X==0&&Y==0&&Z==0)?0:(Z/sqrt(sqr(X)+sqr(Y)+sqr(Z)));
    double phi=atan2(Y, X);
    double r=sqrt(X*X+Y*Y+Z*Z);
    double ralpha=pow(r, 1/Alpha);
    double xi=(ralpha-1)/(ralpha+1);
    double res=0;
    double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR*2-1];
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    for(int m2=0; m2<Ncoefs_angular; m2++)
        gsl_sf_legendre_sphPlm_array(Ncoefs_angular*2-2, m2*2, costheta, legendre_array[m2]);
    for(int l2=0; l2<Ncoefs_angular; l2++)
    {
        int l=2*l2;
        double w=(2*l+1)*Alpha+0.5;
        double resn=0;
        gsl_sf_gegenpoly_array(Ncoefs_radial-1, w, xi, gegenpoly_array);
        for(int n=0; n<Ncoefs_radial; n++)
            {
                double Knl = (4*(n+w)*(n+w)-1)/(8*Alpha*Alpha);
                double resm=0;
                for(int m2=0; m2<=l2; m2++)
                    resm += Knl * SHcoefs[n][l2][m2] * legendre_array[m2][(l2-m2)*2] * cos(m2*2*phi) * (1+(m2>0));
                resn += resm * gegenpoly_array[n];
            }
        res += resn * pow(r, l-2+1/Alpha) * pow(1+ralpha, -(2*l+1)*Alpha-2) / M_SQRTPI;
    }
    return std::max<double>(res,0);
}

void CPotentialBSE::DiffEq(unsigned n, double t, double *v, double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];

    double dFdR=0, dFdTheta=0, dFdPhi=0;   
#ifdef LYAPUNOVVAREQ
    double d2FdR2=0, d2FdRdTheta=0, d2FdRdPhi=0, d2FdTheta2=0, d2FdThetadPhi=0, d2FdPhi2=0;
#endif
    double r2 = sqr(X) + sqr(Y) + sqr(Z);
    double r  = sqrt(r2);
    double rxy2=sqr(X)+sqr(Y);
    double rxy =sqrt(rxy2);
    double theta=atan2(rxy, Z); 
    double phi  =atan2(Y, X);
    double costheta=(Z==0)?0:cos(theta);
    double sintheta=sin(theta);
    double cosphi=(X==0)?0:cos(phi);
    double sinphi=(Y==0)?0:sin(phi);
    double ralpha=pow(r, 1/Alpha);
    double xi=(ralpha-1)/(ralpha+1);
    double legendre_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR*2-1];
    double legendre_deriv_array[MAX_NCOEFS_ANGULAR][MAX_NCOEFS_ANGULAR*2-1];
    double cosmphi[MAX_NCOEFS_ANGULAR], sinmphi[MAX_NCOEFS_ANGULAR];
    double gegenpoly_array[MAX_NCOEFS_RADIAL];
    for(int m2=0; m2<Ncoefs_angular; m2++)
    {
        gsl_sf_legendre_sphPlm_deriv_array(Ncoefs_angular*2-2, m2*2, costheta, legendre_array[m2], legendre_deriv_array[m2]);
        cosmphi[m2] = cos(m2*2*phi) * (1+(m2>0));
        sinmphi[m2] = sin(m2*2*phi) * 2;
    }
    for(int l2=0; l2<Ncoefs_angular; l2++)
    {
        int l=2*l2;
        double w=(2*l+1)*Alpha+0.5;
        gsl_sf_gegenpoly_array(Ncoefs_radial-1, w, xi, gegenpoly_array);
        double multr = pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha) * 2*M_SQRTPI;
        double multdr= (l-(l+1)*ralpha)/((ralpha+1)*r);
        if(multdr!=multdr) multdr=0;  // safety measure to avoid NaN
        for(int n=0; n<Ncoefs_radial; n++)
        {
            {
                double dGdR = (n>0 ? (-n*xi*gegenpoly_array[n] + (n+2*w-1)*gegenpoly_array[n-1])/(2*Alpha*r) : 0);
                double multdFdR = (multdr * gegenpoly_array[n] + dGdR);
#ifdef LYAPUNOVVAREQ
                multd2FdR2 = ((l+1)*(l+2)*sqr(ralpha) + ((1-2*l*(l+1)) - (2*n+1)*(2*l+1)/Alpha - n*(n+1)/sqr(Alpha))*ralpha + l*(l-1))/sqr((1+ralpha)*r)*gegenpoly_array[n] - dGdR*2/r;
#endif
                for(int m2=0; m2<=l2; m2++)
                {
                    int k=(l2-m2)*2;
                    double coef = SHcoefs[n][l2][m2] * multr;
                    dFdR += coef * legendre_array[m2][k] * cosmphi[m2] * multdFdR;
                    dFdTheta += coef * legendre_deriv_array[m2][k] * (-sintheta) * cosmphi[m2] * gegenpoly_array[n];
                    dFdPhi += coef * legendre_array[m2][k] * (-2*m2)*sinmphi[m2] * gegenpoly_array[n];
#ifdef LYAPUNOVVAREQ
                    if(calcLyapunov)
                    {
                        d2FdR2 += coef * legendre_array[m2][k] * cosmphi[m2] * multd2FdR2;
                        d2FdRdTheta += coef * legendre_deriv_array[m2][k] * (-sintheta) * cosmphi[m2] * multdFdR;
                        d2FdRdPhi += coef * legendre_array[m2][k] * (-2*m2)*sinmphi[m2] * multdFdR;
                        d2FdTheta2 += coef * (costheta * legendre_deriv_array[m2][k] - (l*(l+1)-sqr(2*m2/sintheta)) * legendre_array[m2][k]) * cosmphi[m2] * gegenpoly_array[n];
                        d2FdThetadPhi+= coef * legendre_deriv_array[m2][k] * (-sintheta) * (-2*m2)*sinmphi[m2] * gegenpoly_array[n];
                        d2FdPhi2     += coef * legendre_array[m2][k] * cosmphi[m2] * -sqr(2*m2) * gegenpoly_array[n];
                    }
#endif
                }
            }
        }
    }
    if((X==0)||(Y==0)) dFdPhi=0; // to avoid singularity
    if(Z==0) dFdTheta=0;
    if(rxy==0) rxy=BH_SMOOTH; 
    double r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
    double J[3][3]   = { {sintheta*cosphi, sintheta*sinphi, costheta}, {costheta*cosphi/r, costheta*sinphi/r, -sintheta/r}, {-sinphi/rxy, cosphi/rxy, 0} };
    f[N_dim]   = ( dFdR * J[0][0] + dFdTheta * J[1][0] + dFdPhi * J[2][0]) - ((Mbh>0)?(Mbh*X/r3):0) ;
    f[N_dim+1] = ( dFdR * J[0][1] + dFdTheta * J[1][1] + dFdPhi * J[2][1]) - ((Mbh>0)?(Mbh*Y/r3):0) ;
    if((N_dim==3))
        f[N_dim+2] = ( dFdR * J[0][2] + dFdTheta * J[1][2]) - ((Mbh>0)?(Mbh*Z/r3):0);
#ifdef LYAPUNOVVAREQ
    if(calcLyapunov)
    {
        double X2=X*X, Y2=Y*Y, Z2=Z*Z;
        double d2F[3][3] = { { d2FdR2, d2FdRdTheta, d2FdRdPhi }, {d2FdRdTheta, d2FdTheta2, d2FdThetadPhi}, {d2FdRdPhi, d2FdThetadPhi, d2FdPhi2} };
        double m=1/(r*r2);
        double d2R[3][3] = { {(Y2+Z2)*m, -X*Y*m, -X*Z*m}, {-X*Y*m, (X2+Z2)*m, -Y*Z*m}, {-X*Z*m, -Y*Z*m, (X2+Y2)*m} };
        m=1/(rxy*rxy2*r2*r2);
        double d2Theta[3][3] = { 
            {-Z*(2*X2*X2+Y2*X2-Y2*Y2-Z2*Y2)*m, -X*Y*Z*(3*X2+3*Y2+Z2)*m, X*(X2+Y2-Z2)*rxy2*m},
            {-X*Y*Z*(3*X2+3*Y2+Z2)*m, -Z*(2*Y2*Y2+Y2*X2-X2*X2-Z2*X2)*m, Y*(X2+Y2-Z2)*rxy2*m},
            {X*(X2+Y2-Z2)*rxy2*m, Y*(X2+Y2-Z2)*rxy2*m, 2*Z*rxy2*rxy2*m} };
        m=1/(rxy2*rxy2);
        double d2Phi[3][3] = { {2*X*Y*m, (Y2-X2)*m, 0}, {(Y2-X2)*m, -2*X*Y*m, 0}, {0, 0, 0} };
        // convert to Descartes coordinates
        double r5=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 2.5);
        double k0 = (J[0][0]*d2F[0][0]+J[1][0]*d2F[0][1]+J[2][0]*d2F[0][2]);
        double k1 = (J[0][0]*d2F[0][1]+J[1][0]*d2F[1][1]+J[2][0]*d2F[1][2]);
        double k2 = (J[0][0]*d2F[0][2]+J[1][0]*d2F[1][2]+J[2][0]*d2F[2][2]);
        double d2Fdx2 = (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[0][0] + dFdTheta*d2Theta[0][0] + dFdPhi*d2Phi[0][0]) +Mbh*(3*X2/r5-1/r3);
        double d2Fdxdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[0][1] + dFdTheta*d2Theta[0][1] + dFdPhi*d2Phi[0][1]) +Mbh*(3*X*Y/r5);
        double d2Fdxdz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[0][2] + dFdTheta*d2Theta[0][2] + dFdPhi*d2Phi[0][2]) +Mbh*(3*X*Z/r5);
               k0 = (J[0][1]*d2F[0][0]+J[1][1]*d2F[0][1]+J[2][1]*d2F[0][2]);
               k1 = (J[0][1]*d2F[0][1]+J[1][1]*d2F[1][1]+J[2][1]*d2F[1][2]);
               k2 = (J[0][1]*d2F[0][2]+J[1][1]*d2F[1][2]+J[2][1]*d2F[2][2]);
        double d2Fdydx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[1][0] + dFdTheta*d2Theta[1][0] + dFdPhi*d2Phi[1][0]) +Mbh*(3*Y*X/r5);
        double d2Fdy2 = (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[1][1] + dFdTheta*d2Theta[1][1] + dFdPhi*d2Phi[1][1]) +Mbh*(3*Y2/r5-1/r3);
        double d2Fdydz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[1][2] + dFdTheta*d2Theta[1][2] + dFdPhi*d2Phi[1][2]) +Mbh*(3*Y*Z/r5);
               k0 = (J[0][2]*d2F[0][0]+J[1][2]*d2F[0][1]+J[2][2]*d2F[0][2]);
               k1 = (J[0][2]*d2F[0][1]+J[1][2]*d2F[1][1]+J[2][2]*d2F[1][2]);
               k2 = (J[0][2]*d2F[0][2]+J[1][2]*d2F[1][2]+J[2][2]*d2F[2][2]);
        double d2Fdzdx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[2][0] + dFdTheta*d2Theta[2][0] + dFdPhi*d2Phi[2][0]) +Mbh*(3*Z*X/r5);
        double d2Fdzdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[2][1] + dFdTheta*d2Theta[2][1] + dFdPhi*d2Phi[2][1]) +Mbh*(3*Z*Y/r5);
        double d2Fdz2 = (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[2][2] + dFdTheta*d2Theta[2][2] + dFdPhi*d2Phi[2][2]) +Mbh*(3*Z2/r5-1/r3);

        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        f[N_dim*3]   = (wx*d2Fdx2  + wy*d2Fdxdy + wz*d2Fdxdz);
        f[N_dim*3+1] = (wx*d2Fdydx + wy*d2Fdy2  + wz*d2Fdydz);
        if(N_dim==3)
            f[N_dim*3+2] = (wx*d2Fdzdx + wy*d2Fdzdy + wz*d2Fdz2);
    }
#else
    if(calcLyapunov)
    {
        // integrate nearby orbit
        X=v[N_dim*2];
        Y=v[N_dim*2+1];
        Z=(N_dim==3)?v[N_dim*2+2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2] = v[N_dim*3+2];
        r2 = sqr(X) + sqr(Y) + sqr(Z);
        r  = sqrt(r2);
        rxy2=sqr(X)+sqr(Y);
        rxy =sqrt(rxy2);
        theta=atan2(rxy, Z); 
        phi  =atan2(Y, X);
        costheta=(Z==0)?0:cos(theta);
        sintheta=sin(theta);
        cosphi=(X==0)?0:cos(phi);
        sinphi=(Y==0)?0:sin(phi);
        ralpha=pow(r, 1/Alpha);
        xi=(ralpha-1)/(ralpha+1);
        dFdR=0, dFdTheta=0, dFdPhi=0;   

        for(int m2=0; m2<Ncoefs_angular; m2++)
        {
            gsl_sf_legendre_sphPlm_deriv_array(Ncoefs_angular*2-2, m2*2, costheta, legendre_array[m2], legendre_deriv_array[m2]);
            cosmphi[m2] = cos(m2*2*phi) * (1+(m2>0));
            sinmphi[m2] = sin(m2*2*phi) * 2;
        }
        for(int l2=0; l2<Ncoefs_angular; l2++)
        {
            int l=2*l2;
            double w=(2*l+1)*Alpha+0.5;
            gsl_sf_gegenpoly_array(Ncoefs_radial-1, w, xi, gegenpoly_array);
            double multr = pow(r, l) * pow(1+ralpha, -(2*l+1)*Alpha) * 2*M_SQRTPI;
            double multdr= (l-(l+1)*ralpha)/((ralpha+1)*r);
            if(multdr!=multdr) multdr=0;  // safety measure to avoid NaN
            for(int n=0; n<Ncoefs_radial; n++)
            {
                {
                    double dGdR = (n>0 ? (-n*xi*gegenpoly_array[n] + (n+2*w-1)*gegenpoly_array[n-1])/(2*Alpha*r) : 0);
                    double multdFdR = (multdr * gegenpoly_array[n] + dGdR);
                    for(int m2=0; m2<=l2; m2++)
                    {
                        int k=(l2-m2)*2;
                        double coef = SHcoefs[n][l2][m2] * multr;
                        dFdR += coef * legendre_array[m2][k] * cosmphi[m2] * multdFdR;
                        dFdTheta += coef * legendre_deriv_array[m2][k] * (-sintheta) * cosmphi[m2] * gegenpoly_array[n];
                        dFdPhi += coef * legendre_array[m2][k] * (-2*m2)*sinmphi[m2] * gegenpoly_array[n];
                    }
                }
            }
        }
        if((X==0)||(Y==0)) dFdPhi=0; // to avoid singularity
        if(Z==0) dFdTheta=0;
        if(rxy==0) rxy=BH_SMOOTH; 
        double r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
        double J[3][3]   = { {sintheta*cosphi, sintheta*sinphi, costheta}, {costheta*cosphi/r, costheta*sinphi/r, -sintheta/r}, {-sinphi/rxy, cosphi/rxy, 0} };
        f[N_dim*3]   = ( dFdR * J[0][0] + dFdTheta * J[1][0] + dFdPhi * J[2][0]) - ((Mbh>0)?(Mbh*X/r3):0) ;
        f[N_dim*3+1] = ( dFdR * J[0][1] + dFdTheta * J[1][1] + dFdPhi * J[2][1]) - ((Mbh>0)?(Mbh*Y/r3):0) ;
        if((N_dim==3))
            f[N_dim*3+2] = ( dFdR * J[0][2] + dFdTheta * J[1][2]) - ((Mbh>0)?(Mbh*Z/r3):0);

    }
#endif
}

// Dehnen potential by SH expansion (calculate SH coefficients from explicit expression for density)
CPotentialDehnenBSE::CPotentialDehnenBSE(int _N_dim, double _q, double _p, double _Mbh, double _Gamma, int _Ncoefs_radial, int _Ncoefs_angular): 
    CPotentialBSE(_N_dim, _q, _p, _Mbh, _Gamma, 
    (_Gamma==0? 1.0 :   // best-suited values for alpha
    (_Gamma<0.5? 2.0/3 :
    (_Gamma<1.75? 1/(2-_Gamma) : 4.0))),
    _Ncoefs_radial, _Ncoefs_angular)
{
    prepareCoefs();
}

CPotential* CPotentialDehnenBSE::clone() const { return new CPotentialDehnenBSE(*this); }
CPotential::POTENTIALTYPE CPotentialDehnenBSE::PotentialType() const { return PT_DEHNENBSE; }
char* CPotentialDehnenBSE::myName() { return "Dehnen BSE"; }
char* CPotentialDehnenBSE::PotentialName() const { return myName(); }

double CPotentialDehnenBSE::_Rho(double X, double Y, double Z, double t) const
{   // original density
    double m = sqrt(sqr(X) + sqr(Y/q) + sqr(Z/p));
    return (3-Gamma)/(4*M_PI*p*q) * pow(m, -Gamma) * pow(1+m, Gamma-4);
}

#ifdef RHO_DEHNEN_EXACT
double CPotentialDehnenBSE::Rho(double X, double Y, double Z, double t) const
{
    return _Rho(X, Y, Z, t);
}
#endif

struct CPotentialParamDSH{
    const CPotentialDehnenBSE* P;
    int sh_n, sh_l, sh_m;
    double costheta, sintheta, r, Alpha;
};
double intDlm_phi(double phi, void* params)
{
    double costheta= ((CPotentialParamDSH*)params)->costheta;
    double sintheta= ((CPotentialParamDSH*)params)->sintheta;
    double r= ((CPotentialParamDSH*)params)->r;
    const CPotentialDehnenBSE* P=((CPotentialParamDSH*)params)->P;
    int sh_l=((CPotentialParamDSH*)params)->sh_l;
    int sh_m=((CPotentialParamDSH*)params)->sh_m;
    return P->_Rho(r*sintheta*cos(phi), r*sintheta*sin(phi), r*costheta)
        * gsl_sf_legendre_sphPlm(sh_l, sh_m, costheta) * cos(sh_m*phi);
}
double intDlm_theta(double theta, void* params)
{
    double result, error;
    size_t neval;
    ((CPotentialParamDSH*)params)->costheta=cos(theta);
    ((CPotentialParamDSH*)params)->sintheta=sin(theta);
    gsl_function F;
    F.function=&intDlm_phi;
    F.params=params;
    gsl_integration_qng(&F, 0, M_PI_2, 1e-10/((CPotentialParamDSH*)params)->r, 0, &result, &error, &neval);
    return sin(theta)*result;
}
double intDlm_xi(double xi, void* params)
{
    if(xi>=1.0 || xi<=-1.0) return 0;
    double Alpha=((CPotentialParamDSH*)params)->Alpha;
    double r=pow((1+xi)/(1-xi), Alpha);
    int n=((CPotentialParamDSH*)params)->sh_n;
    int l=((CPotentialParamDSH*)params)->sh_l;
    double gp=gsl_sf_gegenpoly_n(n, (2*l+1)*Alpha+0.5, xi);
    ((CPotentialParamDSH*)params)->r=r;
    gsl_function F;
    F.function=&intDlm_theta;
    F.params=params;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, M_PI_2, 1e-10/((CPotentialParamDSH*)params)->r, 0, &result, &error, &neval);
    return result * gp * pow(r, 3-1/Alpha+l) * pow(1+pow(r,1/Alpha), 2-(2*l+1)*Alpha) * Alpha * 4 * 2*M_SQRTPI;
}

void CPotentialDehnenBSE::prepareCoefs()
{
    CPotentialParamDSH PP;
    PP.P=this;
    PP.Alpha=Alpha;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    double interval[2]={-1.0, 1.0};
    gsl_function F;
    F.function=&intDlm_xi;
    F.params=&PP;
    for(int n=0; n<Ncoefs_radial; n++)
        for(int l2=0; l2<Ncoefs_angular; l2++)
        {
            int l=l2*2;
            PP.sh_n=n;  PP.sh_l=l;
            double Knl = (4*sqr(n+(2*l+1)*Alpha+0.5)-1)/8/sqr(Alpha);
            double Inl = 4*M_PI*Alpha * pow(2.0, -4*(2*l+1)*Alpha-2) * gsl_sf_gamma(n+2*(2*l+1)*Alpha+1) / gsl_sf_fact(n) / (n+(2*l+1)*Alpha+0.5) / gsl_pow_2(gsl_sf_gamma((2*l+1)*Alpha+0.5));
            for(int m2=0; m2<=l2; m2++)
            {
                PP.sh_m=m2*2;
                double result, error;
                size_t neval;
                gsl_integration_qagp(&F, interval, 2, 1e-10,0, 1000, w, &result, &error);
                SHcoefs[n][l2][m2] = result / Inl/Knl;
            }
        }
    gsl_integration_workspace_free (w);
}

// Plummer potential by SH expansion (calculate SH coefficients from explicit expression for density)
CPotentialPlummerBSE::CPotentialPlummerBSE(int _N_dim, double _q, double _p, double _Mbh, int _Ncoefs_radial, int _Ncoefs_angular): 
    CPotentialBSE(_N_dim, _q, _p, _Mbh, 0, 0.5, _Ncoefs_radial, _Ncoefs_angular)
{
    prepareCoefs();
}

CPotential* CPotentialPlummerBSE::clone() const { return new CPotentialPlummerBSE(*this); }
CPotential::POTENTIALTYPE CPotentialPlummerBSE::PotentialType() const { return PT_PLUMMERBSE; }
char* CPotentialPlummerBSE::myName() { return "Plummer BSE"; }
char* CPotentialPlummerBSE::PotentialName() const { return myName(); }

double CPotentialPlummerBSE::_Rho(double X, double Y, double Z, double t) const
{   // original density
    double m = sqrt(sqr(X) + sqr(Y/q) + sqr(Z/p));
    return 3/(4*M_PI*p*q) * pow(1+m*m, -2.5);
}

#ifdef RHO_DEHNEN_EXACT
double CPotentialPlummerBSE::Rho(double X, double Y, double Z, double t) const
{
    return _Rho(X, Y, Z, t);
}
#endif

struct CPotentialParamPSH{
    const CPotentialPlummerBSE* P;
    int sh_n, sh_l, sh_m;
    double costheta, sintheta, r;
};
double intPlm_phi(double phi, void* params)
{
    double costheta= ((CPotentialParamPSH*)params)->costheta;
    double sintheta= ((CPotentialParamPSH*)params)->sintheta;
    double r= ((CPotentialParamPSH*)params)->r;
    const CPotentialPlummerBSE* P=((CPotentialParamPSH*)params)->P;
    int sh_l=((CPotentialParamPSH*)params)->sh_l;
    int sh_m=((CPotentialParamPSH*)params)->sh_m;
    return P->_Rho(r*sintheta*cos(phi), r*sintheta*sin(phi), r*costheta)
        * gsl_sf_legendre_sphPlm(sh_l, sh_m, costheta) * cos(sh_m*phi);
}
double intPlm_theta(double theta, void* params)
{
    double result, error;
    size_t neval;
    ((CPotentialParamPSH*)params)->costheta=cos(theta);
    ((CPotentialParamPSH*)params)->sintheta=sin(theta);
    gsl_function F;
    F.function=&intPlm_phi;
    F.params=params;
    gsl_integration_qng(&F, 0, M_PI_2, 1e-10/((CPotentialParamPSH*)params)->r, 0, &result, &error, &neval);
    return sin(theta)*result;
}
double intPlm_xi(double xi, void* params)
{
    if(xi>=1.0 || xi<=-1.0) return 0;
    double r=sqrt((1+xi)/(1-xi));
    int n=((CPotentialParamPSH*)params)->sh_n;
    int l=((CPotentialParamPSH*)params)->sh_l;
    double gp=gsl_sf_gegenpoly_n(n, l+1, xi);
    ((CPotentialParamPSH*)params)->r=r;
    gsl_function F;
    F.function=&intPlm_theta;
    F.params=params;
    double result, error;
    size_t neval;
    gsl_integration_qng(&F, 0, M_PI_2, 1e-10/((CPotentialParamPSH*)params)->r, 0, &result, &error, &neval);
    return result * gp * pow(r, 1+l) * pow(1+r*r, 1.5-l) * 2 * 2*M_SQRTPI;
}

void CPotentialPlummerBSE::prepareCoefs()
{
    CPotentialParamPSH PP;
    PP.P=this;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    double interval[2]={-1.0, 1.0};
    gsl_function F;
    F.function=&intPlm_xi;
    F.params=&PP;
    for(int n=0; n<Ncoefs_radial; n++)
        for(int l2=0; l2<Ncoefs_angular; l2++)
        {
            int l=l2*2;
            PP.sh_n=n;  PP.sh_l=l;
            double Knl = 2*n*(n+2*l+2) + (2*l+1)*(2*l+3)/2.0;
            double Inl = 2*M_PI * pow(2.0, -4*l-4) * gsl_sf_gamma(n+2*l+2) / gsl_sf_fact(n) / (n+l+1) / gsl_pow_2(gsl_sf_gamma(l+1));
            for(int m2=0; m2<=l2; m2++)
            {
                PP.sh_m=m2*2;
                double result, error;
                size_t neval;
                gsl_integration_qagp(&F, interval, 2, 1e-10,0, 1000, w, &result, &error);
                SHcoefs[n][l2][m2] = result / Inl/Knl;
            }
        }
}

// Scale-free potential 
CPotentialScaleFree::CPotentialScaleFree(int _N_dim, double _q, double _p, double _Mbh, double _Gamma): CPotential(_N_dim, _q, _p, _Mbh), Gamma(_Gamma) {}

CPotential* CPotentialScaleFree::clone() const { return new CPotentialScaleFree(*this); }
CPotential::POTENTIALTYPE CPotentialScaleFree::PotentialType() const { return PT_SCALEFREE; }
char* CPotentialScaleFree::myName() { return "Scale-free"; }
char* CPotentialScaleFree::PotentialName() const { return myName(); }

double CPotentialScaleFree::Rho(double X, double Y, double Z, double t) const
{
    return pow(sqrt(sqr(X) + sqr(Y/q) + sqr(Z/p)), -Gamma);
}

struct CPotentialParamSF2{
    const CPotentialScaleFree* P;
    double theta, phi;
};
struct CPotentialParamSF3{
    const CPotentialScaleFree* P;
    double X, Y, Z;
};

double intPhiSF_angular(double s, void* params)
{
    double theta= ((CPotentialParamSF2*)params)->theta;
    double phi  = ((CPotentialParamSF2*)params)->phi;
    const CPotentialScaleFree* P=((CPotentialParamSF2*)params)->P;
    return 2*pow(sqr(s)*( sqr(sin(theta))*(sqr(cos(phi)) + sqr(sin(phi))/(1-(1-sqr(P->q))*sqr(s))) + sqr(cos(theta))/(1-(1-sqr(P->p))*sqr(s)) ), 1-P->Gamma/2) /
        sqrt( (1-(1-sqr(P->q))*sqr(s)) * (1-(1-sqr(P->p))*sqr(s)));
}

double intForceSFx(double s, void* params)
{
    double X= ((CPotentialParamSF3*)params)->X;
    double Y= ((CPotentialParamSF3*)params)->Y;
    double Z= ((CPotentialParamSF3*)params)->Z;
    const CPotentialScaleFree* P=((CPotentialParamSF3*)params)->P;
    return pow(s, 2-P->Gamma) * pow( sqr(X) + sqr(Y)/(1-(1-sqr(P->q))*sqr(s)) + sqr(Z)/(1-(1-sqr(P->p))*sqr(s)), -P->Gamma/2) /
        sqrt( (1-(1-sqr(P->q))*sqr(s)) * (1-(1-sqr(P->p))*sqr(s)));
}
double intForceSFy(double s, void* params)
{
    double X= ((CPotentialParamSF3*)params)->X;
    double Y= ((CPotentialParamSF3*)params)->Y;
    double Z= ((CPotentialParamSF3*)params)->Z;
    const CPotentialScaleFree* P=((CPotentialParamSF3*)params)->P;
    return pow(s, 2-P->Gamma) * pow( sqr(X) + sqr(Y)/(1-(1-sqr(P->q))*sqr(s)) + sqr(Z)/(1-(1-sqr(P->p))*sqr(s)), -P->Gamma/2) /
        sqrt( (1-(1-sqr(P->q))*sqr(s)) * (1-(1-sqr(P->p))*sqr(s))) / (1-(1-sqr(P->q))*sqr(s));
}
double intForceSFz(double s, void* params)
{
    double X= ((CPotentialParamSF3*)params)->X;
    double Y= ((CPotentialParamSF3*)params)->Y;
    double Z= ((CPotentialParamSF3*)params)->Z;
    const CPotentialScaleFree* P=((CPotentialParamSF3*)params)->P;
    return pow(s, 2-P->Gamma) * pow( sqr(X) + sqr(Y)/(1-(1-sqr(P->q))*sqr(s)) + sqr(Z)/(1-(1-sqr(P->p))*sqr(s)), -P->Gamma/2) /
        sqrt( (1-(1-sqr(P->q))*sqr(s)) * (1-(1-sqr(P->p))*sqr(s))) / (1-(1-sqr(P->p))*sqr(s));
}

double CPotentialScaleFree::Phi(double X, double Y, double Z, double t) const
{
    double result, error;
    size_t neval;
    CPotentialParamSF2 PP;
    PP.theta=atan2(sqrt(sqr(X)+sqr(Y)), Z); 
    PP.phi=atan2(Y, X);
    PP.P=this;
    gsl_function F;
    F.function=&intPhiSF_angular;
    F.params=&PP;
    gsl_integration_qng (&F, 0, 1, 0, 1e-5, &result, &error, &neval); 
    return result * 2*M_PI/(2-Gamma)*q*p * pow(sqr(X)+sqr(Y)+sqr(Z),1-Gamma/2) - ((Mbh>0)?(Mbh/sqrt(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z))):0);
}

void CPotentialScaleFree::DiffEq(unsigned n, double t, double *v, double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double mult=-4*M_PI*q*p;
    double r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
    double result, error; 
    size_t neval;
    CPotentialParamSF3 PP;
    PP.X=X; PP.Y=Y; PP.Z=Z;
    PP.P=this;
    gsl_function F;
    F.params=&PP;
    F.function=&intForceSFx;
    gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
    f[N_dim]   = mult * X * result - ((Mbh>0)?(Mbh*X/r3):0);
    F.function=&intForceSFy;
    gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
    f[N_dim+1] = mult * Y * result - ((Mbh>0)?(Mbh*Y/r3):0);
    if((N_dim==3))
    {
        F.function=&intForceSFz;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
        f[N_dim+2] = mult * Z * result - ((Mbh>0)?(Mbh*Z/r3):0);
    }
    if(calcLyapunov)
    {
#ifdef LYAPUNOVVAREQ
        // variational equation - not implemented in scale-free model
        f[N_dim*2]   = 0;
        f[N_dim*2+1] = 0;
        if(N_dim==3) f[N_dim*2+2]=0;
        f[N_dim*3]   = 0;
        f[N_dim*3+1] = 0;
        if(N_dim==3) f[N_dim*3+2]=0;
#else
        // integrate nearby orbit
        X=v[N_dim*2];
        Y=v[N_dim*2+1];
        Z=(N_dim==3)?v[N_dim*2+2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2] = v[N_dim*3+2];
        r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
        PP.X=X; PP.Y=Y; PP.Z=Z;
        F.function=&intForceSFx;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
        f[N_dim*3]   = mult * X * result - ((Mbh>0)?(Mbh*X/r3):0);
        F.function=&intForceSFy;
        gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
        f[N_dim*3+1] = mult * Y * result - ((Mbh>0)?(Mbh*Y/r3):0);
        if((N_dim==3))
        {
            F.function=&intForceSFz;
            gsl_integration_qng (&F, 0, 1, 0, 1e-7, &result, &error, &neval); 
            f[N_dim*3+2] = mult * Z * result - ((Mbh>0)?(Mbh*Z/r3):0);
        }
#endif
    }
}

// scale-free potential - spherical harmonic expansion
CPotentialScaleFreeBSE::CPotentialScaleFreeBSE(int _N_dim, double _q, double _p, double _Mbh, double _Gamma, int _Ncoefs_SF_angular): CPotentialScaleFree(_N_dim, _q, _p, _Mbh, _Gamma), Ncoefs_SF_angular(std::min(_Ncoefs_SF_angular, MAX_NCOEFS_SF_ANGULAR))
{
    prepareCoefs();
}

CPotential* CPotentialScaleFreeBSE::clone() const { return new CPotentialScaleFreeBSE(*this); }
CPotential::POTENTIALTYPE CPotentialScaleFreeBSE::PotentialType() const { return PT_SCALEFREEBSE; }
char* CPotentialScaleFreeBSE::myName() { return "Scale-free SH"; }
char* CPotentialScaleFreeBSE::PotentialName() const { return myName(); }

struct CPotentialParamSFSH{
    const CPotentialScaleFreeBSE* P;
    int sh_l, sh_m;
    double costheta;
};
double intSFlm_phi(double phi, void* params)
{
    double costheta= ((CPotentialParamSFSH*)params)->costheta;
    const CPotentialScaleFreeBSE* P=((CPotentialParamSFSH*)params)->P;
    int sh_l=((CPotentialParamSFSH*)params)->sh_l;
    int sh_m=((CPotentialParamSFSH*)params)->sh_m;
    return pow((1-sqr(costheta))*(1+(1/sqr(P->q)-1)*gsl_pow_2(sin(phi))) + gsl_pow_2(costheta/P->p), -P->Gamma/2)
        * gsl_sf_legendre_sphPlm(sh_l*2, sh_m*2, costheta) * cos(sh_m*2*phi);
}
double intSFlm_theta(double theta, void* params)
{
    double result, error;
    size_t neval;
    ((CPotentialParamSFSH*)params)->costheta=cos(theta);
    gsl_function F;
    F.function=&intSFlm_phi;
    F.params=params;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    gsl_integration_qag(&F, 0, M_PI_2, 1e-12, 0, 1000, 3, w, &result, &error);
    gsl_integration_workspace_free (w);
    return sin(theta)*result;
}
void CPotentialScaleFreeBSE::prepareCoefs()
{
    CPotentialParamSFSH PP;
    PP.P=this;
    if(Gamma<1e-5) Gamma=1e-5;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
    for(PP.sh_l=0; PP.sh_l<Ncoefs_SF_angular; PP.sh_l++)
        for(PP.sh_m=0; PP.sh_m<=PP.sh_l; PP.sh_m++)
        {
            double result, error;
            size_t neval;
            gsl_function F;
            F.function=&intSFlm_theta;
            F.params=&PP;
            gsl_integration_qag(&F, 0, M_PI_2, 1e-12, 0, 1000, 3, w, &result, &error);
            int l=2*PP.sh_l;
            SHcoefs[PP.sh_l][PP.sh_m]= 8*result * (1.0/(l+3-Gamma) + 1.0/(l-2+Gamma)) * 1.0/(2*l+1) * 4*M_PI;
        }
    gsl_integration_workspace_free (w);
    if(Gamma==1e-5) Gamma=0;
    if(q==1) SHcoefs[1][1]=0;
}

double CPotentialScaleFreeBSE::Phi(double X, double Y, double Z, double t) const
{
    double costheta=(X==0&&Y==0&&Z==0)?0:(Z/sqrt(sqr(X)+sqr(Y)+sqr(Z)));
    double phi  =atan2(Y, X);
    double F=0;
    for(int m2=0; m2<Ncoefs_SF_angular; m2++)
    {
        double result_array[MAX_NCOEFS_SF_ANGULAR*2-1];
        gsl_sf_legendre_sphPlm_array(Ncoefs_SF_angular*2-2, m2*2, costheta, result_array);
        for(int k2=0; k2<Ncoefs_SF_angular-m2; k2++)
        {
            int l=(m2+k2)*2;
            F += SHcoefs[m2+k2][m2] * result_array[k2*2] * cos(m2*2*phi) * (1+(m2>0));
        }
    }
    return -F * pow(sqr(X)+sqr(Y)+sqr(Z),1-Gamma/2) - ((Mbh>0)?(Mbh/sqrt(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z))):0);
}

void CPotentialScaleFreeBSE::DiffEq(unsigned n, double t, double *v, double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);

    double theta=atan2(sqrt(sqr(X)+sqr(Y)), Z); 
    double phi  =atan2(Y, X);
    double r2 = sqr(X) + sqr(Y) + sqr(Z);
    double r  = sqrt(r2);
    double rxy2=sqr(X)+sqr(Y);
    double rxy =sqrt(rxy2);
    double costheta=(Z==0)?0:cos(theta);
    double sintheta=sin(theta);
    double cosphi=(X==0)?0:cos(phi);
    double sinphi=(Y==0)?0:sin(phi);
    double F=0, dFdR=0, dFdTheta=0, dFdPhi=0;  // angular part of potential and its derivatives
#ifdef LYAPUNOVVAREQ
    double d2FdR2=0, d2FdRdTheta=0, d2FdRdPhi=0, d2FdTheta2=0, d2FdThetadPhi=0, d2FdPhi2=0;
#endif
    double cosmphi[MAX_NCOEFS_SF_ANGULAR], sinmphi[MAX_NCOEFS_SF_ANGULAR];
    double result_array[MAX_NCOEFS_SF_ANGULAR*2-1], result_deriv_array[MAX_NCOEFS_SF_ANGULAR*2-1];
    for(int m2=0; m2<Ncoefs_SF_angular; m2++)
    {
        cosmphi[m2] = cos(m2*2*phi) * (1+(m2>0));
        sinmphi[m2] = sin(m2*2*phi) * 2;
    }
    for(int m2=0; m2<Ncoefs_SF_angular; m2++)
    {
        gsl_sf_legendre_sphPlm_deriv_array(Ncoefs_SF_angular*2-2, m2*2, costheta, result_array, result_deriv_array);
        for(int k2=0; k2<Ncoefs_SF_angular-m2; k2++)
        {
            int l=(m2+k2)*2;
            double coef=SHcoefs[m2+k2][m2] * pow(r, 2-Gamma);
            F += coef * result_array[k2*2] * cosmphi[m2];
            dFdTheta += coef * result_deriv_array[k2*2] * (-sintheta) * cosmphi[m2];
            dFdPhi   += coef * result_array[k2*2] * (-2*m2)*sinmphi[m2];
#ifdef LYAPUNOVVAREQ
            if(calcLyapunov)
            {
                d2FdTheta2   += coef * (costheta*result_deriv_array[k2*2] - (l*(l+1)-sqr(2*m2/sintheta))*result_array[k2*2]) * cosmphi[m2];
                d2FdThetadPhi+= coef * result_deriv_array[k2*2] * (-sintheta) * (-2*m2)*sinmphi[m2];
                d2FdPhi2     += coef * result_array[k2*2] * cosmphi[m2] * -sqr(2*m2);
            }
#endif
        }
    }
    dFdR = (2-Gamma)*F/r;
    if((X==0)||(Y==0)) dFdPhi=0;
    if(Z==0) dFdTheta=0;
    double J[3][3]   = { {sintheta*cosphi, sintheta*sinphi, costheta}, {costheta*cosphi/r, costheta*sinphi/r, -sintheta/r}, {rxy==0?0:-sinphi/rxy, rxy==0?0:cosphi/rxy, 0} };
    //double J[3][3]   = { {X/r, Y/r, Z/r}, {X*Z/rxy/r2, Y*Z/rxy/r2, -rxy/r2}, {-Y/rxy2, X/rxy2, 0} };
    f[N_dim]   = ( dFdR * J[0][0] + dFdTheta * J[1][0] + dFdPhi * J[2][0]) - ((Mbh>0)?(Mbh*X/r3):0);
    f[N_dim+1] = ( dFdR * J[0][1] + dFdTheta * J[1][1] + dFdPhi * J[2][1]) - ((Mbh>0)?(Mbh*Y/r3):0);
    if((N_dim==3))
        f[N_dim+2] = ( dFdR * J[0][2] + dFdTheta * J[1][2]) - ((Mbh>0)?(Mbh*Z/r3):0);
#ifdef LYAPUNOVVAREQ
    if(calcLyapunov)
    {
        d2FdR2     =(2-Gamma)*(1-Gamma)*F/r2;
        d2FdRdTheta=(2-Gamma)/r*dFdTheta;
        d2FdRdPhi  =(2-Gamma)/r*dFdPhi;
        double X2=X*X, Y2=Y*Y, Z2=Z*Z;
        double d2F[3][3] = { { d2FdR2, d2FdRdTheta, d2FdRdPhi }, {d2FdRdTheta, d2FdTheta2, d2FdThetadPhi}, {d2FdRdPhi, d2FdThetadPhi, d2FdPhi2} };
        double m=1/(r*r2);
        double d2R[3][3] = { {(Y2+Z2)*m, -X*Y*m, -X*Z*m}, {-X*Y*m, (X2+Z2)*m, -Y*Z*m}, {-X*Z*m, -Y*Z*m, (X2+Y2)*m} };
        m=1/(rxy*rxy2*r2*r2);
        double d2Theta[3][3] = { 
            {-Z*(2*X2*X2+Y2*X2-Y2*Y2-Z2*Y2)*m, -X*Y*Z*(3*X2+3*Y2+Z2)*m, X*(X2+Y2-Z2)*rxy2*m},
            {-X*Y*Z*(3*X2+3*Y2+Z2)*m, -Z*(2*Y2*Y2+Y2*X2-X2*X2-Z2*X2)*m, Y*(X2+Y2-Z2)*rxy2*m},
            {X*(X2+Y2-Z2)*rxy2*m, Y*(X2+Y2-Z2)*rxy2*m, 2*Z*rxy2*rxy2*m} };
        m=1/(rxy2*rxy2);
        double d2Phi[3][3] = { {2*X*Y*m, (Y2-X2)*m, 0}, {(Y2-X2)*m, -2*X*Y*m, 0}, {0, 0, 0} };
        // convert to Descartes coordinates
        double r5=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 2.5);
        double k0 = (J[0][0]*d2F[0][0]+J[1][0]*d2F[0][1]+J[2][0]*d2F[0][2]);
        double k1 = (J[0][0]*d2F[0][1]+J[1][0]*d2F[1][1]+J[2][0]*d2F[1][2]);
        double k2 = (J[0][0]*d2F[0][2]+J[1][0]*d2F[1][2]+J[2][0]*d2F[2][2]);
        double d2Fdx2 = (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[0][0] + dFdTheta*d2Theta[0][0] + dFdPhi*d2Phi[0][0]) +Mbh*(3*X2/r5-1/r3);
        double d2Fdxdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[0][1] + dFdTheta*d2Theta[0][1] + dFdPhi*d2Phi[0][1]) +Mbh*(3*X*Y/r5);
        double d2Fdxdz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[0][2] + dFdTheta*d2Theta[0][2] + dFdPhi*d2Phi[0][2]) +Mbh*(3*X*Z/r5);
               k0 = (J[0][1]*d2F[0][0]+J[1][1]*d2F[0][1]+J[2][1]*d2F[0][2]);
               k1 = (J[0][1]*d2F[0][1]+J[1][1]*d2F[1][1]+J[2][1]*d2F[1][2]);
               k2 = (J[0][1]*d2F[0][2]+J[1][1]*d2F[1][2]+J[2][1]*d2F[2][2]);
        double d2Fdydx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[1][0] + dFdTheta*d2Theta[1][0] + dFdPhi*d2Phi[1][0]) +Mbh*(3*Y*X/r5);
        double d2Fdy2 = (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[1][1] + dFdTheta*d2Theta[1][1] + dFdPhi*d2Phi[1][1]) +Mbh*(3*Y2/r5-1/r3);
        double d2Fdydz= (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[1][2] + dFdTheta*d2Theta[1][2] + dFdPhi*d2Phi[1][2]) +Mbh*(3*Y*Z/r5);
               k0 = (J[0][2]*d2F[0][0]+J[1][2]*d2F[0][1]+J[2][2]*d2F[0][2]);
               k1 = (J[0][2]*d2F[0][1]+J[1][2]*d2F[1][1]+J[2][2]*d2F[1][2]);
               k2 = (J[0][2]*d2F[0][2]+J[1][2]*d2F[1][2]+J[2][2]*d2F[2][2]);
        double d2Fdzdx= (k0*J[0][0] + k1*J[1][0] + k2*J[2][0] + dFdR*d2R[2][0] + dFdTheta*d2Theta[2][0] + dFdPhi*d2Phi[2][0]) +Mbh*(3*Z*X/r5);
        double d2Fdzdy= (k0*J[0][1] + k1*J[1][1] + k2*J[2][1] + dFdR*d2R[2][1] + dFdTheta*d2Theta[2][1] + dFdPhi*d2Phi[2][1]) +Mbh*(3*Z*Y/r5);
        double d2Fdz2 = (k0*J[0][2] + k1*J[1][2] + k2*J[2][2] + dFdR*d2R[2][2] + dFdTheta*d2Theta[2][2] + dFdPhi*d2Phi[2][2]) +Mbh*(3*Z2/r5-1/r3);

        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        f[N_dim*3]   = (wx*d2Fdx2  + wy*d2Fdxdy + wz*d2Fdxdz);
        f[N_dim*3+1] = (wx*d2Fdydx + wy*d2Fdy2  + wz*d2Fdydz);
        if(N_dim==3)
            f[N_dim*3+2] = (wx*d2Fdzdx + wy*d2Fdzdy + wz*d2Fdz2);
    }
#else
    if(calcLyapunov)
    {
        // integrate nearby orbit
        X=v[N_dim*2];
        Y=v[N_dim*2+1];
        Z=(N_dim==3)?v[N_dim*2+2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2] = v[N_dim*3+2];
        r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
        theta=atan2(sqrt(sqr(X)+sqr(Y)), Z); 
        phi  =atan2(Y, X);
        r2 = sqr(X) + sqr(Y) + sqr(Z);
        r  = sqrt(r2);
        rxy2=sqr(X)+sqr(Y);
        rxy =sqrt(rxy2);
        F=0, dFdTheta=0, dFdPhi=0;  // angular part of potential and its derivatives
        costheta=(Z==0)?0:cos(theta);
        sintheta=sin(theta);
        cosphi=(X==0)?0:cos(phi);
        sinphi=(Y==0)?0:sin(phi);
        F=0, dFdR=0, dFdTheta=0, dFdPhi=0;  // angular part of potential and its derivatives
        for(int m2=0; m2<Ncoefs_SF_angular; m2++)
        {
            cosmphi[m2] = cos(m2*2*phi) * (1+(m2>0));
            sinmphi[m2] = sin(m2*2*phi) * 2;
        }

        for(int m2=0; m2<Ncoefs_SF_angular; m2++)
        {
            gsl_sf_legendre_sphPlm_deriv_array(Ncoefs_SF_angular*2-2, m2*2, costheta, result_array, result_deriv_array);
            for(int k2=0; k2<Ncoefs_SF_angular-m2; k2++)
            {
                int l=(m2+k2)*2;
                double coef=SHcoefs[m2+k2][m2] * pow(r, 2-Gamma);
                F += coef * result_array[k2*2] * cosmphi[m2];
                dFdTheta += coef * result_deriv_array[k2*2] * (-sintheta) * cosmphi[m2];
                dFdPhi   += coef * result_array[k2*2] * (-2*m2)*sinmphi[m2];
            }
        }

        dFdR = (2-Gamma)*F/r;
        if((X==0)||(Y==0)) dFdPhi=0;
        if(Z==0) dFdTheta=0;
        double J[3][3]   = { {sintheta*cosphi, sintheta*sinphi, costheta}, {costheta*cosphi/r, costheta*sinphi/r, -sintheta/r}, {rxy==0?0:-sinphi/rxy, rxy==0?0:cosphi/rxy, 0} };
        f[N_dim*3]   = ( dFdR * J[0][0] + dFdTheta * J[1][0] + dFdPhi * J[2][0]) - ((Mbh>0)?(Mbh*X/r3):0);
        f[N_dim*3+1] = ( dFdR * J[0][1] + dFdTheta * J[1][1] + dFdPhi * J[2][1]) - ((Mbh>0)?(Mbh*Y/r3):0);
        if((N_dim==3))
            f[N_dim*3+2] = ( dFdR * J[0][2] + dFdTheta * J[1][2]) - ((Mbh>0)?(Mbh*Z/r3):0);
    }
#endif
}

CPotentialHarmonic::CPotentialHarmonic(int _N_dim, double _q, double _p, double _Mbh): CPotential(_N_dim, _q, _p, _Mbh) {};
CPotential* CPotentialHarmonic::clone() const { return new CPotentialHarmonic(*this); }
CPotential::POTENTIALTYPE CPotentialHarmonic::PotentialType() const { return PT_HARMONIC; }
char* CPotentialHarmonic::myName() { return "Harmonic"; }
char* CPotentialHarmonic::PotentialName() const { return myName(); }

double CPotentialHarmonic::Rho(double X, double Y, double Z, double t) const
{
    return 2*(1+1/q/q+1/p/p);
}

double CPotentialHarmonic::Phi(double X, double Y, double Z, double t) const
{
    return X*X + Y*Y/q/q + Z*Z/p/p - Mbh/sqrt(sqr(BH_SMOOTH)+ sqr(X) + sqr(Y) + sqr(Z));
}

void CPotentialHarmonic::DiffEq(unsigned n, double t, double *v, double *f) const
{
    bool calcLyapunov=n>2*N_DIM;
    double X=v[0];
    double Y=v[1];
    double Z=(N_dim==3)?v[2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
    f[0] = v[N_dim];
    f[1] = v[N_dim+1];
    if(N_dim==3)
        f[2] = v[N_dim+2];
    double gradr= Mbh/pow( sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);

    f[N_dim]   = -X*(gradr + 2);
    f[N_dim+1] = -Y*(gradr + 2/q/q);
    if((N_dim==3))
        f[N_dim+2] = -Z*(gradr + 2/p/p);
    if(calcLyapunov)
    {
#ifdef LYAPUNOVVAREQ
        // variational equation
        double Mr3=Mbh/pow(sqr(X) + sqr(Y) + sqr(Z), 1.5);
        double Mr5=Mbh/pow(sqr(X) + sqr(Y) + sqr(Z), 2.5);
        double wx = v[N_dim*2];     // variation vector
        double wy = v[N_dim*2+1];
        double wz = (N_dim==3)? v[N_dim*2+2] : 0;
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2]=v[N_dim*3+2];
        f[N_dim*3]   = -wx * (2     - 3*Mr5*sqr(X) + Mr3) + wy*X*Y*3*Mr5 + wz*X*Z*3*Mr5;
        f[N_dim*3+1] = -wy * (2/q/q - 3*Mr5*sqr(Y) + Mr3) + wx*Y*X*3*Mr5 + wz*Y*Z*3*Mr5;
        if(N_dim==3)
            f[N_dim*3+2] = -wz * (2/p/p - 3*Mr5*sqr(Z) + Mr3) - wx*Z*X*3*Mr5 + wy*Z*Y*3*Mr5;
#else
        // integrate nearby orbit
        X=v[N_dim*2];
        Y=v[N_dim*2+1];
        Z=(N_dim==3)?v[N_dim*2+2]:0;    // for universality: if only 2 dimensions, Z=0 will add nothing
        f[N_dim*2]   = v[N_dim*3];
        f[N_dim*2+1] = v[N_dim*3+1];
        if(N_dim==3)
            f[N_dim*2+2] = v[N_dim*3+2];
        gradr= Mbh/pow( sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
        f[N_dim*3]   = -X*(gradr + 2);
        f[N_dim*3+1] = -Y*(gradr + 2/q/q);
        if((N_dim==3))
            f[N_dim*3+2] = -Z*(gradr + 2/p/p);
#endif
    }
}

// Frozen-N-body potential (adapted from Barnes&Hut treecode from 'hackforce')
CPotentialNB::CPotentialNB(int _N_dim, double _q, double _p, double _Mbh, double _Gamma, double _eps, double _tol,
    int npoints, double _X[], double _Y[], double _Z[], double _mass[]):
    CPotential(_N_dim, _q, _p, _Mbh), Gamma(_Gamma), eps(_eps), tol(_tol)
{
    if(N_dim!=3) return;
    bodytab.resize(npoints);
    for(int i=0; i<npoints; i++)
    {
        bodytab[i].type=BODY;
        bodytab[i].mass=_mass[i];
        bodytab[i].pos[0]=_X[i];
        bodytab[i].pos[1]=_Y[i];
        bodytab[i].pos[2]=_Z[i];
    }
    maketree();
}
CPotentialNB::CPotentialNB(int _N_dim, double _q, double _p, double _Mbh, double _Gamma, double _eps, double _tol,
    int npoints, float _X[], float _Y[], float _Z[], float _mass[]):
    CPotential(_N_dim, _q, _p, _Mbh), Gamma(_Gamma), eps(_eps), tol(_tol)
{
    if(N_dim!=3) return;
    bodytab.resize(npoints);
    for(int i=0; i<npoints; i++)
    {
        bodytab[i].type=BODY;
        bodytab[i].mass=_mass[i];
        bodytab[i].pos[0]=_X[i];
        bodytab[i].pos[1]=_Y[i];
        bodytab[i].pos[2]=_Z[i];
    }
    maketree();
}

CPotential* CPotentialNB::clone() const { return new CPotentialNB(*this); }
CPotential::POTENTIALTYPE CPotentialNB::PotentialType() const { return PT_NB; }
char* CPotentialNB::myName() { return "Nbody"; }
char* CPotentialNB::PotentialName() const { return myName(); }

double CPotentialNB::Phi(double X, double Y, double Z, double t) const
{
    double phi=0;
    vec3 point={X,Y,Z};
    nbb=nbc=0;
    walktree(troot, sqr(rsize), point, CALCPHI, &phi);
    return -phi - ((Mbh>0)?(Mbh/sqrt(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z))):0);
}

bool comparepd(paird elem1, paird elem2)
{
    return(elem1.first<elem2.first);
}

double CPotentialNB::Rho(double X, double Y, double Z, double t) const
{   
    vectorpd vec;
    vec3 point={X,Y,Z};
    // first determine which cell does this point belong to, and its size (which will be used as an estimate of neighbour search radius)
    double searchrad=rsize;
    nodeptr *qptr=(nodeptr*)&troot;
    int xp[N_DIM];
    if(!intcoord(xp, point)) return 0;   // if point is outside root cell, density is zero
    int l = IMAX >> 1;				/* start with top bit       */
    while (*qptr != NULL && l>0) {  /* loop descending tree     */
        if ((*qptr)->type == BODY) {/*   reached a "leaf"?      */
            l=0;
	    }
        else{
	        qptr = &(((cellptr)(*qptr))->subp[subindex(xp, l)]);	/*   move down one level    */
	        l = l >> 1;				/*   and test next bit      */
            searchrad/=2;
        }
    }
    searchrad *= 4;  // to have somewhat more than one neighbour!
    // now look for neighbours inside searchrad
    int numneighbour=0;
    do{
        vec.clear();
        walktreedens(troot, rsize*rsize, searchrad*searchrad, point, xp, IMAX >> 1, &vec);
        numneighbour=vec.size();
        searchrad*=2;
    } while(numneighbour<TREECODE_DENSITY_KTH_NEIGHBOUR);
    std::sort(vec.begin(), vec.end(), comparepd);
    int kth=std::min<int>(vec.size(), TREECODE_DENSITY_KTH_NEIGHBOUR);
    if(kth<3) return 0; // failure to get density
    double mass=0;
    double rad=0, smoothrad=vec[kth-1].first;
    for(int k=0; k<vec.size(); k++)
    {
        double radrel=vec[k].first/smoothrad;  // smoothing radius is the distance to Kth neighbour
        double weight=0;  
#ifdef TREECODE_DENSITY_SPLINE
        if(radrel<0.5) weight=8-48*radrel*radrel+48*radrel*radrel*radrel;  // use spline weighting function with compact support
        else if(radrel<1) weight=16*(1-radrel)*(1-radrel)*(1-radrel);
#else
        if(radrel<1) weight=15.0/8*(1-radrel*radrel);  // use simpler Ferrer F1 kernel
#endif
        mass += vec[k].second*weight;
        if(weight==0) k=vec.size(); // all other neighbours are farther than spline support
    }
    mass /= (M_PI*pow(smoothrad, 3.0));  // for spline
    return mass;
}

void CPotentialNB::DiffEq(unsigned n, double t, double *v, double *f) const
{
    double X=v[0];
    double Y=v[1];
    double Z=v[2];
    f[0] = v[3];
    f[1] = v[4];
    f[2] = v[5];
    f[3]=f[4]=f[5]=0;
    f[6]=rsize;
    if(n<2*N_DIM) return;   // do not live in Flatland...
    nbb=nbc=0;
    vec3 point={X,Y,Z};
    f[N_DIM*2];  // init min distance
    walktree(troot, sqr(rsize), point, CALCACC, f+N_DIM);
    if(Mbh>0)
    {
        double r3=pow(sqr(BH_SMOOTH) + sqr(X) + sqr(Y) + sqr(Z), 1.5);
        f[N_dim]   += - ((Mbh>0)?(Mbh*X/r3):0);
        f[N_dim+1] += - ((Mbh>0)?(Mbh*Y/r3):0);
        f[N_dim+2] += - ((Mbh>0)?(Mbh*Z/r3):0);
    }
}

void CPotentialNB::maketree()    // initialize tree from set of bodies
{
    int celltabsize = (int) (100+bodytab.size()*0.52);  // should be enough
    bool result=true; 
    do{
        troot=NULL;
        for(int d=0; d<N_DIM; d++) rmin[d]=-2.0;
        rsize=4.0;  // init box size
        celltab.clear();
        celltab.reserve(celltabsize); 
        result=true; 
        int b=0;
        for(b=0; b<bodytab.size() && result; b++)
        {
	        result = expandbox(b);   /* expand root to fit */
	        result&= appendtree(b);  /* insert into tree   */
	    }
        if(!result)
        {
            if(celltab.size()>=celltab.capacity() && b>bodytab.size()/2)  // error due to insufficient allocation of celltab
                celltabsize = (int) (celltab.capacity() *1.02* bodytab.size()/b); // increase it and repeat
            else
            {
                troot=0;
                celltab.clear();
                return;  // unknown error...
            }
        }
    } while(!result);
    
    centerofmass(troot);				/* find c-of-m coordinates  */
}

bool CPotentialNB::expandbox(const int b)  // expand bounding box if needed to accomodate for a new body
{
    int xtmp[N_DIM];
    while (! intcoord(xtmp, bodytab[b].pos)) {  /* expand box (rarely)      */
       vec3 rmid;
       for(int d=0; d<N_DIM; d++)
            rmid[d]=rmin[d]+0.5*rsize;          /*   find box midpoint      */
        for(int d=0; d<N_DIM; d++)              /*   loop over dimensions   */
            if (bodytab[b].pos[d] < rmid[d])    /*     is p left of mid?    */
                rmin[d] -= rsize;               /*       extend to left     */
        rsize = 2.0 * rsize;                    /*   double length of box   */
        if (troot != NULL) {                    /*   repot existing tree?   */
            cell c;                             /*     create new root cell */
            int xmid[N_DIM];
            if(!intcoord(xmid, rmid)) return false;   /*     locate old root cell */
            int s = subindex(xmid, IMAX >> 1);  /*     find old tree index  */
            c.subp[s] = troot;                  /*     graft old on new     */
            if(celltab.size()>=celltab.capacity()) return false;  /* cannot reallocate vector, otherwise all pointers will be confused */
            celltab.push_back(c);
            troot = (nodeptr) (&(celltab.back()));             /*     plant new tree       */
        }
    }
    return true;
}

bool CPotentialNB::appendtree(const int b)			// append a body to the tree, creating new cell if necessary
{
    int xp[N_DIM], xq[N_DIM];
    nodeptr *qptr;

    intcoord(xp, bodytab[b].pos);   /* form integer coords      */
    int l = IMAX >> 1;				/* start with top bit       */
    qptr = &troot;                  /* start with tree root     */
    while (*qptr != NULL) {         /* loop descending tree     */
        if(l==0) {                  /*   dont run out of bits   */
            (*qptr)->mass += bodytab[b].mass;  /* if two particles coincide, just sum up their mass */
            return true;
        }
        if ((*qptr)->type == BODY) {/*   reached a "leaf"?      */
            cell c;                 /*     alloc a new cell     */
	        if(!intcoord(xq, (*qptr)->pos)) return false;   /*     get integer coords   */
	        c.subp[subindex(xq, l)] = *qptr;	/*     put body in cell     */
            if(celltab.size()>=celltab.capacity()) return false;  /* cannot reallocate vector, otherwise all pointers will be confused */
	        celltab.push_back(c);
	        *qptr = (nodeptr) (&(celltab.back()));		/*     link cell in tree    */
	    }
	    qptr = &(((cellptr)(*qptr))->subp[subindex(xp, l)]);	/*   move down one level    */
	    l = l >> 1;				/*   and test next bit      */
    }
    *qptr = (nodeptr) (&(bodytab[b]));			/* found place, store p     */
    return true;
}

bool CPotentialNB::intcoord(    // compute integer coords, return false if point is out of bounds for current root cell
		    int xp[N_DIM],   /* integerized coordinate vec3 [0,IMAX) */
		    const vec3 rp)   const   /* double coordinate vec3 (system coords) */
{
    bool inb = true;					/* use to check bounds      */
    for (int k = 0; k < N_DIM; k++) {		/* loop over dimensions     */
        double xsc = (rp[k] - rmin[k]) / rsize; /*   scale to range [0,1)   */
        if (0.0 <= xsc && xsc < 1.0)            /*   within unit interval?  */
            xp[k] = (int) floor(IMAX * xsc);    /*     then integerize      */
        else                                    /*   out of range           */
            inb = false;                        /*     then remember that   */
    }
    return inb;
}

int CPotentialNB::subindex(   // determine the subcell that a given position belongs to
		   const int x[N_DIM],	       /* integerized coordinates of particle */
		   const int l)	 const      /* current level of tree */
{
    int i = 0;                                      /* sum index in i           */
    for (int k = 0; k < N_DIM; k++)                  /* check each dimension     */
        if (x[k] & l)                           /*   if beyond midpoint     */
            i += NSUB >> (k + 1);               /*     skip over subcells   */
    return (i);
}

void CPotentialNB::centerofmass(const nodeptr q)      // recursive center-of-mass finder
{
    if(q==NULL) return;
    if (q->type == CELL) {                      /* is this a cell?          */
        q->mass = 0.0;                          /*   init total mass        */
        for(int d=0; d<N_DIM; d++) q->pos[d]=0;	/*   and c. of m.           */
        for(int i=0; i< NSUB; i++) {            /*   loop over subcells     */
            nodeptr r = ((cellptr)q)->subp[i];
            if (r != NULL) {                    /*     does subcell exist?  */
                centerofmass(r);                /*       find subcell cm    */
                q->mass += r->mass;             /*       sum total mass     */
                for(int d=0; d<N_DIM; d++)
                    q->pos[d] += r->mass * r->pos[d];
            }
        }
        if(q->mass>0)
            for(int d=0; d<N_DIM; d++)
                q->pos[d] /= q->mass;
#ifdef TREECODE_QUADRUPOLE
        for(int d1=0; d1<N_DIM; d1++)
            for(int d2=0; d2<N_DIM; d2++)
                ((cellptr)q)->quad[d1][d2]=0;
        for (int i = 0; i < NSUB; i++) {		/*   loop over subnodes     */
            nodeptr r = ((cellptr)q)->subp[i];
            if (r != NULL) {			/*     does subnode exist?  */
                vec3 dr={r->pos[0]-q->pos[0], r->pos[1]-q->pos[1], r->pos[2]-q->pos[2]};  /* displacement vect. */
                double drsq=sqr(dr[0])+sqr(dr[1])+sqr(dr[2]);
                for(int d1=0; d1<N_DIM; d1++)
                    for(int d2=0; d2<N_DIM; d2++)
                    {
                        ((cellptr)q)->quad[d1][d2] += (3*dr[d1]*dr[d2] - (d1==d2?drsq:0)) * r->mass;
                        if(r->type==CELL)    /*       if subnode is cell */
                            ((cellptr)q)->quad[d1][d2] += ((cellptr)r)->quad[d1][d2];  /*         use its moment   */
                    }
            }
        }
#endif
    }
}

void CPotentialNB::walktree(const nodeptr p,   // pointer to interacting node or cell 
		   const double dsq,                   // cell size squared
           const vec3 pos0,                    // position to calculate quantity in question
           const WALKOP operation,             // operation to perform during tree-walk
           double data[])  const               // data accumulated according to the operation
{
    if(p==NULL) return;
    nodeptr *pp;
    vec3 dr={p->pos[0]-pos0[0], p->pos[1]-pos0[1], p->pos[2]-pos0[2]};
    double distsq=sqr(dr[0])+sqr(dr[1])+sqr(dr[2]);
    double distsq0=distsq;
    // test whether cell should be opened
    if(p->type==CELL && (tol*tol*distsq < dsq/* || distsq<sqr(config.nbpotEps)*/)) {
        for(int c=0; c<NSUB; c++)
            if(((cellptr)p)->subp[c] != NULL)  /* walk into sub-cells */
                walktree(((cellptr)p)->subp[c], dsq/4.0, pos0, operation, data);
    }
    else if(distsq>0) 
    {   // perform requested operation on leaf or unopened cell
        distsq += eps*eps;                            /* use Plummer softening   */
        double dist = sqrt(distsq);
        double phii = p->mass / dist;
#ifdef TREECODE_QUADRUPOLE
        vec3 quaddr;
        double drquaddr=0;
        if(p->type == CELL)                        /* if cell, add quad. term  */
            for(int d1=0; d1<N_DIM; d1++)
            {
                quaddr[d1] = ((cellptr)p)->quad[d1][0] * dr[0] + ((cellptr)p)->quad[d1][1] * dr[1] + ((cellptr)p)->quad[d1][2] * dr[2];
                drquaddr += quaddr[d1]*dr[d1];
            }
        double dr5inv = 1.0/(distsq * distsq * dist);  /*   dr ** (-5)             */
        double phiquad = -0.5 * dr5inv * drquaddr;     /*   quad. part of poten.   */
#else
        double phiquad=0;
#endif
        switch(operation){
            case CALCPHI: 
                data[0] += phii - phiquad; 
                break;
            case CALCACC: 
                for(int d=0; d<N_DIM; d++)
                    data[d] += dr[d] * phii/distsq;
#ifdef TREECODE_QUADRUPOLE
                if(p->type == CELL) {                       /* if cell, add quad. term  */
                    phiquad = 5.0 * phiquad / distsq;         /*   save for acceleration  */
                    for(int d=0; d<N_DIM; d++)
                        data[d] -= dr[d]*phiquad + quaddr[d]*dr5inv;
                }
#endif
                { 
                    if(dist<data[N_DIM]) data[N_DIM]=dist;   // update minimum distance
                    //double fftime2=sqrt(distsq-eps*eps)*dist/phii;
                    //if(fftime2<data[N_DIM+1]) data[N_DIM+1]=fftime2;  // update minimum free-fall time squared
                }
                break;
        }
    }
    if (p->type == BODY)
	    nbb++;
	else
	    nbc++;
    
}

void CPotentialNB::walktreedens(const nodeptr p,   // pointer to interacting node or cell 
		   const double dsq,      // cell size squared
           const double searchradsq, // search radius squared
           const vec3 pos0,       // position to calculate quantity in question
           const int xp[N_DIM],   // integerized coordinates of the point in question
           const int lev,         // current grid level
           vectorpd* data)  const // array to store neighbours
{
    if(p==NULL) return;
    vec3 dr={p->pos[0]-pos0[0], p->pos[1]-pos0[1], p->pos[2]-pos0[2]};
    double distsq=sqr(dr[0])+sqr(dr[1])+sqr(dr[2]);
    if(p->type==CELL) {  
        if(distsq < dsq*3 + searchradsq) {  // worst case - CoM may lie on the other side of diagonal
            for(int c=0; c<NSUB; c++)
                if(((cellptr)p)->subp[c] != NULL)  /* walk into sub-cells */
                    walktreedens(((cellptr)p)->subp[c], dsq/4, searchradsq, pos0, xp, lev>>1, data);
        }
        else {  // point is far from our CoM, but lies within one of child cells, which still needs to be opened
            int subcell=subindex(xp, lev);
            if(((cellptr)p)->subp[subcell] != NULL)  /* walk into this sub-cell */
                walktreedens(((cellptr)p)->subp[subcell], dsq/4, searchradsq, pos0, xp, lev>>1, data);
        }
    }
    else if (distsq < searchradsq)   // this is a leaf within search radius
        data->push_back(paird(sqrt(distsq),p->mass));
}

}