#include "defines.h"
#include <QFile>
#include <QTextStream>
#include <QProcess>
#define _USE_MATH_DEFINES
#include <cmath>
#include <algorithm>
#include <gsl/gsl_math.h>
#include <glpk.h>    // GNU linear programming kit - optimization problem solver
#include "schwarzschild.h"
#include "core.h"

namespace smile{

CSpatialGrid::CSpatialGrid(const CPotential* _potential, const int _numShells, const int _linesPerSegment):
potential(_potential), 
linesPerSegment(_linesPerSegment), 
numShells(_numShells),
cellsPerSegment(linesPerSegment*linesPerSegment),
cellsPerShell(3*cellsPerSegment),
cellsInModel(numShells*cellsPerShell)
{
    cellBoundary.resize(linesPerSegment+1);
    shellBoundary.resize(numShells+1);

    for(int i=0; i<linesPerSegment; i++)
        cellBoundary[i] = tan(M_PI/4/linesPerSegment*i);
    cellBoundary[linesPerSegment] = 1.0;  // exact value
    shellBoundary[0]=0;
    double Gamma = potentialGamma();
    switch(potential->PotentialType())
    {  // everything except Dehnen/plummer is quite arbitrary and should not be used!
    case CPotential::PT_LOG:
        for(int i=0; i<numShells; i++)
            shellBoundary[i+1] = (i+1.0)/(numShells+1) * 2;
        break;
    case CPotential::PT_BSE:
    case CPotential::PT_NB:
    case CPotential::PT_DEHNEN:
    case CPotential::PT_DEHNENBSE:  // Dehnen
    case CPotential::PT_PLUMMERBSE:
        for(int i=0; i<numShells; i++)
        {
            double f = (i+1.0)/(numShells+1);
            if(potential->PotentialType()==CPotential::PT_PLUMMERBSE)
                shellBoundary[i+1] = 1/sqrt(pow(f,-2.0/3)-1);   
            else
            {
                double b=pow(f, 1/(3-Gamma));
                shellBoundary[i+1] = b/(1-b);
            }
        }
        if(potential->PotentialType()==CPotential::PT_PLUMMERBSE)
            totalMass = pow(sqr(shellBoundary[numShells])/(1+sqr(shellBoundary[numShells])), 1.5);
        else
            totalMass = pow(shellBoundary[numShells]/(1+shellBoundary[numShells]), 3-Gamma);
        break;
    case CPotential::PT_SCALEFREE: 
    case CPotential::PT_SCALEFREEBSE:   // self-similar cusp, should not really be used
        for(int i=0; i<numShells; i++)
            shellBoundary[i+1] = pow((i+1.0)/(numShells+1), 1/(3-((CPotentialScaleFree*)potential)->Gamma)) * 2;
        break;
    default:
        return;
    }
}

double CSpatialGrid::potentialGamma() const
{
    double Gamma=-1;
    switch(potential->PotentialType())
    {
        case CPotential::PT_DEHNEN:      Gamma=((CPotentialDehnen*)potential)->Gamma; break;
        case CPotential::PT_BSE: 
        case CPotential::PT_DEHNENBSE:    Gamma=((CPotentialBSE*)potential)->Gamma; break;
        case CPotential::PT_SCALEFREE:   
        case CPotential::PT_SCALEFREEBSE: Gamma=((CPotentialScaleFree*)potential)->Gamma; break;
        case CPotential::PT_NB:          Gamma=((CPotentialNB*)potential)->Gamma; break;
        default: Gamma=-1;   // N/A
    }
    return Gamma;
}

void CSpatialGrid::cellCenter(int C, double& X, double& Y, double& Z) const
{
    int nshell = C / cellsPerShell;
    if(nshell<0 || nshell>=numShells) 
    {
        X=0; Y=0; Z=0;
        return;
    }
    int indseg = C % cellsPerShell;
    int nseg = indseg / cellsPerSegment;
    int ind1 =(indseg % cellsPerSegment) % linesPerSegment;
    int ind2 =(indseg % cellsPerSegment) / linesPerSegment;
    double ratio1 = (cellBoundary[ind1] + cellBoundary[ind1+1])/2;
    double ratio2 = (cellBoundary[ind2] + cellBoundary[ind2+1])/2;
    double r = (shellBoundary[nshell] + shellBoundary[nshell+1])/2;
    double coord0 = r/sqrt(1 + sqr(ratio1) + sqr(ratio2));
    double coord1 = ratio1*coord0;
    double coord2 = ratio2*coord0;
    switch(nseg)
    {
    case 0:
        X=coord0;
        Y=coord1;
        Z=coord2;
        break;
    case 1:
        Y=coord0;
        Z=coord1;
        X=coord2;
        break;
    case 2:
        Z=coord0;
        X=coord1;
        Y=coord2;
        break;
    }
    Y *= potential->q;
    Z *= potential->p;
    if(whichCell(X,Y,Z) != C)
        X=0;
}

int CSpatialGrid::whichCell(double X, double Y, double Z) const
{
    X=fabs(X); Y=fabs(Y/potential->q); Z=fabs(Z/potential->p);
    int nseg=-1, nshell=0;
    double coord0, coord1, coord2;
    if((X>=Y)&&(X>Z))
    {
        nseg=0;
        coord0=X;
        coord1=Y;
        coord2=Z;
    } else
    if((Y>=Z)&&(Y>X))
    {
        nseg=1;
        coord0=Y;
        coord1=Z;
        coord2=X;
    } else
    if((Z>=X)&&(Z>Y))
    {
        nseg=2;
        coord0=Z;
        coord1=X;
        coord2=Y;
    }
    if(nseg==-1)
        return 0;  // should not happen, but who knows?...
    double ratio1 = coord1/coord0;
    double ratio2 = coord2/coord0;
    int ind1=0, ind2=0;
    for(int i=1; i<linesPerSegment; i++)
    {
        if(ratio1>=cellBoundary[i]) ind1=i;
        if(ratio2>=cellBoundary[i]) ind2=i;
    }
    int indseg = ind1 + ind2*linesPerSegment + nseg*cellsPerSegment;
    double m=sqrt(sqr(X)+sqr(Y)+sqr(Z));
    for(int i=0; i<=numShells; i++)
        if(m>=shellBoundary[i]) nshell=i;
    int ncell = nshell*cellsPerShell + indseg;
    return (ncell<cellsInModel)?ncell:cellsInModel;
}

/// Class for performing Schwarzschild modelling
CSchwarzschildModel::CSchwarzschildModel(const CPotential* _potential, const int _numShells, const int _linesPerSegment)
{
    grid = new CSpatialGrid(_potential, _numShells, _linesPerSegment);
    orbits=NULL;   // will be initialized later
    // calculate mass in cells
    shellEnergy.resize(grid->numShells+1);
    shellTorbx.resize(grid->numShells+1);
    shellEnergy[0]=shellTorbx[0]=0;
    double ip[3]={0,0,0}, iv[3]={0,0,0}, lf[3]={1,1,1};

    for(int i=1; i<=grid->numShells; i++)
    {
        shellEnergy[i]=grid->potential->Phi(grid->shellBoundary[i], 0, 0);
        shellTorbx[i]=grid->potential->longaxisperiod(shellEnergy[i]);
    }
    cellMass.resize(grid->cellsInModel+1);
    cellPenalty.resize(grid->cellsInModel+1);
    
    double M=0;  // total mass
    if(grid->potential->PotentialType()!=CPotential::PT_NB)
    for(int c=0; c<grid->cellsInModel; c++)
    {
        int nshell = c / grid->cellsPerShell;
        int indseg = c % grid->cellsPerShell;
        int nseg = indseg / grid->cellsPerSegment;
        int ind1 =(indseg % grid->cellsPerSegment) % grid->linesPerSegment;
        int ind2 =(indseg % grid->cellsPerSegment) / grid->linesPerSegment;
        double dr1 = grid->cellBoundary[ind1+1] - grid->cellBoundary[ind1];
        double dr2 = grid->cellBoundary[ind2+1] - grid->cellBoundary[ind2];
        double mass=0;
        for(int i1=0; i1<NUM_SUBDIV_ANGULAR; i1++)
            for(int i2=0; i2<NUM_SUBDIV_ANGULAR; i2++)
            {
                double r1l = grid->cellBoundary[ind1] + dr1*i1/NUM_SUBDIV_ANGULAR;
                double r1h = grid->cellBoundary[ind1] + dr1*(i1+1)/NUM_SUBDIV_ANGULAR;
                double r2l = grid->cellBoundary[ind2] + dr2*i2/NUM_SUBDIV_ANGULAR;
                double r2h = grid->cellBoundary[ind2] + dr2*(i2+1)/NUM_SUBDIV_ANGULAR;
                double phill = asin(r1l*r2l/sqrt((1+sqr(r1l))*(1+sqr(r2l))));
                double philh = asin(r1l*r2h/sqrt((1+sqr(r1l))*(1+sqr(r2h))));
                double phihl = asin(r1h*r2l/sqrt((1+sqr(r1h))*(1+sqr(r2l))));
                double phihh = asin(r1h*r2h/sqrt((1+sqr(r1h))*(1+sqr(r2h))));
                double area = (phill+phihh-philh-phihl);
                double ratio1 = (r1l+r1h)/2;
                double ratio2 = (r2l+r2h)/2;
                double r=0, rprev, rho=0, rhoprev;
                for(int ir=0; ir<=NUM_SUBDIV_RADIAL; ir++)
                {
                    rprev=r; rhoprev=rho;
                    r = grid->shellBoundary[nshell] + (grid->shellBoundary[nshell+1] - grid->shellBoundary[nshell]) * ir / NUM_SUBDIV_RADIAL;
                    if(r==0) r=grid->shellBoundary[1]*0.01/NUM_SUBDIV_RADIAL;
                    double coord0 = r/sqrt(1 + sqr(ratio1) + sqr(ratio2));
                    double coord1 = ratio1*coord0;
                    double coord2 = ratio2*coord0;
                    double X, Y, Z;
                    switch(nseg)
                    {
                    case 0:
                        X=coord0;
                        Y=coord1;
                        Z=coord2;
                        break;
                    case 1:
                        Y=coord0;
                        Z=coord1;
                        X=coord2;
                        break;
                    case 2:
                        Z=coord0;
                        X=coord1;
                        Y=coord2;
                        break;
                    }
                    Y *= grid->potential->q;
                    Z *= grid->potential->p;
                    rho = grid->potential->Rho(X, Y, Z);
                    if(ir>0)
                    {
                        double gamma = - log(rho/rhoprev)/log(r/rprev);
                        if(!gsl_finite(gamma)) gamma=0; // avoid NaN
                        if(gamma<0)
                            gamma=0; 
                        double intrad = rho*pow(r,gamma) * (gamma!=3 ? (pow(r, 3-gamma)-pow(rprev, 3-gamma))/(3-gamma) : log(r/rprev));
                        if(!gsl_finite(intrad)) intrad=0;
                        mass +=  area * intrad * 8*grid->potential->q*grid->potential->p;
                    }
                }
            }
        cellMass[c] = mass;
        cellPenalty[c]=0;
        M += mass;
    }
    else  // frozen Nbody potential - simply count bodies in each cell
    {
        CPotentialNB* potential=(CPotentialNB*)grid->potential;
        for(std::vector<CPotentialNB::body>::const_iterator body=potential->bodytab.begin(); body!=potential->bodytab.end(); body++)
        {
            int cell=grid->whichCell(body->pos[0], body->pos[1], body->pos[2]);
            cellMass[cell] += body->mass;
            M += body->mass;
        }
    }
    grid->totalMass=M;
    for(int c=0; c<=grid->cellsInModel; c++)
        cellMass[c]/=M;
}

CSchwarzschildModel::~CSchwarzschildModel()
{
    delete grid;
}

// driver routines
double CSchwarzschildModel::solveOptimizationProblem(bool QP)
{
    if(orbits==NULL || orbits->OrbitList.size()==0) 
        return -1;
    // solve linear/quadratic optimization problem
    prepareModel();
    // call solver; might also use seed values in orbitWeight (not implemented yet)
    int result = doOptimizationProblem(QP);;
    updateModel();
    if(result<0) return result; else 
    return calcCellPenalty();
}

double CSchwarzschildModel::solveLucyIterationsOptimization(int &nIter)
{
    if(orbits==NULL || orbits->OrbitList.size()==0) 
        return -1;
    // solve optimization problem
    prepareModel();
    nIter=0;
    // call solver
    nIter = doLucyIterationsOptimization();
    updateModel();
    return calcCellPenalty();
}

// below follow private lower-level functions
// pre-processing 
void CSchwarzschildModel::prepareModel()
{
    emit calcProgress("Preparing model");
    maxShellSuperOrbit=(config.maxShellSuperOrbit<=0 || config.maxShellSuperOrbit>grid->numShells) ? grid->numShells : config.maxShellSuperOrbit;
    orbits->markChaoticOrbits();
    try{
    if(config.chaoticSuperOrbit)
        weightFactor.resize(maxShellSuperOrbit, config.chaoticWeightFactor);
    linearMatrix.clear();
    vectorf row;
    if(config.chaoticSuperOrbit)   // count chaotic orbits
    {
        linearMatrix.resize(maxShellSuperOrbit);
        for(int s=0; s<maxShellSuperOrbit; s++)
        {   // first "maxShellSuperOrbit" orbits are for chaotic super-orbit at each energy shell
            linearMatrix[s].assign(grid->cellsInModel+(config.sm_constrainBeta?1:0)*grid->numShells, 0);    // reserve space for super-orbit cell weight array
            numChaoticInShell.push_back(0);
        }
        for(int o=0; o<orbits->OrbitList.size(); o++)
        {
            if(orbits->OrbitList[o].isChaotic && orbits->OrbitList[o].shellIndex<maxShellSuperOrbit)
            {
                int s = orbits->OrbitList[o].shellIndex;
                for(int c=0; c<grid->cellsInModel; c++) 
                {
                    linearMatrix[s][c] += orbits->OrbitList[o].celltimes[c];
                    if(config.sm_constrainBeta)
                    {
                        double beta = config.sm_betaIn + (config.sm_betaOut-config.sm_betaIn)*(c/grid->cellsPerShell)/grid->numShells;
                        linearMatrix[s][c/grid->cellsPerShell+grid->cellsInModel] += orbits->OrbitList[o].celltimes[c]*(orbits->OrbitList[o].cellVt[c] - 2*(1-beta)*orbits->OrbitList[o].cellVr[c]);
                    }
                }
                numChaoticInShell[s]++;
            }
            else
            {
                row.assign(orbits->OrbitList[o].celltimes.begin(), orbits->OrbitList[o].celltimes.end());
                row.pop_back();  // eliminate last cell (which goes to infinity)
                if(config.sm_constrainBeta)
                {
                    for(int s=0; s<grid->numShells; s++)
                        row.push_back(0);
                    for(int c=0; c<grid->cellsInModel; c++)
                    {
                        double beta = config.sm_betaIn + (config.sm_betaOut-config.sm_betaIn)*(c/grid->cellsPerShell)/grid->numShells;
                        row[c/grid->cellsPerShell+grid->cellsInModel] += orbits->OrbitList[o].celltimes[c]*(orbits->OrbitList[o].cellVt[c] - 2*(1-beta)*orbits->OrbitList[o].cellVr[c]);
                    }
                }
                linearMatrix.push_back(row);
            }
        }
        for(int s=0; s<maxShellSuperOrbit; s++)
            if(numChaoticInShell[s]>0)
            {
                for(int c=0; c<linearMatrix[s].size(); c++) 
                    linearMatrix[s][c] /= numChaoticInShell[s];
                weightFactor[s] /= numChaoticInShell[s];
            }
    }
    else
    {
        for(int o=0; o<orbits->OrbitList.size(); o++)
        {
            row.assign(orbits->OrbitList[o].celltimes.begin(), orbits->OrbitList[o].celltimes.end());
            row.pop_back();  // eliminate last cell (which goes to infinity)
            if(config.sm_constrainBeta)
            {
                for(int s=0; s<grid->numShells; s++)
                {
                    double beta = config.sm_betaIn+(config.sm_betaOut-config.sm_betaIn)*s/grid->numShells;
                    double q=0;
                    for(int c=s*grid->cellsPerShell; c<(s+1)*grid->cellsPerShell; c++)
                        q += orbits->OrbitList[o].celltimes[c]*(orbits->OrbitList[o].cellVt[c] - 2*(1-beta)*orbits->OrbitList[o].cellVr[c]);
                    row.push_back(q);
                }
            }
            linearMatrix.push_back(row);
        }
    }
    // initialize rhs
    rhs.assign(grid->cellsInModel+(config.sm_constrainBeta?1:0)*grid->numShells, 0);
    constraintWeight.assign(grid->cellsInModel+(config.sm_constrainBeta?1:0)*grid->numShells, 0);
    for(int c=0; c<grid->cellsInModel; c++)
    {
        rhs[c] = cellMass[c];
        constraintWeight[c] = 1.0;
        if(config.sm_constrainBeta)
        {
            constraintWeight[c/grid->cellsPerShell+grid->cellsInModel] = SCHW_CONSTRAINT_BETA_WEIGHT*grid->cellsPerShell;
        }
    }
    orbitWeight.resize(linearMatrix.size(), 0);
    weightFactor.resize(linearMatrix.size());
    if(config.chaoticSuperOrbit)
    {
        int oi=maxShellSuperOrbit;
        for(int o=0; o<orbits->OrbitList.size(); o++)
            if(orbits->OrbitList[o].isChaotic && orbits->OrbitList[o].shellIndex<maxShellSuperOrbit)
            {
                int s=orbits->OrbitList[o].shellIndex;
                weightFactor[s] = config.chaoticWeightFactor;
            }
            else
            {
                weightFactor[oi] = orbits->OrbitList[o].isChaotic ? config.chaoticWeightFactor : 0.0;
                oi++;
            }
    }
    else
    {
        for(int o=0; o<orbits->OrbitList.size(); o++)
        {
            double wf=log10(orbits->OrbitList[o].lfccdiff)-log10(config.chaoticMinFreqDiff) + 0.5;   // `soft' transition from reg to chaos, based on freq.diff.rate
            if(wf<0) wf=0; if(wf>1) wf=1;
            if(orbits->OrbitList[o].lambda>config.chaoticMinLambda) wf=1;
            weightFactor[o]= orbits->OrbitList[o].shellIndex<maxShellSuperOrbit ? wf*config.chaoticWeightFactor : 0.0;
        }
    }
    }
    catch(const std::bad_alloc&) {
        emit calcProgress("Not enough memory");
        return;
    }
}

// post-processing
void CSchwarzschildModel::updateModel(bool finish)
{
    if(finish) emit calcProgress("Finishing model");
    // stats
    nOrbitsNonZero=0;
    sumSquaredWeights=0;
    massRegular=0;
    massChaotic=0;
    if(config.chaoticSuperOrbit)
    {
        int oi=config.chaoticSuperOrbit ? maxShellSuperOrbit : 0;
        for(int o=0; o<orbits->OrbitList.size(); o++)
        {
            if(orbits->OrbitList[o].isChaotic && orbits->OrbitList[o].shellIndex<maxShellSuperOrbit)
            {
                int s = orbits->OrbitList[o].shellIndex;
                orbits->OrbitList[o].orbitWeight = orbitWeight[s]/numChaoticInShell[s];
                if(orbits->OrbitList[o].orbitWeight < SCHW_MIN_ORBIT_WEIGHT_USED/orbits->OrbitList.size()) 
                    orbits->OrbitList[o].orbitWeight=0;
                massChaotic += orbits->OrbitList[o].orbitWeight;
            }
            else
            {
                orbits->OrbitList[o].orbitWeight = orbitWeight[oi];
                if(orbits->OrbitList[o].orbitWeight < SCHW_MIN_ORBIT_WEIGHT_USED/orbits->OrbitList.size()) 
                    orbits->OrbitList[o].orbitWeight=0;
                oi++;
                if(orbits->OrbitList[o].isChaotic)
                    massChaotic += orbits->OrbitList[o].orbitWeight;   // it may still be chaotic, but not belonging to super-orbit
                else
                    massRegular += orbits->OrbitList[o].orbitWeight;
            }
            sumSquaredWeights += sqr(orbits->OrbitList[o].orbitWeight);
            if(orbits->OrbitList[o].orbitWeight>0) nOrbitsNonZero++;
        }
    }
    else
        for(int o=0; o<orbits->OrbitList.size(); o++)
        {
            orbits->OrbitList[o].orbitWeight=orbitWeight[o];
            if(orbits->OrbitList[o].orbitWeight < SCHW_MIN_ORBIT_WEIGHT_USED/orbits->OrbitList.size()) 
                orbits->OrbitList[o].orbitWeight=0;
            sumSquaredWeights += sqr(orbits->OrbitList[o].orbitWeight);
            if(orbits->OrbitList[o].orbitWeight>0) nOrbitsNonZero++;
            if(orbits->OrbitList[o].isChaotic)
                massChaotic += orbits->OrbitList[o].orbitWeight;
            else
                massRegular += orbits->OrbitList[o].orbitWeight;
        }
    if(!finish) return;   // just progress report in Lucy iterations
    for(int o=0; o<linearMatrix.size(); o++)
        linearMatrix[o].clear();
    linearMatrix.clear();
    orbitWeight.clear();
    weightFactor.clear();
    numChaoticInShell.clear();
}

// core optimization routines
int CSchwarzschildModel::doOptimizationProblem(bool QP)
{
    int numOrbitsUsed = orbitWeight.size();
    int numCellsUsed = rhs.size();
    if(!config.sm_useBPMPD)   // use GNU linear programming kit
    {
        glp_prob *lp;
        lp = glp_create_prob();
        glp_set_obj_dir(lp, GLP_MIN);
        glp_add_rows(lp, numCellsUsed);
        for(int c=0; c<numCellsUsed; c++)
            glp_set_row_bnds(lp, c+1, GLP_FX, rhs[c], rhs[c]);
        glp_add_cols(lp, numOrbitsUsed+2*numCellsUsed);
        if(config.sm_maxWeight>0)
            for(int c=0; c<numOrbitsUsed; c++)
                glp_set_col_bnds(lp, c+1, GLP_DB, 0.0, config.sm_maxWeight);
        else
            for(int c=0; c<numOrbitsUsed; c++)
                glp_set_col_bnds(lp, c+1, GLP_LO, 0.0, 0.0);
        for(int c=0; c<2*numCellsUsed; c++)  // lambda, mu - slack variables (should be zero after optimization)
        {
            glp_set_col_bnds(lp, c+1+numOrbitsUsed, GLP_LO, 0.0, 0.0);
            glp_set_obj_coef(lp, c+1+numOrbitsUsed, constraintWeight[c/2]);
        }
        std::vector<int> ic(1);
        std::vector<int> io(1);
        std::vector<double> matrix(1);
        int ind=1;
        for(int o=0; o<numOrbitsUsed; o++)
        {
            for(int c=0; c<numCellsUsed; c++)
            {
                if(linearMatrix[o][c]!=0)
                {
                    ic.push_back(c+1);
                    io.push_back(o+1);
                    matrix.push_back(linearMatrix[o][c]);
                }
            }
            if(weightFactor[o]!=0)
                glp_set_obj_coef(lp, o+1, weightFactor[o]/numOrbitsUsed);
        }
        for(int c=0; c<numCellsUsed; c++)
        {

            ic.push_back(c+1);
            io.push_back(numOrbitsUsed+2*c+1);
            matrix.push_back(1.0);
            ic.push_back(c+1);
            io.push_back(numOrbitsUsed+2*c+2);
            matrix.push_back(-1.0);
        }
        glp_load_matrix(lp, ic.size()-1, &(ic.front()), &(io.front()), &(matrix.front()));

        emit calcProgress("Starting optimization");

        //int status = glp_simplex(lp, NULL);
        int status = glp_interior(lp, NULL);
        for(int o=0; o<numOrbitsUsed; o++)
        {
            double v = glp_ipt_col_prim(lp, o+1);
            orbitWeight[o] = (v>=SCHW_MIN_ORBIT_WEIGHT_USED/numOrbitsUsed) ? v:0;
        }
        glp_delete_prob(lp); 
        return -status;  // 0 for success, <0 for failure
    }
    else   // solve optimization problem using external solver BPMPD
    {
        // first write out MPS file
        emit calcProgress("Writing file");

        QString dataFileName = core->TempDir+"data.mps";
        QFile file(dataFileName);
        file.open(QFile::WriteOnly | QFile::Text);
        QTextStream out(&file);
        QString Str;
        out << "NAME  testmp" << endl;
        out << "ROWS" << endl << " N  OBJFUN" << endl;
        for(int c=0; c<numCellsUsed; c++)
            out << Str.sprintf(" E  R%07d", c+1) << endl;
        if(config.sm_maxWeight>0)
        {
            for(int o=0; o<numOrbitsUsed; o++)
                out << Str.sprintf(" L  W%07d", o+1) << endl;
        }
        out << "COLUMNS";
        int parity=0;
        for(int o=0; o<numOrbitsUsed; o++)
        {
            if(weightFactor[o]!=0)
            {
                out << endl << Str.sprintf("    C%07d  OBJFUN    %12G ", o+1, weightFactor[o]/numOrbitsUsed);
                parity=1;
            }
            if(config.sm_maxWeight>0)
            {
                if(parity==0) 
                    out << endl << Str.sprintf("    C%07d", o+1);
                out << Str.sprintf("  W%07d             1 ", o+1);
                parity=1-parity;
            }
            for(int c=0; c<numCellsUsed; c++)
                if(linearMatrix[o][c]!=0)
                {
                    if(parity==0) 
                        out << endl << Str.sprintf("    C%07d", o+1);
                    out << Str.sprintf("  R%07d  %12G ", c+1, linearMatrix[o][c]);
                    parity = 1-parity;
                }
                parity=0;
        }
        out << endl;
        for(int c=0; c<numCellsUsed; c++)
        {
            out << Str.sprintf("    L%07d  OBJFUN    %12G   R%07d  %12G", c+1, constraintWeight[c], c+1,  1.0) << endl;
            out << Str.sprintf("    M%07d  OBJFUN    %12G   R%07d  %12G", c+1, constraintWeight[c], c+1, -1.0) << endl;
        }
        out << "RHS";
        parity=0;
        for(int c=0; c<numCellsUsed; c++)
        {
            if(parity==0)
                out << endl << "    RHS1    ";
            out << Str.sprintf("  R%07d  %12G ", c+1, rhs[c]);
            parity=1-parity;
        }
        if(config.sm_maxWeight>0)
        {
            for(int o=0; o<numOrbitsUsed; o++)
            {
                if(parity==0)
                    out << endl << "    RHS1    ";
                out << Str.sprintf("  W%07d  %12G ", o+1, config.sm_maxWeight);
                parity=1-parity;
            }
        }
        if(QP)
        {   // add Quadratic optimization block
            out << endl << "QUADOBJ";
            for(int o=0; o<numOrbitsUsed; o++)
                out << endl << Str.sprintf("    C%07d  C%07d  %12G", o+1, o+1, SCHW_CONSTRAINT_QUAD_WEIGHT);
        }
        out << endl << "ENDATA" << endl;
        file.close();

        // then run BPMPD
        emit calcProgress("Starting optimization");
        QProcess process;
        process.setWorkingDirectory(core->TempDir);  // to read .par file correctly
        process.start((core->AppDir+"bpmpd"), QStringList("data") );
        if(!process.waitForStarted()) return -2;  // report error
        QProcess::ProcessState st=process.state();
        QProcess::ProcessError er=process.error();
        if(st!=QProcess::Running) return -2-er;  // report error
        process.waitForFinished(-1);
        file.remove();
        // finally, read results
        emit calcProgress("Reading results");
        QFile fileo(core->TempDir+"data.out");
        int status=-1;
        if(fileo.open(QIODevice::ReadOnly))
        {
            for(int c=0; c<grid->cellsInModel; c++)
                cellPenalty[c]=0;
            for(int o=0; o<numOrbitsUsed; o++)
                orbitWeight[o]=0;
            QTextStream in(&fileo);
            QString S;
            int numColumnsResult=0;
            do{
                S = in.readLine();
                if(S.contains("Limit of the evaluation version is reached")) return -13;  // BPMPD: sadly we won't know the answer...
                bool ok=false;
                QStringList fields=S.trimmed().split(QRegExp("\\s+"));
                int c=fields[0].mid(1,7).toInt(&ok);
                if(fields[0].mid(0,1)=="C" && ok && c>0 && c<=numOrbitsUsed)
                {
                    double v=fields[1].toDouble(&ok);
                    if(ok)
                    {
                        bool used=(v>=SCHW_MIN_ORBIT_WEIGHT_USED/numOrbitsUsed);
                        orbitWeight[c-1] = used?v:0;
                        numColumnsResult++;
                    }
                }
            }
            while(!in.atEnd());
            status=(numColumnsResult==numOrbitsUsed ? 0 : -14);
            fileo.close();
            fileo.remove();
        }
        return status;
    }
}

int CSchwarzschildModel::doLucyIterationsOptimization()
{
    QFile f(core->TempDir+"lucy.log");
    f.open(QFile::WriteOnly | QFile::Text);
    QTextStream logfile(&f);
    stopflag=false;
    // prepare iterations
    int niter=0;
    int numOrbitsUsed = orbitWeight.size();
    for(int o=0; o<numOrbitsUsed; o++)
        orbitWeight[o]=1.0/numOrbitsUsed;
    vectord P(grid->cellsInModel+1);
    vectord currWeight(orbitWeight);
    vectord totalWeight(numOrbitsUsed);
    for(int o=0; o<numOrbitsUsed; o++)
    {
        totalWeight[o]=0;
        for(int c=0; c<grid->cellsInModel; c++)
            totalWeight[o] += linearMatrix[o][c];
    }
    double diff=0, bestdiff=1e10;
    int bestdiffstepnum=-1, numCellsInfeas=0;
    do
    {
        diff=0;
        numCellsInfeas=0;
        // 1st stage
        for(int c=0; c<grid->cellsInModel; c++)
        {
            double mass=0;
            for(int o=0; o<numOrbitsUsed; o++)
                mass += linearMatrix[o][c]*currWeight[o];
            P[c] = cellMass[c] / mass;
            diff += fabs(mass-cellMass[c]);
            if(fabs(mass-cellMass[c])/cellMass[c] > SCHW_MAX_CELLMASS_REL_DIFFERENCE)
                numCellsInfeas++;
        }
        if(diff<bestdiff)
        {
            bestdiff=diff;
            bestdiffstepnum=niter;
            orbitWeight.assign(currWeight.begin(), currWeight.end());
        }
        // 2nd stage
        int numOrbitsNonzero=0;
        for(int o=0; o<numOrbitsUsed; o++)
          if(currWeight[o]>0)
          {
            double Q=0;
            for(int c=0; c<grid->cellsInModel; c++)
                Q += linearMatrix[o][c]*P[c];
            if(totalWeight[o]>0) 
                currWeight[o] *= Q/totalWeight[o];
            if(currWeight[o]<SCHW_MIN_ORBIT_WEIGHT_USED/numOrbitsUsed) currWeight[o]=0;
            else numOrbitsNonzero++;
          }
        if(niter%SCHW_ITERATIONS_UPDATE==0)
        {
            updateModel(false);
            emit calcProgress("Iteration #"+QString::number(niter)+"\ndiff="+QString::number(diff)+
                "\norbits used="+QString::number(numOrbitsNonzero)+"\ncells infeasible="+QString::number(numCellsInfeas));
            logfile << niter << "\t" << diff << "\t" << numOrbitsNonzero << "\t" << numCellsInfeas << "\n";
            logfile.flush();
        }
        niter++;
    }
    while(!stopflag && ((niter<SCHW_MIN_ITERATIONS) || (niter<SCHW_MAX_ITERATIONS) && (niter-bestdiffstepnum<SCHW_MAX_GROWING_ITERATIONS)
        && (diff>SCHW_MAX_CELLMASS_REL_DIFFERENCE/grid->cellsInModel) && (numCellsInfeas>0)));
    return niter;
}

double CSchwarzschildModel::calcCellPenalty()
{
    double diff=0;
    nCellsInfeasible=0;
    for(int c=0; c<grid->cellsInModel; c++)
    {
        double mass=0;
        for(int o=0; o<orbits->OrbitList.size(); o++)
            mass += orbits->OrbitList[o].orbitWeight * orbits->OrbitList[o].celltimes[c];
        cellPenalty[c] = mass - cellMass[c];
        if(fabs(cellPenalty[c]/cellMass[c])>SCHW_MAX_CELLMASS_REL_DIFFERENCE) 
            nCellsInfeasible++; 
        else 
            cellPenalty[c]=0;
        diff+=fabs(cellPenalty[c]);
    }
    return diff;
}

bool CSchwarzschildModel::exportGrid(QString fileName)
{
    QFile file(fileName);
    if (!file.open(QFile::WriteOnly | QFile::Text)) 
        return false;
    QTextStream out(&file);
    out << "x\ty\tz\tshell\tEshell\trshell\tTorbx\tM\tdM/M\tNOtotal\tNOused\tsigmaR\tsigmaT\tbeta\tchaoticfrac" << endl;
    std::vector<int> ntotal(grid->numShells), nused(grid->numShells);
    std::vector<double> shellMass(grid->numShells), shellPenalty(grid->numShells), shellVr2(grid->numShells), shellVt2(grid->numShells);
    for(int c=0; c<grid->cellsInModel; c++)
    {
        int ishell=c/grid->cellsPerShell;
        double X, Y, Z;
        grid->cellCenter(c, X, Y, Z);
        int ot=0, ou=0;
        double vr2=0, vt2=0, Mreg=0, Mcha=0;
        for(int o=0; o<orbits->OrbitList.size(); o++)
            if(orbits->OrbitList[o].celltimes[c]>0)
            {
                ot++;
                ntotal[ishell]++;
                if(orbits->OrbitList[o].orbitWeight>0) 
                {
                    ou++;
                    nused[ishell]++;
                    vr2 += orbits->OrbitList[o].cellVr[c] * orbits->OrbitList[o].celltimes[c] * orbits->OrbitList[o].orbitWeight;
                    vt2 += orbits->OrbitList[o].cellVt[c] * orbits->OrbitList[o].celltimes[c] * orbits->OrbitList[o].orbitWeight;
                    if(orbits->OrbitList[o].isChaotic)
                        Mcha += orbits->OrbitList[o].celltimes[c] * orbits->OrbitList[o].orbitWeight;
                    else
                        Mreg += orbits->OrbitList[o].celltimes[c] * orbits->OrbitList[o].orbitWeight;
                }
            }

        out << X << "\t" << Y << "\t" << Z << "\t" << 
            (ishell+1) << "\t" << shellEnergy[ishell+1] << "\t" << grid->shellBoundary[ishell+1] << "\t" << shellTorbx[ishell+1] << "\t" <<
            cellMass[c] << "\t" << cellPenalty[c]/cellMass[c] << "\t" << ot << "\t" << ou << "\t" << 
            sqrt(vr2/cellMass[c]) << "\t" << sqrt(vt2/cellMass[c]) << "\t" << (1-vt2/2/vr2) << "\t" << (Mcha/(Mcha+Mreg)) <<
            endl;
        shellMass[ishell]+=cellMass[c];
        shellPenalty[ishell]+=cellPenalty[c];
        shellVr2[ishell]+=vr2;
        shellVt2[ishell]+=vt2;
    }
    out << endl;
    for(int s=0; s<grid->numShells; s++)
        out << grid->shellBoundary[s+1] << "\t" << 0 << "\t" << 0 << "\t" << 
            (s+1) << "\t" << shellEnergy[s+1] << "\t" << grid->shellBoundary[s+1] << "\t" << shellTorbx[s+1] << "\t" <<
            shellMass[s] << "\t" << shellPenalty[s]/shellMass[s] << "\t" << ntotal[s] << "\t" << nused[s] << "\t" << 
            sqrt(shellVr2[s]/shellMass[s]) << "\t" << sqrt(shellVt2[s]/shellMass[s]) << "\t" << (1-shellVt2[s]/2/shellVr2[s]) << "\t\t" << 
            orbits->getOrbitPopulation(0, s*grid->cellsPerShell, (s+1)*grid->cellsPerShell-1) << endl;
    file.close();
    return true;
}

int CSchwarzschildModel::exportNbodyPrepare(int numPoints, QString fileName, int refineFactor)
{
    int numorb=0, niter=0;
    double pointmass0=grid->totalMass/numPoints;
    exportNbodyOrbitMass.assign(orbits->OrbitList.size(), 0);
    exportNbodyNumPoints = numPoints;
    exportNbodyFileName = fileName;
    // mass refinement ranking is based on the condition of minimum distance reached, measured by rplC - OR - energy
    std::vector<double> refineBoundary(refineFactor+1);
    std::vector<double> orbitRank(orbits->OrbitList.size());
    for(int o=0; o<orbits->OrbitList.size(); o++)
    {
        orbitRank[o] = orbits->OrbitList[o].Energy;
    }
    std::vector<double> sortedRank(orbitRank);
    std::sort(sortedRank.begin(), sortedRank.end());
    for(int rf=0; rf<=refineFactor; rf++)
        refineBoundary[rf] = sortedRank[ (sortedRank.size()-1) / ((int)pow(2.0, rf)) ];
    do{
        numorb=0;
        for(int o=0; o<orbits->OrbitList.size(); o++)
        {
            double orbitmass = pointmass0; 
            if(refineFactor>0)
            {
                for(int rf=0; rf<=refineFactor; rf++)
                    if(orbitRank[o] <= refineBoundary[rf]) orbitmass=pointmass0*pow(2.0, refineFactor-rf);
            }
            exportNbodyOrbitMass[o]=orbitmass;
            numorb += (int)(orbits->OrbitList[o].orbitWeight*grid->totalMass/orbitmass+0.5);
        }
        // now correct point mass
        pointmass0 = pointmass0*numorb/numPoints;
        niter++;
    } 
    while(numorb>0 && numorb!=numPoints && niter<42);
    if(numorb==0) return -1;
    int needReintegrate = 0;
    double sumWeight=0;
    int maxsampling=0;
    for(int o=0; o<orbits->OrbitList.size(); o++)
    {
        int numsp=(int)(orbits->OrbitList[o].orbitWeight*grid->totalMass/exportNbodyOrbitMass[o]+0.5);
        int NSP=orbits->OrbitList[o].trajsample[0].size();
        if(numsp>0) 
        {
            //if(ncur-nprev>NSP)
            if(numsp>NSP)
            {
                needReintegrate++;
                orbits->OrbitList[o].numSamplingPoints = numsp;
                orbits->OrbitList[o].state=0;
            }
            if(numsp>maxsampling) maxsampling=numsp;
        }
    }
    return needReintegrate;
}

void CSchwarzschildModel::exportNbody()
{
    int numPoints=exportNbodyNumPoints;
    vectorf pointmassind;
    double sumWeight=0;
    vectorf points[N_DIM*2];
    std::vector<int> pointindex;
    for(int d=0; d<N_DIM*2; d++) 
        points[d].reserve(numPoints);
    pointindex.reserve(numPoints);
    pointmassind.reserve(numPoints);
    int maxsampling=0, norbitsused=0;
    // controlling center mass position, net momentum and rotation
    double CX=0, CY=0, CZ=0, VX=0, VY=0, VZ=0, LX=0, LY=0, LZ=0, Lhm=0, Chm=0, Vhm=0;
    for(int o=0; o<orbits->OrbitList.size(); o++)
    {
//        sumWeight+=orbits->OrbitList[o].orbitWeight;
        double orbitmass=exportNbodyOrbitMass[o];
        int numsp=(int)(orbits->OrbitList[o].orbitWeight*grid->totalMass/orbitmass+0.5);
        int NSP=orbits->OrbitList[o].trajsample[0].size();
        if(numsp>0)
        {
            //if(ncur-nprev>maxsampling) maxsampling=ncur-nprev;
            if(numsp>maxsampling) maxsampling=numsp;
            norbitsused++;
            for(int p=0; p<std::min(/*ncur-nprev*/numsp, NSP); p++)
            {
                double xv[N_DIM*2];
                for(int d=0; d<N_DIM*2; d++)
                    xv[d] = orbits->OrbitList[o].trajsample[d][p];
                {   //!!! do not emit points with very extreme kinetic to total energy ratio
                    double W=grid->potential->Phi(xv[0], xv[1], xv[2]);
                    double K=(sqr(xv[3])+sqr(xv[4])+sqr(xv[5]))/2;
                    if(W<0 && (K+W>=0 || -K/(K+W)>sqrt(numPoints*1.0)*0.2)) continue;  // otherwise it will distort virial equilibrium too much
                }
                double lx=orbitmass*(xv[1]*xv[5]-xv[2]*xv[4]), 
                       ly=orbitmass*(xv[2]*xv[3]-xv[0]*xv[5]), 
                       lz=orbitmass*(xv[0]*xv[4]-xv[1]*xv[3]);
                // decide whether to flip the orbit, and about which plane.
                int nflip=0;
                double absL = fabs(LX+lx)+fabs(LY+ly)+fabs(LZ+lz) + 
                    (fabs(CX+orbitmass*xv[0])+fabs(CY+orbitmass*xv[1])+fabs(CZ+orbitmass*xv[2])) *
                    (fabs(VX+orbitmass*xv[3])+fabs(VY+orbitmass*xv[4])+fabs(VZ+orbitmass*xv[5])) / NBEXPORT_XV_FACTOR;
                double absLx= fabs(LX+lx)+fabs(LY-ly)+fabs(LZ-lz) + 
                    (fabs(CX-orbitmass*xv[0])+fabs(CY+orbitmass*xv[1])+fabs(CZ+orbitmass*xv[2])) *
                    (fabs(VX-orbitmass*xv[3])+fabs(VY+orbitmass*xv[4])+fabs(VZ+orbitmass*xv[5])) / NBEXPORT_XV_FACTOR;
                if(absLx<absL) { absL=absLx; nflip=1; }
                double absLy= fabs(LX-lx)+fabs(LY+ly)+fabs(LZ-lz) + 
                    (fabs(CX+orbitmass*xv[0])+fabs(CY-orbitmass*xv[1])+fabs(CZ+orbitmass*xv[2])) *
                    (fabs(VX+orbitmass*xv[3])+fabs(VY-orbitmass*xv[4])+fabs(VZ+orbitmass*xv[5])) / NBEXPORT_XV_FACTOR;
                if(absLy<absL) { absL=absLy; nflip=2; }
                double absLz= fabs(LX-lx)+fabs(LY-ly)+fabs(LZ+lz) + 
                    (fabs(CX+orbitmass*xv[0])+fabs(CY+orbitmass*xv[1])+fabs(CZ-orbitmass*xv[2])) *
                    (fabs(VX+orbitmass*xv[3])+fabs(VY+orbitmass*xv[4])+fabs(VZ-orbitmass*xv[5])) / NBEXPORT_XV_FACTOR;
                if(absLz<absL) { absL=absLz; nflip=3; }
                if(nflip>0)
                {   xv[nflip-1]=-xv[nflip-1]; xv[nflip-1+N_DIM]=-xv[nflip-1+N_DIM]; }
                for(int d=0; d<N_DIM*2; d++)
                    points[d].push_back(xv[d]);
                pointindex.push_back(o);
                pointmassind.push_back(orbitmass);
                // accumulate total momentum and angular momentum
                LX+=orbitmass*(xv[1]*xv[5]-xv[2]*xv[4]);
                LY+=orbitmass*(xv[2]*xv[3]-xv[0]*xv[5]);
                LZ+=orbitmass*(xv[0]*xv[4]-xv[1]*xv[3]);
                CX+=orbitmass*xv[0];
                CY+=orbitmass*xv[1];
                CZ+=orbitmass*xv[2];
                VX+=orbitmass*xv[3];
                VY+=orbitmass*xv[4];
                VZ+=orbitmass*xv[5];
                sumWeight+=orbitmass;
            }
        }
        if(sumWeight>0.5 && Lhm==0 && Vhm==0 && Chm==0)
        {   // store half-mass values
            Chm = sqrt(CX*CX+CY*CY+CZ*CZ)/sumWeight;
            Vhm = sqrt(VX*VX+VY*VY+VZ*VZ)/sumWeight;
            Lhm = sqrt(LX*LX+LY*LY+LZ*LZ)/sumWeight;
        }
    }
    numPoints = points[0].size();
    CX/=sumWeight;
    CY/=sumWeight;
    CZ/=sumWeight;
    VX/=sumWeight;
    VY/=sumWeight;
    VZ/=sumWeight;
    LX/=sumWeight;
    LY/=sumWeight;
    LZ/=sumWeight;
    for(int i=0; i<numPoints; i++)
    {
        points[3][i] -= VX;
        points[4][i] -= VY;
        points[5][i] -= VZ;
    }   // cancel total momentum

    std::vector<int> cellnp(grid->cellsInModel+1);
    std::vector<double> cellvr2(grid->cellsInModel+1), cellvt2(grid->cellsInModel+1), cellmass(grid->cellsInModel+1);
    std::vector<double> cellvx2(grid->cellsInModel+1), cellvy2(grid->cellsInModel+1), cellvz2(grid->cellsInModel+1);

    stopflag=false;
    {   // write orbits to Nbody snapshot file
        CSnapShotWriter snap(exportNbodyFileName+".nb");
        QString strGamma="";
        if(grid->potentialGamma()>=0)
            strGamma = " Gamma="+QString::number(grid->potentialGamma());
        if(grid->potential->Mbh>0)
            strGamma += " Mbh="+QString::number(grid->potential->Mbh);
        snap.WriteHistory(QString("SMILE potential=")+QString(grid->potential->PotentialName()) + 
            " b/a="+QString::number(grid->potential->q) + " c/a="+QString::number(grid->potential->p) + strGamma +
            " nshells="+QString::number(grid->numShells) + " ncells="+QString::number(grid->cellsInModel) + 
            " norbitsused="+QString::number(norbitsused) + " VERSION="+QString(VERSION));
        snap.WritePhase(points, pointmassind, 0);
    }   // object destroyed and file closed after this
    emit calcProgress("Number of bodies created="+QString::number(numPoints)+"\nnum orbits used="+QString::number(norbitsused) + 
        "\nmax sampling points="+QString::number(maxsampling) + 
        " \nAvg.X="+QString::number(sqrt(CX*CX+CY*CY+CZ*CZ)) + " (" + QString::number(Chm) + 
        ")\nAvg.V="+QString::number(sqrt(VX*VX+VY*VY+VZ*VZ)) + " (" + QString::number(Vhm) + ") [cancelled]" +
        " \nAvg.L="+QString::number(sqrt(LX*LX+LY*LY+LZ*LZ)) + " (" + QString::number(Lhm) + 
        ")"+(stopflag||!config.sm_exportNbodyOrbitsFile?"":"\nWriting orbits to text file..")
        );
    if(stopflag) return;  
    // write orbits in text file
    if(config.sm_exportNbodyOrbitsFile)
    {
        try{
            COrbitLibrary* nbodies = new COrbitLibrary(grid->potential, NULL);
            nbodies->OrbitList.reserve(numPoints);
            for(int i=0; i<numPoints; i++)
            {
                if(points[0][i] || points[1][i] || points[2][i])
                {
                    double r=sqrt(points[0][i]*points[0][i]+points[1][i]*points[1][i]+points[2][i]*points[2][i]);
                    double vr2=sqr(points[0][i]*points[3][i]+points[1][i]*points[4][i]+points[2][i]*points[5][i])/sqr(r);
                    double vt2=(sqr(points[0][i]*points[4][i]-points[1][i]*points[3][i])+sqr(points[1][i]*points[5][i]-points[2][i]*points[4][i])+sqr(points[2][i]*points[3][i]-points[0][i]*points[5][i]))/sqr(r);
                    // store grid
                    int c=grid->whichCell(points[0][i], points[1][i], points[2][i]);
                    cellnp[c]++;
                    cellvr2[c] += vr2*pointmassind[i];
                    cellvt2[c] += vt2*pointmassind[i];
                    cellmass[c]+= pointmassind[i];
                    cellvx2[c] += points[3][i]*points[3][i]*pointmassind[i];
                    cellvy2[c] += points[4][i]*points[4][i]*pointmassind[i];
                    cellvz2[c] += points[5][i]*points[5][i]*pointmassind[i];

                    COrbitDesc* src = &(orbits->OrbitList[pointindex[i]]);
                    double ip[3], iv[3];
                    ip[0]=points[0][i];
                    ip[1]=points[1][i];
                    ip[2]=points[2][i];
                    iv[0]=points[3][i];
                    iv[1]=points[4][i];
                    iv[2]=points[5][i];
                    COrbitDesc od=COrbitDesc(src->potential, NULL, ip, iv, src->time, src->timeStep, src->timeUnit, src->calcLyapunov);
                    for(int c=0; c<N_DIM; c++) od.lf[c]=src->lf[c];
                    od.lfccdiff=src->lfccdiff;
                    od.lambda=src->lambda;
                    od.Description=src->Description;
                    od.isChaotic=src->isChaotic;
                    od.shellIndex=src->shellIndex;
                    od.orbitIndex=pointindex[i];
                    od.orbitWeight=pointmassind[i];            
                    od.Energy=src->Energy;
                    od.Ediff=src->Ediff;
                    nbodies->OrbitList.push_back(od);
                }
            }
            bool result=nbodies->exportOrbits(exportNbodyFileName, false);
            delete nbodies;
            if(!result || numPoints==0)
                return;
        }
        catch(const std::bad_alloc&) {
            return;  // not enough memory
        }
        if(stopflag) return;  
    }

    emit calcProgress("Number of bodies created="+QString::number(numPoints)+"\nnum orbits used="+QString::number(norbitsused) + 
        "\nmax sampling points="+QString::number(maxsampling) + 
        " \nAvg.X="+QString::number(sqrt(CX*CX+CY*CY+CZ*CZ)) + " (" + QString::number(Chm) + 
        ")\nAvg.V="+QString::number(sqrt(VX*VX+VY*VY+VZ*VZ)) + " (" + QString::number(Vhm) + ") [cancelled]" +
        " \nAvg.L="+QString::number(sqrt(LX*LX+LY*LY+LZ*LZ)) + " (" + QString::number(Lhm) + 
        ")\nCalculating virial ratio.."
        );
    // calculate virial ratio and other statistics
    CPotentialNB treePot(3, grid->potential->q, grid->potential->p, grid->potential->Mbh, 0, 0, TREECODE_DEFAULT_THETA, numPoints, &(points[0].front()), &(points[1].front()), &(points[2].front()), &(pointmassind.front()));
    double T=0, W=0;
    double r2max=0, v2max=0;//, rijmin=1e4; ..skip separation calculation..
    for(int i=0; i<numPoints && !stopflag; i++)
    {
        double v2 = sqr(points[3][i])+sqr(points[4][i])+sqr(points[5][i]);
        T += pointmassind[i]*v2/2;
        if(v2>v2max) v2max=v2;
        double r2 = sqr(points[0][i])+sqr(points[1][i])+sqr(points[2][i]);
        if(r2>r2max) r2max=r2;
        W += (treePot.Phi(points[0][i], points[1][i], points[2][i])/2 - treePot.Mbh/sqrt(sqr(BH_SMOOTH) + sqr(points[0][i]) + sqr(points[1][i]) + sqr(points[2][i]))/2) * pointmassind[i];
    }
    exportNbodyOrbitMass.clear();
    if(stopflag) 
    emit calcProgress("Number of bodies created="+QString::number(numPoints)+"\nnum orbits used="+QString::number(norbitsused) + "\nmax sampling points="+QString::number(maxsampling) + 
        " \nAvg.X="+QString::number(sqrt(CX*CX+CY*CY+CZ*CZ)) + " (" + QString::number(Chm) + 
        ")\nAvg.V="+QString::number(sqrt(VX*VX+VY*VY+VZ*VZ)) + " (" + QString::number(Vhm) + ") [cancelled]" +
        " \nAvg.L="+QString::number(sqrt(LX*LX+LY*LY+LZ*LZ)) + " (" + QString::number(Lhm) + ")"
        );
    else
    emit calcProgress("Number of bodies created="+QString::number(numPoints)+"\nnum orbits used="+QString::number(norbitsused) + "\nmax sampling points="+QString::number(maxsampling) + 
        " \nAvg.X="+QString::number(sqrt(CX*CX+CY*CY+CZ*CZ)) + " (" + QString::number(Chm) + 
        ")\nAvg.V="+QString::number(sqrt(VX*VX+VY*VY+VZ*VZ)) + " (" + QString::number(Vhm) + ") [cancelled]" +
        " \nAvg.L="+QString::number(sqrt(LX*LX+LY*LY+LZ*LZ)) + " (" + QString::number(Lhm) + 
        ")\nmax r="+QString::number(sqrt(r2max)) + "\nmax v="+QString::number(sqrt(v2max)) + 
        "\nT="+QString::number(T) + "\nW="+QString::number(-W) + "\nvirial ratio="+QString::number(-2*T/W)
        );
}

void CSchwarzschildModel::stop()
{
    stopflag=true;
}

/// helper class that writes NEMO-compatible snapshot file
CSnapShotWriter::CSnapShotWriter(QString name, bool append)
{
    snap.open(name.toStdString().c_str(), std::ios::binary | (append?std::ios_base::app : std::ios_base::trunc));
    level=0;
}
//                                                                            
CSnapShotWriter::~CSnapShotWriter()
{
   snap.close();
}
//                                                                            
void CSnapShotWriter::PutInt(const char*name,int var)
{
   snap.put(-110);
   snap.put(9);
   snap.put('i');
   PutZString(name);
   snap.write((const char*)&var,4);
}
//                                                                            
void CSnapShotWriter::PutDouble(const char* name,double d)
{
   snap.put(-110);
   snap.put(9);
   snap.put('d');
   PutZString(name);
   snap.write((const char*)&d,8);
}
//                                                                            
void CSnapShotWriter::PutDoubleArray(const char*name,int demi,const int*dem,const double*ar)
{
   snap.put(-110);
   snap.put(11);
   snap.put('d');
   PutZString(name);
   snap.write((const char*)dem,demi*4);
   int buf=0;
   snap.write((const char*)&buf,4);
   buf=1;
   for (int i=0;i<demi;i++)
     buf*=dem[i];
   snap.write((const char*)ar,sizeof(double)*buf);
}
//                                                                            
void CSnapShotWriter::PutFloatArray(const char*name,int demi,const int*dem,const float*ar)
{
   snap.put(-110);
   snap.put(11);
   snap.put('f');
   PutZString(name);
   snap.write((const char*)dem,demi*4);
   int buf=0;
   snap.write((const char*)&buf,4);
   buf=1;
   for (int i=0;i<demi;i++)
     buf*=dem[i];
   snap.write((const char*)ar,sizeof(float)*buf);
}
//                                                                            
void CSnapShotWriter::PutIntArray(const char*name,int demi,const int*dem,const int* ar)
{
   snap.put(-110);
   snap.put(11);
   snap.put('i');
   PutZString(name);
   snap.write((const char*)dem,demi*4);
   int buf=0;
   snap.write((const char*)&buf,4);
   buf=1;
   for (int i=0;i<demi;i++)
     buf*=dem[i];
   snap.write((const char*)ar,sizeof(int)*buf);
}
//                                                                            
void CSnapShotWriter::PutCharArray(const char* name,int demi,const int*dem,const char* ar)
{
   snap.put(-110);
   snap.put(11);
   snap.put('c');
   PutZString(name);
   snap.write((const char*)dem,demi*4);
   int buf=0;
   snap.write((const char*)&buf,4);
   buf=1;
   for (int i=0;i<demi;i++)
     buf*=dem[i];
   snap.write((const char*)ar,buf);
}
//                                                                            
void CSnapShotWriter::StartLevel(const char*name)
{
   level++;
   snap.put(-110);
   snap.put(9);
   snap.put('(');
   PutZString(name);
}
//                                                                            
void CSnapShotWriter::EndLevel()
{
   level--;
   if (level<0)
     { // should raise error
     }
   snap.put(-110);
   snap.put(9);
   snap.put(')');
   snap.put(0);
}
//                                                                            
void CSnapShotWriter::PutString(const char* name,const char* str)
{
   int len=strlen(str);
   PutCharArray(name,1,&len,str);
}
//                                                                            
void CSnapShotWriter::PutZString(const char*name)
{
   snap.put(0);
   snap.write(name,strlen(name));
   snap.put(0);
}

void CSnapShotWriter::WriteHistory(QString new_h)
{
   if (new_h.size() > 0)
       PutString("History",new_h.toStdString().c_str());
}
void CSnapShotWriter::WritePhase(
             const std::vector<float>* xv,
             const std::vector<float>& m,
			 double t)
{       
   int nbody     = m.size();
   float* phase = new float[nbody * 6];
   float* mass  = new float[nbody];
   
   for (int i = 0 ; i < nbody ; i++)
   {
        mass[i] = m[i];
        for (size_t j = 0 ; j < 6 ; j++)
            phase[ i * 6 + j ] = xv[j][i];
   }
   
   StartLevel("SnapShot");
   StartLevel("Parameters");
   PutInt("Nobj", nbody);
   PutDouble("Time", t);
   EndLevel();
   StartLevel("Particles");
   PutInt("CoordSystem",0201402);
   int tmp_dem[3];
   tmp_dem[0] = nbody;
   PutFloatArray("Mass",1,tmp_dem,mass);
   tmp_dem[0] = nbody;
   tmp_dem[1] = 2;
   tmp_dem[2] = 3;
   PutFloatArray("PhaseSpace",3,tmp_dem,phase);
   EndLevel();
   EndLevel();
   
   delete [] phase;
   delete [] mass;
}

}